/*
 * Copyright 2010, 2011, 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "AIDriver.h"
#include <cmath>
#include <cassert>

AIDriver::AIDriver()
:	maxSpeed(1000.0)
{
}

AIDriver::AIDriver(double maxSpeed)
:	maxSpeed(maxSpeed)
{
}

int AIDriver::drive(int carId, const Track& track)
{
	double nearestWallAngle;
	double nearestWallDistance = track.getCarDistanceNearestWall(carId, nearestWallAngle);

	double speed = track.getCarForwardSpeed(carId);
	
	double tileAngle = track.getCarTileAngle(carId);
		
	// car is heading towards nearest wall
	if (std::fabs(nearestWallAngle) < 0.5 * PI) 
	{
		if (std::fabs(tileAngle) > 0.1)
		{
			// car has low speed
			if (std::fabs(speed) < 3 && nearestWallDistance < 1.01 * CAR_RADIUS)
			{
				return DRIVER_BRAKE;
			}
			else if (speed < 0.0)
			{
				return DRIVER_BRAKE | (tileAngle > 0 ? DRIVER_LEFT : DRIVER_RIGHT);
			}
		}
	}
	
	int flags = 0;
	
	if (speed < 0.0)
	{
		// to avoid going out and in of above code repeatedly
		flags = DRIVER_ACCELERATE;
	}
	else if (nearestWallDistance < 2.0 + CAR_RADIUS)
	{
		flags = DRIVER_ACCELERATE;
		if (nearestWallAngle < 0.0)
		{
			flags |= DRIVER_LEFT;
		}
		else 
		{
			flags |= DRIVER_RIGHT;
		}
	}
	else
	{
		double angle;
		if (std::fabs(tileAngle) > PI_HALF)
		{
			// If the car is facing the wrong direction the best 
			// driving angle is decided by the current tile
			angle = -tileAngle;
		}
		else
		{
			angle = track.getBestDrivingAngle(carId, 2.0, 4);
		}

		if (angle < -0.2)
		{
			flags = DRIVER_RIGHT;
			if (speed > 70.5 && angle < -0.75)
			{
				flags |= DRIVER_BRAKE;
			}
			else
			{
				flags |= DRIVER_ACCELERATE;
			}
		}
		else if (angle > 0.2)
		{
			flags = DRIVER_LEFT;
			if (speed > 70.5 && angle > 0.75)
			{
				flags |= DRIVER_BRAKE;
			}
			else
			{
				flags |= DRIVER_ACCELERATE;
			}
		}
		else
		{
			flags = DRIVER_ACCELERATE;
		}
	}

	if ((flags & DRIVER_ACCELERATE) && speed > maxSpeed)
	{
		flags &= ~DRIVER_ACCELERATE;
	}

	return flags;
}

bool AIDriver::isAI() const
{
	return true;
}

AIDriver& AIDriver::getInstance(int difficultyLevel)
{
	assert(difficultyLevel >= 0 && difficultyLevel < 5);
	static AIDriver instance[5] =
	{
		AIDriver(40.0), // 0 - very easy
		AIDriver(50.0), // 1 - easy
		AIDriver(60.0), // 2 - medium
		AIDriver(68.0), // 3 - hard
		AIDriver()      // 4 - very hard
	};
	return instance[difficultyLevel];
}

