/*
 * Copyright 2010, 2011, 2013, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "CarSelectionState.h"
#include "SurfaceBank.h"
#include "AIDriver.h"
#include "StateManager.h"
#include "RaceSetup.h"
#include "JoinNetworkGame.h"
#include "Network.h"
#include "CreateNetworkGame.h"
#include "RacePreparationMenu.h"
#include "NetworkMenu.h"
#include "RaceState.h"
#include "TimeTrialMenu.h"
#include "SelectTrack.h"
#include "str.h"
#include <algorithm>

CarSelectionState::CarSelectionState(CarSelectionMode mode, Network* net, int player, Uint8 reservedCars)
:	mode(mode),
	selectedCar(0),
	player(player),
	reservedCars(reservedCars),
	net(net),
	playerText(80),
	rotation(0),
	isSelected(false),
	hoverCar(-1),
	timeoutText(160)
{	
	menu.setTitle("select your car!");
	
	nextCar();
	updateCars();
	
	playerText.addTo(this);
	menu.addRenderObjects(this);
	
	if (mode == CAR_SELECTION_NETWORK_JOIN)
	{
		timeoutStart = SDL_GetTicks();
		timeoutText.addTo(this);
		timeoutSeconds = 0;
	}
}

void CarSelectionState::setStateInfo(int)
{
	const MenuEvent& me = StateManager::getInstance().getLastMenuEvent();
	if  (me.type == MENU_MOUSE_CLICK1 ||
	     me.type == MENU_MOUSE_CLICK2)
	{
		mouseMove(me.x, me.y);
	}
}

void CarSelectionState::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	
	menu.onEvent(me);
	
	switch (me.type)
	{
	case MENU_DEC:
		prevCar();
		break;
	case MENU_INC:
		nextCar();
		break;
	case MENU_ACTION:
		selectCar();
		break;
	case MENU_ESC:
		switch (mode)
		{
		case CAR_SELECTION_NORMAL:
			sm.enter(new RacePreparationMenu(Settings::race.tournament), -1);
			break;
		case CAR_SELECTION_NETWORK_CREATE:
		case CAR_SELECTION_NETWORK_JOIN:
			net->quitGame();
			sm.enter(new NetworkMenu(net), -1);
			break;
		case CAR_SELECTION_TIME_TRIAL:
			sm.enter(new TimeTrialMenu(), -1);
		}
		break;
	case MENU_MOUSE_MOVE:
		mouseMove(me.x, me.y);
		break;
	case MENU_MOUSE_CLICK1:
		mouseClick(me.x, me.y);
		break;
	default:
		break;
	}
}

bool CarSelectionState::selectCar(int i, bool select)
{
	if (i < 0)
	{
		i += 8;
	}
	else
	{
		i = i % 8;
	}
	
	if (i == selectedCar && select == isSelected)
	{
		return true;
	}
	
	if ((reservedCars >> i) & 1)
	{
		return false;
	}
	
	rotation = 0.0;
	renderObjects[selectedCar].setRotation(0.0);
	isSelected = select;
	selectedCar = i;
	return true;
}

void CarSelectionState::nextCar()
{
	for (int i = (isSelected ? 1 : 0); i < 8; ++i)
	{
		if (selectCar(selectedCar + i))
		{
			break;
		}
	}
}
void CarSelectionState::prevCar()
{
	for (int i = (isSelected ? 1 : 0); i < 8; ++i)
	{
		if (selectCar(selectedCar - i))
		{
			break;
		}
	}
}

void CarSelectionState::selectCar()
{
	chosenCars.push_back(selectedCar);
	reservedCars |= (1 << selectedCar);
	++player;
	
	if (proceed())
	{
		return;
	}
	
	updateCars();
	nextCar();
}

void CarSelectionState::updateCars()
{
	if (mode == CAR_SELECTION_TIME_TRIAL)
	{
		playerText.setText(Settings::timetrial.name);
	}
	else
	{
		playerText.setText(Settings::players.humanNames[player]);
	}
	
	std::vector<int> carIds;
	for (int i = 0; i < 8; i++)
	{
		removeRenderObject(renderObjects[i]);
		if (!((reservedCars >> i) & 1))
		{
			carIds.push_back(i);
		}
	}
	
	const int CAR_SPACE = 20;
	int w = carIds.size() * CAR_SPACE - 13;
	int x = (GAME_WIDTH - w) / 2;
	
	for (std::size_t i = 0; i < carIds.size(); i++)
	{
		int j = carIds[i];
		renderObjects[j].setSurface(SurfaceBank::getCar(j));
		renderObjects[j].moveTo(x, 110);
		x += CAR_SPACE;
		addRenderObject(renderObjects[j]);
	}
}

void CarSelectionState::mouseMove(int x, int y)
{
	for (int i = 0; i < 8; ++i)
	{
		if (renderObjects[i].isHit(x, y) && selectCar(i))
		{
			hoverCar = i;
			return;
		}
	}
	if (hoverCar != -1)
	{
		selectCar(selectedCar, false);
		hoverCar = -1;
	}
}

void CarSelectionState::mouseClick(int x, int y)
{
	hoverCar = selectedCar;
	int oldSelectedCar = selectedCar;
	mouseMove(x, y);
	if (oldSelectedCar == selectedCar && hoverCar == selectedCar)
	{
		selectCar();
	}
}

void CarSelectionState::update()
{
	if (mode == CAR_SELECTION_NORMAL && proceed())
	{
		return;
	}
	
	if (net)
	{
		net->step();
	}
	
	menu.update();
	
	if (isSelected)
	{
		rotation += 0.07;
		renderObjects[selectedCar].setRotation(rotation);
	}
	
	if (mode == CAR_SELECTION_NETWORK_JOIN)
	{
		int newTimeoutSeconds = (timeoutStart + 16000 - static_cast<int>(SDL_GetTicks())) / 1000;
		
		if (newTimeoutSeconds != timeoutSeconds)
		{
			timeoutSeconds = newTimeoutSeconds;
			if (timeoutSeconds <= 0)
			{
				selectCar();
			}
			else
			{
				timeoutText.setText("The car will be autoselected in " + int2string(timeoutSeconds) + " seconds");
			}
		}
	}
}

bool CarSelectionState::proceed()
{
	StateManager& sm = StateManager::getInstance();
	
	switch (mode)
	{
	case CAR_SELECTION_NORMAL:
		if (player >= Settings::race.humanPlayers)
		{
			// give remaining cars to the computer players randomly
			for (int i = 0; i < 8; i++)
			{
				if (!((reservedCars >> i) & 1))
				{
					chosenCars.push_back(i);
				}
			}
			std::random_shuffle(chosenCars.begin() + Settings::race.humanPlayers, chosenCars.end());
			chosenCars.resize(Settings::race.humanPlayers + Settings::race.computerPlayers);
			
			std::vector<Driver*> drivers;
			std::vector<std::string> names;
			for (int i = 0; i < Settings::race.humanPlayers; i++)
			{
				drivers.push_back(sm.getDriver(i));
				names.push_back(Settings::players.humanNames[i]);
			}
			for (int i = 0; i < Settings::race.computerPlayers; i++)
			{
				int difficulty = Settings::players.computerDifficulties[i];
				if (difficulty < 0)
				{
					difficulty = Settings::race.difficulty;
				}
				drivers.push_back(&AIDriver::getInstance(difficulty));
				names.push_back(Settings::players.computerNames[i]);
			}
			sm.enter(new RaceState(Map::generate(), RaceSetup(drivers, chosenCars, names, Settings::race.laps, Settings::race.tracks, Settings::race.tournament, Settings::race.collisions)), 0);
			return true;
		}
		break;
	case CAR_SELECTION_NETWORK_CREATE:
		if (player >= Settings::network.players)
		{
			net->createGame(reservedCars, chosenCars);
			sm.enter(net->getNewGameState(), 0);
			return true;
		}
		break;
	case CAR_SELECTION_NETWORK_JOIN:
		sm.enter(new JoinNetworkGame(net, chosenCars.back()), 0);
		return true;
	case CAR_SELECTION_TIME_TRIAL:
		Settings::timetrial.car = selectedCar;
		sm.enter(new SelectTrack(), 0);
		return true;
	}
	return false;
}
