/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ChatInput.h"
#include "GameState.h"
#include "Settings.h"
#include "StateManager.h"
#include "Font.h"
#include <cctype>

ChatInput::ChatInput()
:	ready(false),
	gameState(),
	iPlayer(0),
	textWidth(0)
{
	renderObject.moveTo(2 , std::max(120U, (Settings::network.maxChatMessages + 1) * (Settings::network.chatFont ? 10 : 8)));
	addDefaultPlayers();
}

void ChatInput::setGameState(GameState* newGameState)
{
	if (gameState)
	{
		gameState->removeRenderObject(renderObject);
	}
	if (newGameState)
	{
		newGameState->addRenderObject(renderObject);
	}
	gameState = newGameState;
}

bool ChatInput::onEvent(const SDL_Event& event)
{
	if (!gameState)
	{
		return false;
	}
	
	if (event.type == SDL_KEYDOWN)
	{
		if (event.key.keysym.sym == SDLK_UP)
		{
			if (!text.empty() && !ready)
			{
				if (++iPlayer >= players.size())
				{
					iPlayer = 0;
				}
				updateRenderObject();
				return true;
			}
		}
		else if (event.key.keysym.sym == SDLK_DOWN)
		{
			if (!text.empty() && !ready)
			{
				if (--iPlayer >= players.size())
				{
					iPlayer = players.size() - 1;
				}
				updateRenderObject();
				return true;
			}
		}
		else if (event.key.keysym.sym == SDLK_BACKSPACE)
		{
			if (!text.empty())
			{
				if (SDL_GetModState() & KMOD_CTRL)
				{
					std::size_t p = text.find_last_not_of(' ');
					if (p == std::string::npos)
					{
						p = 0;
					}
					else
					{
						p = text.find_last_of(' ', p);
						if (p == std::string::npos)
						{
							p = 0;
						}
						else
						{
							p += 1;
						}
					}
					text.resize(p);
				}
				else
				{
					text.resize(text.size() - 1);
				}
				updateRenderObject();
				return true;
			}
		}
		else if (event.key.keysym.sym == SDLK_KP_ENTER || 
		         event.key.keysym.sym == SDLK_RETURN)
		{
			if (!ready && !text.empty())
			{
				ready = true;
				return true;
			}
		}
		else if (event.key.keysym.sym == SDLK_ESCAPE)
		{
			if (!text.empty())
			{
				text.clear();
				updateRenderObject();
				return true;
			}
		}
		else if (!ready)
		{
			char c = unicode2char(event.key.keysym.unicode);
			int newTextWidth = textWidth + getFont(Settings::network.chatFont ? BIG_FONT : SMALL_FONT).calculateTextWidth(std::string(1, c));
			if (std::isprint(c) && newTextWidth < GAME_WIDTH)
			{
				text += c;
				updateRenderObject();
				return true;
			}
		}
	}

	return false;
}

void ChatInput::updateRenderObject()
{
	if (empty())
	{
		return;
	}
	Surface textSurface(text.empty() ? "" : players[iPlayer].name + ": " + text, Settings::network.chatFont ? BIG_FONT : SMALL_FONT, carColor(players[iPlayer].car), TEXT_BORDER_COLOR);
	textWidth = textSurface.width();
	renderObject.setSurface(textSurface);
}

bool ChatInput::isReady() const
{
	return ready;
}

std::string ChatInput::fetch()
{
	std::string retStr;
	if (!text.empty() && !empty())
	{
		retStr = players[iPlayer].name + ": " + text;
		text.clear();
		updateRenderObject();
		ready = false;
	}
	return retStr;
}

Color ChatInput::fetchCar() const
{
	return empty() ? -1 : players[iPlayer].car;
}

void ChatInput::addPlayer(const std::string& name, int car)
{
	Player p;
	p.name = name;
	p.car = car;
	players.push_back(p);
}

void ChatInput::removePlayer(int car)
{
	std::vector<Player>::iterator it;
	for (it = players.begin(); it !=  players.end(); ++it)
	{
		if (it->car == car)
		{
			if (it == players.begin() + iPlayer)
			{
				iPlayer = 0;
				updateRenderObject();
			}
			players.erase(it);
			break;
		}
	}
}

bool ChatInput::empty() const
{
	return players.empty();
}

void ChatInput::clear()
{
	setGameState(0);
	text.clear();
	ready = false;
	players.clear();
	renderObject.setSurface(Surface());
}

void ChatInput::addDefaultPlayers()
{
	for (int i = 0; i < Settings::network.players; ++i)
	{
		addPlayer(Settings::players.humanNames[i]);
	}
}
