/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "GraphicsSettingsMenu.h"
#include "StateManager.h"
#include "SettingsMenu.h"
#include <sstream>
#include <algorithm>

std::string int2string(int);

bool resolutionSorter(const SDL_Rect& r1, const SDL_Rect& r2)
{
	if (r1.w == r2.w)
	{
		return r1.h < r2.h;
	}
	else
	{
		return r1.w < r2.w;
	}
}

// The scaler order in the settings menu has been changed. To maintain 
// backward compatibility with old settings files it is important
// that the ID of the scalers remain the same. This array can be used
// to map between the position number and the ID.
static int SCALER_ID_MAP[] = {0, 2, 1, 3};

GraphicsSettingsMenu::GraphicsSettingsMenu()
:	scaler(SCALER_ID_MAP[Settings::graphics.scaler]),
	stretch(Settings::graphics.stretch ? (Settings::graphics.keepAspectRatio ? 1 : 2) : 0),
	fullscreen(Settings::graphics.fullscreen),
	pointer(0),
	scalePointer(Settings::graphics.scalePointer),
	oldSettings(Settings::graphics),
	save(false),
	cancel(false)
{
	menu.setTitle("Graphics Settings");
	
	menu.addItem("Scaling", &scaler, 0, 3, "nearest neighbor", "hqx", "bilinear", "bilinear (less blur)");
	menu.addItem("Stretch", &stretch, 0, 2, "no", "keep aspect ratio", "fill");
	
	const std::vector<std::string>& pointers = Pointer::getList();
	std::vector<std::string>::const_iterator it = std::find(pointers.begin(), pointers.end(), Settings::graphics.pointer);
	if (it != pointers.end())
	{
		pointer = it - pointers.begin();
	}
	menu.addItem("Pointer type", &pointer, pointers);
	menu.addItem("Scale pointer", &scalePointer, 0, 2, "no", "only in fullscreen", "always");
	
	menu.addItem("Start game in", &fullscreen, 0, 1, "window", "fullscreen");
	
	std::vector<std::string> resolutions;
	resolutions.push_back("same as desktop");
	SDL_Rect** modes = SDL_ListModes(0, SDL_FULLSCREEN);
	if(!modes || modes == reinterpret_cast<SDL_Rect**>(-1))
	{
		for(int i = 1; i <= 5; ++i)
		{				
			resolutions.push_back(int2string(GAME_WIDTH * i) + "*" + int2string(GAME_HEIGHT * i));
		}
	}
	else
	{
		std::vector<SDL_Rect> rects;
		for(int i = 0; modes[i]; ++i)
		{
			rects.push_back(*modes[i]);
		}
		std::sort(rects.begin(), rects.end(), resolutionSorter);
		for(std::size_t i = 0; i < rects.size(); ++i)
		{
			resolutions.push_back(int2string(rects[i].w) + "*" + int2string(rects[i].h));
		}
	}
	menu.addItem("Fullscreen resolution", &resolution, 12, resolutions);
	
	menu.addItem("Save settings", &save);
	menu.addItem("Cancel", &cancel);
	
	menu.addRenderObjects(this);
	
	onValueUpdate(&resolution);
}

GraphicsSettingsMenu::~GraphicsSettingsMenu()
{
	if (!save)
	{
		StateManager::getInstance().updateGraphicsSettings(oldSettings);
	}
}

void GraphicsSettingsMenu::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	
	if (menu.onEvent(me))
	{
		if (save)
		{			
			sm.enter(new SettingsMenu(), -1);
			return;
		}
	}
	else if (me.type == MENU_ESC)
	{
		cancel = true;
	}
	if (cancel)
	{
		// The settings are restored in the destructor, not here.
		sm.enter(new SettingsMenu(), -1);
	}
}

void GraphicsSettingsMenu::validateTypedValue(const std::string* strValue, std::string& inputString)
{
	if (strValue == &resolution && !inputString.empty())
	{
		std::size_t i = inputString.size() - 1;
		char c = inputString[i];
		if (!std::isdigit(c))
		{
			if ((c == 'x' || c == ' ' || c == '*') && i > 0 && std::find(inputString.begin(), inputString.end() - 1, '*') == inputString.end() - 1)
			{
				inputString[i] = '*';
			}
			else
			{
				inputString.resize(i);
			}
		}
	}
}

void GraphicsSettingsMenu::onValueUpdate(int* intValue)
{
	StateManager& sm = StateManager::getInstance();
	if (intValue == &scaler)
	{
		sm.updateScaler(SCALER_ID_MAP[scaler]);
	}
	else if (intValue == &stretch)
	{
		bool newStretch = (stretch > 0);
		bool newKeepAspectRatio = (stretch == 1);
		sm.updateStretch(newStretch, newKeepAspectRatio);
	}
	else if (intValue == &fullscreen)
	{
		// This will not have any effect until next time the game is started.
		Settings::graphics.fullscreen = fullscreen;
	}
	else if (intValue == &pointer)
	{
		sm.updatePointerType(Pointer::getList()[pointer]);
	}
	else if (intValue == &scalePointer)
	{
		sm.updatePointerScaling(scalePointer);
	}
}

void GraphicsSettingsMenu::onValueUpdate(std::string* strValue)
{
	if (strValue == &resolution)
	{		
		int w = 0;
		int h = 0;
		std::istringstream iss(*strValue);
		if (iss >> w && iss.ignore() && iss >> h && w != 0 && h != 0)
		{
			Settings::graphics.fullscreenWidth  = std::min(w, 10000);
			Settings::graphics.fullscreenHeight = std::min(h, 10000);
		}
		
		if (resolution.find("desktop") != std::string::npos ||
		    Settings::graphics.fullscreenWidth == 0 || 
		    Settings::graphics.fullscreenHeight == 0)
		{
			Settings::graphics.fullscreenWidth  = 0;
			Settings::graphics.fullscreenHeight = 0;
			resolution = "same as desktop";
		}
		else
		{
			resolution = int2string(Settings::graphics.fullscreenWidth) + 
			             "*" + 
			             int2string(Settings::graphics.fullscreenHeight);
		}
		menu.notifyValueUpdate(&resolution);
	}
}
