/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "HumanPlayerSettingsMenu.h"
#include "StateManager.h"
#include "Driver.h"
#include "InputDevice.h"
#include "Settings.h"
#include "SettingsMenu.h"
#include <sstream>

static std::string controlLabel[] = 
{
	"Accelerate: ",
	"Brake: ",
	"Left: ",
	"Right: "
};

HumanPlayerSettingsMenu::HumanPlayerSettingsMenu()
:	player(1),
	device(0),
	saveAction(false),
	cancelAction(false),
	mode(-1)
{
	menu.setTitle("Human Player Settings");
	
	menu.addItem("Player", &player, 1, 8);
	menu.addItem("Name", &name);
	
	StateManager& sm = StateManager::getInstance();
	std::vector<std::string> deviceNames = sm.getDeviceNames();
	menu.addItem("Device", &device, deviceNames);
	
	for (int i = 0; i < 4; ++i)
	{
		enterControlMode[i] = false;
		menu.addItem(controlLabel[i], &enterControlMode[i]);
	}
	
	menu.addItem("Save settings", &saveAction);
	menu.addItem("Cancel", &cancelAction);
	
	menu.addRenderObjects(this);
	
	for (int i = 0; i < 8; ++i)
	{
		names[i] = Settings::players.humanNames[i];
		
		drivers[i].resize(deviceNames.size());
		Driver* driver = sm.getDriver(i);
		int deviceIndex = driver->getDeviceIndex();
		
		Driver* driverCopy = sm.getDevice(deviceIndex)->createDriver();
		driverCopy->setDeviceIndex(deviceIndex);
		std::stringstream ss;
		driver->save(ss);
		ss.ignore(80, '\n');
		driverCopy->load(ss);
		
		drivers[i][deviceIndex] = driverCopy;
		deviceIndices[i] = deviceIndex;
	}
	updatePlayer();
}

HumanPlayerSettingsMenu::~HumanPlayerSettingsMenu()
{
	StateManager& sm = StateManager::getInstance();
	for (int i = 0; i < 8; ++i)
	{
		for (std::size_t j = 0; j < drivers[i].size(); ++j)
		{
			if (sm.getDriver(i) != drivers[i][j])
			{
				delete drivers[i][j];
			}
		}
	}
}

void HumanPlayerSettingsMenu::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	
	if (mode != -1)
	{
		int i = player - 1;
		Driver* driver = drivers[i][deviceIndices[i]];
		if (driver->setControl(event, mode))
		{
			menu.makeActive(ITEM_SELECTED);
			updateControl(mode);
			mode = -1;
			return;
		}
		
		if (me.type == MENU_ESC)
		{
			menu.makeActive(ITEM_SELECTED);
			mode = -1;
			return;
		}
	}
	else
	{
		int oldPlayer = player;
		int oldDevice = device;
		
		if (menu.onEvent(me))
		{
			if (saveAction)
			{
				for (int i = 0; i < 8; ++i)
				{
					for (std::size_t j = 0; j < drivers[i].size(); ++j)
					{
						if (drivers[i][j])
						{
							if (deviceIndices[i] == int(j))
							{
								sm.setDriver(i, drivers[i][j]);
							}
						}
					}
					Settings::players.humanNames[i] = names[i];
				}
				sm.enter(new SettingsMenu(), -1);
			}
			else if (cancelAction)
			{
				sm.enter(new SettingsMenu(), -1);
			}
			else if (player != oldPlayer)
			{
				updatePlayer();
			}
			else if (device != oldDevice)
			{
				updateDevice();
			}
			else
			{
				for (int i = 0; i < 4; ++i)
				{
					if (enterControlMode[i])
					{
						menu.makeActive();
						enterControlMode[i] = false;
						mode = i;
						break;
					}
				}
			}
		}
		else if (me.type == MENU_ESC)
		{
			sm.enter(new SettingsMenu(), -1);
		}
	}
}

void HumanPlayerSettingsMenu::onValueUpdate(std::string* strValue)
{
	if (strValue == &name)
	{
		names[player - 1] = name;
	}
}

void HumanPlayerSettingsMenu::updatePlayer()
{
	int i = player - 1;
	device = deviceIndices[i];
	menu.notifyValueUpdate(&device);
	updateDevice();
	name = names[i];
	menu.notifyValueUpdate(&name);
}

void HumanPlayerSettingsMenu::updateDevice()
{
	int i = player - 1;
	deviceIndices[i] = device;
	Driver*& driver = drivers[i][device];
	if (!driver)
	{
		driver = StateManager::getInstance().getDevice(device)->createDriver();
		driver->setDeviceIndex(device);
	}
	for (int i = 0; i < 4; ++i)
	{
		updateControl(i);
	}
}

void HumanPlayerSettingsMenu::updateControl(int i)
{
	Driver* driver = drivers[player - 1][deviceIndices[player - 1]];
	menu.setLabel(3 + i, controlLabel[i] + driver->getControlName(i));
}
