/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Internet.h"
#include "Settings.h"
#include "Version.h"
#include <sstream>
#include <algorithm>

const Uint32 MIN_INTERVAL =  1000; // 1 second
const Uint32 MAX_INTERVAL = 60000; // 1 minute

const Uint32 SERVER_REFRESH = 600000; // 10 minutes

// Bots should use a special range of IDs so that they can be
// easily identified and ignored in the game stats. This is mainly 
// for allowing programs to help with TCP hole punching.
const Uint32 BOT_ID_RANGE_BEGIN = 4000000000U;

// C++03 doesn't guarantee 64 bit stream operators so
// here are a few simplified overloads just in case.

// This wrapper class is used to avoid ambiguity. If a match is found in
// the standard library it will be used instead.
struct U64StreamWrapper
{
	U64StreamWrapper(Uint64& ref) : ref(ref) {}
	Uint64& ref;
};

std::istream& operator>>(std::istream& is, U64StreamWrapper w)
{
	Uint64& value = w.ref;
	value = 0;
	is >> std::ws;
	if (!std::isdigit(is.peek()))
	{
		is.setstate(std::ios::failbit);
	}
	else
	{
		do
		{
			value = (is.get() - '0') + value * 10;
		} while (std::isdigit(is.peek()));
	}
	return is;
}

std::ostream& operator<<(std::ostream& os, U64StreamWrapper w)
{
	Uint64 value = w.ref;
	std::string str;
	do
	{
		int digit = value % 10;
		str += '0' + digit;
		value /= 10;
	} while (value != 0);
	std::reverse(str.begin(), str.end());
	return os << str;
}

std::string Internet::serverMessage(const std::string& data)
{
	// It's fine to access the server info from the thread because they 
	// are not modified while the network is running.
	return "GET " + serverPath + " HTTP/1.0\r\n"
	       "Host: " + serverHost + "\r\n"
	       "Connection: close\r\n"
	       "User-Agent: BrumBrumRally/" BBR_VERSION "\r\n"
	       "Brum-Data: " + data + "\r\n"
	       "\r\n";
}

int server_thread(void* arg)
{
	Internet* net = static_cast<Internet*>(arg);
	IPaddress ip;
	
	// Resolve server name
	while (SDLNet_ResolveHost(&ip, net->serverHost.c_str(), net->serverPort) != 0) 
	{
		SDL_mutexP(net->mutex);		
		do 
		{
			net->refresh = false;
			net->newStatus = "Server not found!";
			if (net->stopThread)
			{
				SDL_mutexV(net->mutex);
				return 0;
			}
			SDL_CondWait(net->cond, net->mutex);
		} while (!net->refresh);
		SDL_mutexV(net->mutex);
	}

	// Fetch server list
	while (true)
	{
		SDL_mutexP(net->mutex);
		if (net->stopThread)
		{
			SDL_mutexV(net->mutex);
			break;
		}
		SDL_mutexV(net->mutex);
		
		TCPsocket server = SDLNet_TCP_Open(&ip);
		
		if (!server)
		{
			SDL_mutexP(net->mutex);
			net->newStatus = "No server connection!";
			SDL_mutexV(net->mutex);
		}
		else
		{
			std::ostringstream requestData;
			
			SDL_mutexP(net->mutex);
			requestData << net->playerID << ',' << net->privatePort << ',' << net->publicPort;
			SDL_mutexV(net->mutex);
			
			std::string request = net->serverMessage(requestData.str());
			
			SDLNet_TCP_Send(server,  request.c_str(), request.size());
			
			const int BUFFER_SIZE = 128;
			const std::size_t MAX_RESPONSE_LENGTH = 5000;
			char buffer[BUFFER_SIZE];
			
			int count;
			std::string response;
			while ((count = SDLNet_TCP_Recv(server, buffer, BUFFER_SIZE)) > 0)
			{
				response.insert(response.size(), buffer, count);
				if (response.length() >= MAX_RESPONSE_LENGTH)
				{
					break;
				}
			}
			
			SDLNet_TCP_Close(server);
			std::istringstream iss(response);
			
			// Check response code
			std::string line;
			std::getline(iss, line);
			if (line.find("200") == std::string::npos)
			{
				SDL_mutexP(net->mutex);
				net->newStatus = "Server error!";
				SDL_mutexV(net->mutex);
			}
			else
			{
				// Skip to content
				while (std::getline(iss, line) && line.size() > 1);
				
				Uint64 newPlayerID;
				if (iss >> newPlayerID)
				{
					std::vector<ServerPlayer> newList;
					Uint64 id;
					Uint32 ip;
					Uint16 port;
					while (iss >> id >> ip >> port && newList.size() < Settings::network.maxConnections)
					{
						ServerPlayer sp;
						sp.id = id;
						sp.connectAddress = createAddress(ip, port);
						sp.nextConnectionAttempt = 0;
						sp.nextConnectionInterval = MIN_INTERVAL;
						newList.push_back(sp);
					}
					
					SDL_mutexP(net->mutex);
					net->newList.swap(newList);
					net->newPlayerID = newPlayerID;
					net->newListReady = true;
					net->listTime = SDL_GetTicks();
					SDL_mutexV(net->mutex);
				}
				else
				{
					iss.clear();
					SDL_mutexP(net->mutex);
					std::getline(iss, net->newStatus);
					SDL_mutexV(net->mutex);
				}
			}
		}
		
		SDL_mutexP(net->mutex);
		net->refresh = false;
		do 
		{
			if (net->stopThread)
			{
				SDL_mutexV(net->mutex);
				break;
			}
			SDL_CondWait(net->cond, net->mutex);
		} while (!net->refresh);
		SDL_mutexV(net->mutex);
	}
	
	// Remove from server list
	TCPsocket server = SDLNet_TCP_Open(&ip);
	if (server)
	{
		std::ostringstream requestData;
		SDL_mutexP(net->mutex);
		requestData << "quit," << net->playerID;
		SDL_mutexV(net->mutex);

		std::string request = net->serverMessage(requestData.str());
		SDLNet_TCP_Send(server,  request.c_str(), request.size());
		SDLNet_TCP_Close(server);
	}
	
	return 0;
}


Internet::Internet()
:	iPlayer(0),
	privatePort(0),
	publicPort(0),
	playerID(0),
	thread(0),
	mutex(SDL_CreateMutex()),
	cond(SDL_CreateCond()),
	refresh(false),
	stopThread(false),
	newListReady(false),
	listTime(0),
	newPlayerID(0),
	serverPort(80)
{
	if (!error.empty())
	{
		return;
	}
	
	if (Settings::network.internetPorts.empty())
	{
		for (privatePort = 65535; privatePort >= 1024; --privatePort)
		{
			if ((socket = SDLNet_UDP_Open(privatePort)))
			{
				break;
			}
		}
	}
	else
	{
		for (std::size_t i = 0; i < Settings::network.internetPorts.size(); ++i)
		{
			privatePort = Settings::network.internetPorts[i];
			if ((socket = SDLNet_UDP_Open(privatePort)))
			{
				break;
			}
		}
	}
	publicPort = privatePort;
	
	// Parse server info
	std::string server = Settings::network.server;
	
	// Remove protocol
	const std::string protocol = "http://";
	if (server.compare(0, protocol.size(), protocol) == 0)
	{
		server.erase(server.begin(), server.begin() + protocol.size());
	}
	
	// Split path and host
	std::size_t pos = server.find('/');
	if (pos != std::string::npos)
	{
		serverHost = server.substr(0, pos);
		serverPath = server.substr(pos);
	}
	else
	{
		serverHost = server;
		serverPath = "/";
	}
	
	// If path is a directory add default server filename.
	if (serverPath[serverPath.size() - 1] == '/')
	{
		serverPath += "brumnet.php";
	}
	
	// Get protocol
	pos = serverHost.find(':');
	if (pos != std::string::npos)
	{
		std::istringstream iss(serverHost.substr(pos + 1));
		Uint16 port;
		if (iss >> port && port > 0)
		{
			serverPort = port;
		}
		serverHost.resize(pos);
	}
	
	if (!socket)
	{
		error = "No available ports!";
	}
	else
	{
		thread = SDL_CreateThread(server_thread, this);
		if (!thread)
		{
			error = "No thread!";
		}
		else
		{
			setStatus("Connecting...");
		}
	}
}

Internet::~Internet()
{
	SDL_mutexP(mutex);
	stopThread = true;
	SDL_CondSignal(cond);
	SDL_mutexV(mutex);
	
	SDL_WaitThread(thread, 0);
	
	SDL_DestroyMutex(mutex);
	SDL_DestroyCond(cond);
}

void Internet::step()
{
	if (connectionType != NET_GAME_PLAYER)
	{
		SDL_mutexP(mutex);
		if (newListReady)
		{
			playerList.swap(newList);
			newList.clear();
			playerID = newPlayerID;
			setStatus("");
			newListReady = false;
		}
		if (!newStatus.empty())
		{
			setStatus(newStatus);
			newStatus.clear();
		}
		SDL_mutexV(mutex);

		// Try connect to players in list
		if (!playerList.empty() && connections.size() < Settings::network.maxConnections)
		{
			if (iPlayer >= playerList.size())
			{
				iPlayer = 0;
			}
			
			ServerPlayer& p = playerList[iPlayer];
			
			if (!isConnected(p.id) && p.nextConnectionAttempt <= SDL_GetTicks())
			{
				msg.reset();
				msg.write8(MSG_I_CONNECT);
				msg.write64(playerID);
				msg.write64(p.id);
				msg.write16(p.connectAddress.host != INADDR_BROADCAST ? getPort(p.connectAddress) : 0);
				send(p.connectAddress);
				
				p.nextConnectionAttempt = SDL_GetTicks() + p.nextConnectionInterval;
				p.nextConnectionInterval = std::min(2 * p.nextConnectionInterval, MAX_INTERVAL);
			}
			
			++iPlayer;
		}
	}
	
	Network::step();
}

void Internet::handleMessage(Uint8 msgType, IPaddress address)
{
	switch (msgType)
	{
		case MSG_I_CONNECT:
		case MSG_I_CONNECT_ACK:
		{
			if (connectionType == NET_GAME_PLAYER)
			{
				break;
			}

			if (connections.size() >= Settings::network.maxConnections)
			{
				break;
			}

			Uint64 srcId = msg.read64();
			Uint64 dstId = msg.read64();

			if (dstId != playerID)
			{
				break;
			}

			Uint16 dstPort = msg.read16();
			bool isLocalConnection = (dstPort == 0);
			
			if (!isConnected(srcId))
			{
				Connection& connection = connections[address];
				connection.isBot = (srcId >= BOT_ID_RANGE_BEGIN);
				if (connection.lastRecive != 0)
				{
					// The connection is already established but with another ID.
					
					std::map<Uint64, ConnectedPlayer>::iterator it;
					for (it = idMap.begin(); it != idMap.end(); ++it)
					{
						if (it->second.address == address)
						{
							if (srcId > it->first)
							{
								// This is a more up-to-date connection
								// so we better get rid of the old one.
								idMap.erase(it);
								
								// Reset some of the necessary fields.
								connection.lastChatMessageRecived = 0;
								connection.lastChatMessageDelivered = 0;
								connection.lastTimestamp = 0;
								
								std::map<IPaddress, JoinInProgress>::iterator joinIt = joinInProgress.find(address);
								if (joinIt != joinInProgress.end())
								{
									joinIt->second.lastTimestamp = 0;
								}
								
								// We don't touch gameInfo because it
								// might need special clean-up code.
								
								break;
							}
							else
							{
								// We already have a more up-to-date connection
								// so we better just ignore this one.
								// This should normally never happen.
								return;
							}
						}
					}
					
				}
				connection.lastRecive = SDL_GetTicks();
					
				ConnectedPlayer& cp = idMap[srcId];
				cp.address = address;
				cp.port = dstPort;
			}
			
			if (!isLocalConnection && dstPort != publicPort)
			{
				std::map<Uint16, int> portCounts;
				std::map<Uint64, ConnectedPlayer>::iterator it;
				
				// Count the public ports.
				for (it = idMap.begin(); it != idMap.end(); ++it)
				{
					if (it->second.port != 0)
					{
						++portCounts[it->second.port];
					}
				}
				
				int maxCount = portCounts[publicPort];
				Uint16 newPort = publicPort;
				std::map<Uint16, int>::iterator countIt;
				
				// Find the most common public port.
				for (countIt = portCounts.begin(); countIt != portCounts.end(); ++countIt)
				{
					if (countIt->second > maxCount)
					{
						maxCount = countIt->second;
						newPort = countIt->first;
					}
				}
				
				if (newPort != publicPort)
				{
					// Update public port.
					SDL_mutexP(mutex);
					publicPort = newPort;
					refresh = true;
					SDL_CondSignal(cond);
					SDL_mutexV(mutex);
				}
			}
			
			if (msgType == MSG_I_CONNECT)
			{
				msg.reset();
				msg.write8(MSG_I_CONNECT_ACK);
				msg.write64(playerID);
				msg.write64(srcId);
				msg.write16(isLocalConnection ? 0 : getPort(address));
				send(address);
			}
			
			break;
		}
		default:
		{
			Network::handleMessage(msgType, address);
		}
	}
}

void Internet::onConnectionTimeout(const IPaddress& addr)
{
	std::map<Uint64, ConnectedPlayer>::iterator it;
	for (it = idMap.begin(); it != idMap.end(); ++it)
	{
		if (it->second.address == addr)
		{
			idMap.erase(it);
			break;
		}
	}
}

void Internet::refreshAsNeeded()
{
	Uint32 now = SDL_GetTicks();
	if (connectionType == NET_GAME_PLAYER)
	{
		// This happens when we change from GAME_PLAYER to LOBBY.
		
		SDL_mutexP(mutex);
		if (listTime + SERVER_REFRESH <= now)
		{
			refresh = true;
			setStatus("Refreshing...");
			SDL_CondSignal(cond);
		}
		SDL_mutexV(mutex);
		
		if (!refresh)
		{
			std::vector<ServerPlayer>::iterator it;
			for (it = playerList.begin(); it != playerList.end(); ++it)
			{
				it->nextConnectionAttempt = 0;
				it->nextConnectionInterval = MIN_INTERVAL;
			}
		}
	}
	else if (connectionType == NET_GAME_OWNER)
	{
		// This happens after GAME_OWNER has finished a race.
		
		SDL_mutexP(mutex);
		if (listTime + SERVER_REFRESH <= now)
		{
			refresh = true;
			// We don't show any message here because it can be disturbing 
			// while playing but if there is an error it will be shown.
			SDL_CondSignal(cond);
		}
		SDL_mutexV(mutex);
	}
}

bool Internet::isConnected(Uint64 id) const
{
	return idMap.find(id) != idMap.end();
}

std::string Internet::name() const
{
	return "Internet";
}
