/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "JoinNetworkGame.h"
#include "StateManager.h"
#include "Network.h"
#include "StateManager.h"
#include "NetworkMenu.h"
#include "ErrorState.h"
#include "CarSelectionState.h"
#include "JoinNetworkGameFinal.h"

const Uint32 JOIN_INTERVALL = 200;
const int JOIN_ATTEMPTS     = 25;

JoinNetworkGame::JoinNetworkGame(Network* net, const IPaddress& joinAddress)
:	net(net),
	lastSend(0),
	attempts(0),
	selectCar(-1),
	joinState(JOIN_NOT_STARTED),
	waitText("Please wait...", 100)
{
	net->setGameState(this, false);
	
	menu.setTitle("Joining game");
	
	waitText.addTo(this);
	
	menu.addRenderObjects(this);
	net->join(joinAddress); // Should this really be done here?
}

JoinNetworkGame::JoinNetworkGame(Network* net, int car)
:	net(net),
	lastSend(0),
	attempts(0),
	selectCar(car),
	waitText("Please wait...", 100)
{
	JoinStatus status;
	net->getJoinStatus(status);
	joinState = status.state;
	
	net->setGameState(this, false);
	
	if (status.state == JOIN_MORE_CARS)
	{
		menu.setTitle("Selecting car");
	}
	else
	{
		menu.setTitle("Joining game");
	}
	
	waitText.addTo(this);
	
	menu.addRenderObjects(this);
}

void JoinNetworkGame::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	menu.onEvent(me);
	if (me.type == MENU_ESC)
	{
		net->quitGame();
		sm.enter(new NetworkMenu(net), -1);
	}
}

void JoinNetworkGame::update()
{
	menu.update();
	net->step();
	
	JoinStatus status;
	
	StateManager& sm = StateManager::getInstance();
	
	if (net->getJoinStatus(status))
	{
		if (status.state == JOIN_ERROR)
		{
			if (joinState == JOIN_NOT_STARTED)
			{
				sm.enter(new ErrorState("Join error", "The game is full!", new NetworkMenu(net)), 0);
			}
			else
			{
				sm.enter(new ErrorState("Join error", "The game could not be joined!", new NetworkMenu(net)), 0);
			}
			return;
		}
		
		if (status.state == JOIN_HAS_ENDED)
		{
			net->join();
			sm.enter(net->getNewGameState(), 0);
			return;
		}
		
		if (joinState != JOIN_NO_MORE_CARS && status.state == JOIN_NO_MORE_CARS)
		{
			if (status.joinedPlayers < Settings::network.players)
			{
				sm.enter(new JoinNetworkGameFinal(net, status.joinedPlayers), 0);
			}
			else
			{
				joinState = JOIN_NO_MORE_CARS;
				attempts = 0;
				lastSend = 0;
			}
			return;
		}
		
		if (status.state == JOIN_MORE_CARS)
		{
			sm.enter(new CarSelectionState(CAR_SELECTION_NETWORK_JOIN, net, status.joinedPlayers, status.reservedCars), 0);
			return;
		}
	}
	
	if (lastSend + JOIN_INTERVALL < SDL_GetTicks())
	{
		if (attempts >= JOIN_ATTEMPTS)
		{
			sm.enter(new ErrorState("Join error", "The game did not respond!", new NetworkMenu(net)), 0);
			return;
		}
		switch (joinState)
		{
		case JOIN_NOT_STARTED:
			net->sendJoinStart();
			break;
		case JOIN_MORE_CARS:
			net->sendJoinCar(selectCar);
			break;
		case JOIN_NO_MORE_CARS:
			net->sendJoinEnd();
			break;
		}
		lastSend = SDL_GetTicks();
		++attempts;
	}
}
