/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "LAN.h"
#include "Settings.h"
#include <ctime>

LAN::LAN()
:	port(0),
	nextBroadcast(0),
	nextBroadcastInterval(MIN_BROADCAST_INTERVAL),
	timestamp(std::time(0))
{
	if (!error.empty())
	{
		return;
	}
	
	std::vector<Uint16>::iterator it;
	for (it = Settings::network.lanPorts.begin(); it != Settings::network.lanPorts.end() && !socket; ++it)
	{
		port = *it;
		socket = SDLNet_UDP_Open(port);
	}
	
	if (!socket)
	{
		error = "No available ports!";
	}
	
	// Use something random until we know our real IP address
	localAddress.host = SDL_Swap32(rand()) ^ SDL_GetTicks() ^ SDL_ThreadID();
	
	SDLNet_Write16(port, &localAddress.port);
}

void LAN::step()
{
	if (connectionType == NET_LOBBY)
	{
		Uint32 now = SDL_GetTicks();
		if (now >= nextBroadcast)
		{
			broadcast();
			now = SDL_GetTicks();
			nextBroadcast = now + nextBroadcastInterval;
			nextBroadcastInterval = std::min(2 * nextBroadcastInterval, MAX_BROADCAST_INTERVAL);
		}
	}
	
	Network::step();
}

void LAN::handleMessage(Uint8 msgType, IPaddress address)
{
	switch (msgType)
	{
		case MSG_LAN_BROADCAST:
		{
			if (connectionType == NET_GAME_PLAYER)
			{
				break;
			}
			IPaddress unverifiedAddress = msg.readAddress();
			if (localAddress == address || localAddress == unverifiedAddress)
			{
				break;
			}
			
			if (connections.find(address) == connections.end())
			{
				// New connection!
				if (connections.size() >= Settings::network.maxConnections)
				{
					break;
				}
			}
			
			Uint32 newTimestamp = msg.read32();
			
			Connection& connection = connections[address];
			Uint32& timestamp = timestamps[address];
			if (connection.lastRecive != 0 && newTimestamp > timestamp)
			{
				// The connection is already established but is outdated.
				
				// Reset some of the necessary fields.
				connection.lastChatMessageRecived = 0;
				connection.lastChatMessageDelivered = 0;
				connection.lastTimestamp = 0;
				
				std::map<IPaddress, JoinInProgress>::iterator joinIt = joinInProgress.find(address);
				if (joinIt != joinInProgress.end())
				{
					joinIt->second.lastTimestamp = 0;
				}
				
				// We don't touch gameInfo because it
				// might need special clean-up code.
			}
			
			connection.lastRecive = SDL_GetTicks();
			timestamp = newTimestamp;
			
			msg.reset();
			msg.write8(MSG_LAN_BROADCAST_ACK);
			msg.writeAddress(address);
			send(address);
			
			break;
		}
		case MSG_LAN_BROADCAST_ACK:
		{
			if (connectionType == NET_GAME_PLAYER)
			{
				break;
			}
			localAddress = msg.readAddress();
			
			if (connections.find(address) == connections.end())
			{
				// New connection!
				if (connections.size() >= Settings::network.maxConnections)
				{
					break;
				}
			}
			
			connections[address].lastRecive = SDL_GetTicks();
			
			break;
		}
		default:
		{
			Network::handleMessage(msgType, address);
		}
	}
}

std::string LAN::name() const
{
	return "LAN";
}

void LAN::broadcast()
{
	msg.reset();
	msg.write8(MSG_LAN_BROADCAST);
	msg.writeAddress(localAddress);
	msg.write32(timestamp);
	
	IPaddress addr;
	addr.host = INADDR_BROADCAST;
	
	std::vector<Uint16>::iterator it;
	for (it = Settings::network.lanPorts.begin(); it != Settings::network.lanPorts.end(); ++it)
	{
		SDLNet_Write16(*it, &addr.port);
		send(addr);
	}
}

void LAN::onConnectionTimeout(const IPaddress& addr)
{
	timestamps.erase(addr);
}

void LAN::refreshAsNeeded()
{
	if (connectionType == NET_GAME_PLAYER)
	{
		// This happens when we change from GAME_PLAYER to LOBBY.
		
		// Reset broadcast timings to quickly set up connections again.
		nextBroadcast = 0;
		nextBroadcastInterval = MIN_BROADCAST_INTERVAL;
	}
}
