/*
 * Copyright 2010, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Map.h"
#include "TileSet.h"
#include "Settings.h"
#include "NetworkMessage.h"
#include "BinaryFile.h"
#include <algorithm>
#include <numeric>
#include <cassert>

Map Map::generate()
{
	// Try to generate a map that satisfies the tracks settings.
	const int MIN_ATTEMPTS = 100;
	const int MAX_DURATION = 200;
	
	Map bestMap;
	{
		Map map;
		double bestScore = 1000;
		
		int attempts = 0;
		int duration = 0;
		Uint32 start = SDL_GetTicks();
		do
		{
			map.makeTrack();
			double score = map.getScore();
			if (score < bestScore)
			{
				bestScore = score;
				bestMap = map;
			}
			++attempts;
			duration = SDL_GetTicks() - start;
		} while (bestScore > 0 && (duration < MAX_DURATION || attempts < MIN_ATTEMPTS));
	}
	
	// Put track in the center of the map.
	bestMap.center();
	
	// Generate tileset decoration.
	const TileSet& tileSet = TileSet::getRandomTileSet();
	bestMap.tileset = tileSet.getName();
	int gfxLimit = tileSet.getGfxTileLimit();
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			bestMap.decoration[y][x] = std::rand() % gfxLimit;
		}
	}
	
	// Find a suitable finish line.
	std::vector<std::pair<int, int> > possibleFinishLines;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (bestMap[y][x] == ROAD_HORIZONTAL || bestMap[y][x] == ROAD_VERTICAL)
			{
				possibleFinishLines.push_back(std::make_pair(x, y));
			}
		}
	}
	int gi = std::rand() % possibleFinishLines.size();
	bestMap.finishX = possibleFinishLines[gi].first;
	bestMap.finishY = possibleFinishLines[gi].second;
	
	// Pick a driving direction.
	if (bestMap[bestMap.finishY][bestMap.finishX] == ROAD_HORIZONTAL)
	{
		bestMap.finishDir = (std::rand() % 2) ? LEFT : RIGHT;
	}
	else
	{
		bestMap.finishDir = (std::rand() % 2) ? UP : DOWN;
	}
	
	return bestMap;
}

void Map::zeroFill()
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			map[y][x] = 0;
		}
	}
}

int Map::getRoadCount() const
{
	int count = 0;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0)
			{
				count++;
			}
		}
	}
	return count;
}

static double calcError(int value, int minSetting, int maxSetting, int minConstant, int maxConstant)
{
	double error = 0.0;
	if (value < minSetting)
	{
		error = (minSetting - value) / double(minSetting - minConstant);
	}
	else if (value > maxSetting)
	{
		error = (value - maxSetting) / double(maxConstant - maxSetting);
	}
	return error * error;
}

double Map::getScore() const
{
	// Find a non-empty tile to start traversing on.
	int x;
	int y;
	for (y = 0; y < MAX_MAP_Y; y++)
	{
		for (x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0)
			{
				goto end_loop;
			}
		}
	}
end_loop:
	int length = 0;
	int corners = 0;
	int straighaway = 0;
	int cornerSeries = 0;
	int longestStraighaway = 0;
	int longestCornerSeries = 0;
	int incompleteCornerSeries = 0;
	int minX = x;
	int maxX = x;
	int minY = y;
	int maxY = y;
	
	// Traverse tiles.
	int endX = x;
	int endY = y;
	int prevDir = DOWN;
	do
	{
		int tile = map[y][x];
		++length;
		if (tile == ROAD_HORIZONTAL || tile == ROAD_VERTICAL)
		{
			++straighaway;
			if (incompleteCornerSeries == 0)
			{
				// The traversal always start on a corner so there is no
				// chance of setting a value of zero here.
				incompleteCornerSeries = cornerSeries;
			}
			cornerSeries = 0;
			longestStraighaway = std::max(longestStraighaway, straighaway);
		}
		else
		{
			++corners;
			straighaway = 0;
			++cornerSeries;
			longestCornerSeries = std::max(longestCornerSeries, cornerSeries);
		}
		
		minX = std::min(minX, x);
		maxX = std::max(maxX, x);
		maxY = std::max(maxY, y);
		
		int dir = tile & ~prevDir;
		switch (dir)
		{
		case LEFT:
			prevDir = RIGHT;
			--x;
			break;
		case RIGHT:
			prevDir = LEFT;
			++x;
			break;
		case UP:
			prevDir = DOWN;
			--y;
			break;
		case DOWN:
			prevDir = UP;
			++y;
			break;
		}
	} while (x != endX || y != endY);
	
	longestCornerSeries = std::max(longestCornerSeries, cornerSeries + incompleteCornerSeries);
	
	int width = maxX - minX + 1;
	int height = maxY - minY + 1;
	int cornerPercentage = 100 * corners / length;
	
	double error[6] = {};
	
	error[0] = calcError(length, 
	                     Settings::tracks.minLength, 
	                     Settings::tracks.maxLength, 
	                     MIN_TRACK_LENGTH, 
	                     MAX_TRACK_LENGTH);
	error[1] = calcError(width, 
	                     Settings::tracks.minWidth, 
	                     Settings::tracks.maxWidth, 
	                     MIN_TRACK_WIDTH, 
	                     MAX_TRACK_WIDTH);
	error[2] = calcError(height, 
	                     Settings::tracks.minHeight, 
	                     Settings::tracks.maxHeight, 
	                     MIN_TRACK_HEIGHT, 
	                     MAX_TRACK_HEIGHT);
	error[3] = calcError(cornerPercentage, 
	                     Settings::tracks.minCorners, 
	                     Settings::tracks.maxCorners, 
	                     MIN_TRACK_CORNERS, 
	                     MAX_TRACK_CORNERS);
	error[4] = calcError(longestStraighaway, 
	                     Settings::tracks.minLongestStraighaway, 
	                     Settings::tracks.maxLongestStraighaway, 
	                     MIN_LONGEST_STRAIGHTAWAY, 
	                     MAX_LONGEST_STRAIGHTAWAY);
	error[5] = calcError(longestCornerSeries, 
	                     Settings::tracks.minLongestCornerSeries, 
	                     Settings::tracks.maxLongestCornerSeries, 
	                     MIN_LONGEST_CORNER_SERIES, 
	                     MAX_LONGEST_CORNER_SERIES);

	return std::accumulate(error, error + 6, 0.0);
}

const Uint8* Map::operator[](int y) const
{
	return map[y];
}

void Map::makeTrack()
{
	 // make sure the map is empty before we start
	zeroFill();
	
	bool verticalSplit = std::rand() & 1;
	
	if (verticalSplit)
	{
		int splitX = std::rand() % (MAX_MAP_X - 1) + 1;
		
		int connectionY1 = std::rand() % MAX_MAP_Y;
		int connectionY2 = std::rand() % (MAX_MAP_Y - 3);
		if (connectionY2 >= connectionY1 - 1)
		{
			connectionY2 += 3;
		}
		
		makeConnection(0, splitX - 1, 0, MAX_MAP_Y - 1, splitX - 1, connectionY1, RIGHT, splitX - 1, connectionY2, RIGHT);
		makeConnection(splitX, MAX_MAP_X - 1, 0, MAX_MAP_Y - 1, splitX, connectionY1, LEFT, splitX, connectionY2, LEFT);
	}
	else
	{
		int splitY = std::rand() % (MAX_MAP_Y - 1) + 1;
		
		int connectionX1 = std::rand() % MAX_MAP_X;
		int connectionX2 = std::rand() % (MAX_MAP_X - 3);
		if (connectionX2 >= connectionX1 - 1)
		{
			connectionX2 += 3;
		}
		makeConnection(0, MAX_MAP_X - 1, 0, splitY - 1, connectionX1, splitY - 1, DOWN, connectionX2, splitY - 1, DOWN);
		makeConnection(0, MAX_MAP_X - 1, splitY, MAX_MAP_Y - 1, connectionX1, splitY, UP, connectionX2, splitY, UP);
	}
}

void Map::makeConnection(int lowerX, int upperX, int lowerY, int upperY, int connectionX1, int connectionY1, int connectionDir1, int connectionX2, int connectionY2, int connectionDir2)
{
	if (connectionX1 == connectionX2 && connectionY1 == connectionY2)
	{
		// ORing two directions gives the connection
		map[connectionY1][connectionX1] = connectionDir1 | connectionDir2;
	}
	else
	{
		bool verticalSplit;
		if (connectionX1 == connectionX2)
		{
			verticalSplit = false;
		}
		else if (connectionY1 == connectionY2)
		{
			verticalSplit = true;
		}
		else 
		{
			verticalSplit = std::rand() & 1;
		}
		
		if (verticalSplit)
		{
			// make sure we know connection 1 is to the left
			// and connection 2 to the right
			if (connectionX1 > connectionX2)
			{
				std::swap(connectionX1, connectionX2);
				std::swap(connectionY1, connectionY2);
				std::swap(connectionDir1, connectionDir2);
			}

			// calculate the spliting plane
			int splitX = std::rand() % (connectionX2 - connectionX1) + connectionX1 + 1;

			int splitConnectionY = std::rand() % (upperY - lowerY + 1) + lowerY;
			
			// left side
			makeConnection(lowerX, splitX - 1,            // lo/up x
			  lowerY, upperY,                             // lo/up y
			  connectionX1, connectionY1, connectionDir1, // con. 1
			  splitX - 1, splitConnectionY, RIGHT);       // con. 2

			// right side
			makeConnection(splitX, upperX,                // lo/up x
			  lowerY, upperY,                             // lo/up y
			  connectionX2, connectionY2, connectionDir2, // con. 1
			  splitX, splitConnectionY, LEFT);            // con. 2
		}
		else
		{
			// make sure we know connection 1 is over connection 2
			if (connectionY1 > connectionY2)
			{
				std::swap(connectionX1, connectionX2);
				std::swap(connectionY1, connectionY2);
				std::swap(connectionDir1, connectionDir2);
			}

			// calculate the spliting plane
			int splitY = std::rand() % (connectionY2 - connectionY1) + connectionY1 + 1;
			int splitConnectionX = std::rand() % (upperX - lowerX + 1) + lowerX;

			// up side
			makeConnection(lowerX, upperX,                // lo/up x
			  lowerY, splitY - 1,                         // lo/up y
			  connectionX1, connectionY1, connectionDir1, // con. 1
			  splitConnectionX, splitY - 1, DOWN);        // con. 2

			// down side
			makeConnection(lowerX, upperX,                // lo/up x
			  splitY, upperY,                             // lo/up y
			  connectionX2, connectionY2, connectionDir2, // con. 1
			  splitConnectionX, splitY, UP);              // con. 2
		}
	}
}

void Map::center()
{
	int minX = MAX_MAP_Y;
	int minY = MAX_MAP_Y;
	int maxX = 0;
	int maxY = 0;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0)
			{
				minX = std::min(minX, x);
				minY = std::min(minY, y);
				maxX = std::max(maxX, x);
				maxY = std::max(maxY, y);
			}
		}
	}
	
	// The code below will work even if the map is empty
	
	int spaceLeft  = minX;
	int spaceRight = MAX_MAP_X - maxX - 1;
	int moveSpaceLeft = (spaceRight - spaceLeft) / 2;
	
	int spaceUp   = minY;
	int spaceDown = MAX_MAP_Y - maxY - 1;
	int moveSpaceUp = (spaceDown - spaceUp) / 2;
	
	if (moveSpaceLeft != 0 || moveSpaceUp != 0)
	{
		// move!
		Map temp = *this;
		zeroFill();
		for (int y = minY; y <= maxY; y++)
		{
			for (int x = minX; x <= maxX; x++)
			{
				map[(y + moveSpaceUp) % MAX_MAP_Y][(x + moveSpaceLeft) % MAX_MAP_X] = temp.map[y][x];
			}
		}
	}
}

int Map::getFinishX() const
{
	return finishX;
}

int Map::getFinishY() const
{
	return finishY;
}

int Map::getFinishDir() const
{
	return finishDir;
}

int Map::getDecoration(int x, int y) const
{
	return decoration[y][x];
}

std::string Map::getTilesetName() const
{
	return tileset;
}

void Map::write(NetworkMessage& msg) const
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			msg.write8(map[y][x]);
			msg.write8(decoration[y][x]);
		}
	}
	msg.writeString(tileset);
	msg.write8(finishX);
	msg.write8(finishY);
	msg.write8(finishDir);
}

void Map::read(NetworkMessage& msg)
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			map[y][x] = msg.read8();
			decoration[y][x] = msg.read8();
		}
	}
	tileset = msg.readString();
	finishX = msg.read8();
	finishY = msg.read8();
	finishDir = msg.read8();
}

void Map::write(OutputBinaryFile& out) const
{
	// Write decoration.
	Uint8 dMax = 0;
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			Uint8 d = decoration[y][x];
			if (d > dMax)
			{
				dMax = d;
			}
		}
	}
	int dBits = calculateBitsNeeded(dMax);
	out.write(8, dBits);
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			out.write(dBits, decoration[y][x]);
			out.write(4, map[y][x]);
		}
	}
	
	out.write(4, finishX);
	out.write(3, finishY);
	out.write(4, finishDir);
	out.writeS(tileset);
}

void Map::read(InputBinaryFile& in)
{
	// Read decoration.
	int dBits = in.read(8);
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			decoration[y][x] = in.read(dBits);
			map[y][x] = in.read(4);

		}
	}
	
	finishX = in.read(4);
	finishY = in.read(3);
	finishDir = in.read(4);
	tileset = in.readS();	
}

bool Map::isValid() const
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			if (map[y][x] != 0               &&
			    map[y][x] != ROAD_HORIZONTAL &&
			    map[y][x] != ROAD_VERTICAL   &&
			    map[y][x] != ROAD_LEFT_UP    &&
			    map[y][x] != ROAD_LEFT_DOWN  &&
			    map[y][x] != ROAD_RIGHT_UP   &&
			    map[y][x] != ROAD_RIGHT_DOWN)
			{
				return false;
			}
			if ((map[y][x] & LEFT) && (x == 0 || (map[y][x - 1] & RIGHT) == 0))
			{
				return false;
			}
			if ((map[y][x] & RIGHT) && (x == MAX_MAP_X - 1 || (map[y][x + 1] & LEFT) == 0))
			{
				return false;
			}
			if ((map[y][x] & UP) && (y == 0 || (map[y - 1][x] & DOWN) == 0))
			{
				return false;
			}
			if ((map[y][x] & DOWN) && (y == MAX_MAP_Y - 1 || (map[y + 1][x] & UP) == 0))
			{
				return false;
			}
		}
	}
	if (finishX < 0 || finishX >= MAX_MAP_X ||
	    finishY < 0 || finishY >= MAX_MAP_Y)
	{
		return false;
	}
	if (finishDir != LEFT  && 
	    finishDir != RIGHT && 
	    finishDir != UP    && 
	    finishDir != DOWN)
	{
		return false;
	}
	if ((finishDir == LEFT || finishDir == RIGHT) && map[finishY][finishX] != ROAD_HORIZONTAL)
	{
		return false;
	}
	if ((finishDir == UP || finishDir == DOWN) && map[finishY][finishX] != ROAD_VERTICAL)
	{
		return false;
	}
	return true;
}
