/*
 * Copyright 2011, 2013, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Menu.h"
#include "GameState.h"
#include "Color.h"
#include "StateManager.h"
#include "Font.h"
#include "str.h"
#include <cstdarg>
#include <cassert>
#include <algorithm>
#include <cmath>

Menu::Menu()
:	selectedItem(0),
	hoverItem(-1),
	selectedItemState(ITEM_SELECTED),
	titleRotation(0),
	titleOldRotation(0),
	titleRotationSpeed(0),
	titleRotationSpeedLimit(0),
	titleDrag(false),
	state(0)
{
}

void Menu::updateItem(std::size_t i)
{
	if (i >= items.size())
	{
		return;
	}
	
	Item& item = items[i];
	std::string text = item.label;
	if (i == selectedItem && selectedItemState == ITEM_ACTIVE)
	{
		text += inputString;
	}
	else if (item.strValue)
	{
		text += *item.strValue;
	}
	else if (item.intValue)
	{
		std::size_t iStrVal = *item.intValue - item.min;
		if (iStrVal < item.strValues.size())
		{
			text += item.strValues[iStrVal];
		}
		else
		{
			text += int2string(*item.intValue);
		}
	}
	if (i != selectedItem || selectedItemState != ITEM_ACTIVE)
	{
		text += item.suffix;
	}
	Color textColor = MENU_INACTIVE_COLOR;
	if (i == selectedItem)
	{
		if (selectedItemState == ITEM_SELECTED)
		{
			textColor = MENU_ACTIVE_COLOR;
		}
		else if (selectedItemState == ITEM_ACTIVE)
		{
			textColor = MENU_ACTIVE_COLOR2;
		}
	}
	Surface s(text, BIG_FONT, textColor, TEXT_BORDER_COLOR);
	item.renderObject.setSurface(s);
}

bool Menu::onEvent(MenuEvent event)
{
	switch (event.type)
	{
	case MENU_ACTION:
		action();
		return true;
	case MENU_NEXT:
		next();
		return true;
	case MENU_PREV:
		prev();
		return true;
	case MENU_INC:
		inc();
		return true;
	case MENU_DEC:
		dec();
		return true;
	case MENU_ESC:
		return escape();
	case MENU_FIRST:
		first();
		return true;
	case MENU_LAST:
		last();
		return true;
	case MENU_MIN:
		min();
		return true;
	case MENU_MAX:
		max();
		return true;
	case MENU_MOUSE_MOVE:
		mouseMove(event.x, event.y);
		return true;
	case MENU_MOUSE_CLICK1:
		click(1, event.x, event.y);
		return true;
	case MENU_MOUSE_CLICK2:
		click(2, event.x, event.y);
		return true;
	case MENU_MOUSE_RELEASE:
		release();
		return true;
	case MENU_TEXT:
		input(event.c);
		return true;
	case MENU_ERASE:
		erase();
		return true;
	case MENU_DELETE:
		return del();
	}
	return false;
}

bool Menu::onEvent(const SDL_Event& event)
{
	return onEvent(StateManager::getInstance().getMenuEvent(event));
}

int Menu::getItem(int x, int y)
{
	for (std::size_t i = 0; i < items.size(); ++i)
	{
		if (items[i].renderObject.isHit(x, y))
		{
			return i;
		}
	}
	return -1;
}

void Menu::addItem(const std::string& label, bool* fireAction)
{	
	Item item;
	item.label = label;
	item.fireAction = fireAction;
	item.intValue = 0;
	item.strValue = 0;
	item.min = 0;
	item.max = 0;
	item.wrap = false;
	items.push_back(item);
	
	if (fireAction)
	{
		*fireAction = false;
	}
}

void Menu::addItem(const std::string& label, int* value, int min, int max, const char* firstStrValue, ...)
{
	assert(value);
	assert(*value >= min);
	assert(*value <= max);
	
	Item item;
	item.label = label + ": ";
	item.fireAction = 0;
	item.intValue = value;
	item.strValue = 0;
	item.min = min;
	item.max = max;
	
	if (firstStrValue)
	{
		item.wrap = true;
		std::va_list vl;
		va_start(vl,firstStrValue);
		item.strValues.push_back(firstStrValue);
		for (int i = min + 1; i <= max; i++)
		{
			firstStrValue = va_arg(vl, const char*);
			item.strValues.push_back(firstStrValue);
		}
		va_end(vl);
	}
	else
	{
		item.wrap = false;
	}
	items.push_back(item);
}

void Menu::addItem(const std::string& label, int* value, const std::vector<std::string>& strValues)
{
	assert(value);
	assert(*value >= 0);
	assert(unsigned(*value) < strValues.size());
	
	Item item;
	item.label = label + ": ";
	item.fireAction = 0;
	item.intValue = value;
	item.strValue = 0;
	item.min = 0;
	item.max = strValues.size() - 1;
	item.strValues = strValues;
	item.wrap = true;
	items.push_back(item);
}

void Menu::addItem(const std::string& label, std::string* value, int max)
{
	assert(value);
	
	Item item;
	item.label = label + ": ";
	item.fireAction = 0;
	item.intValue = 0;
	item.strValue = value;
	item.min = 0;
	item.max = max;
	item.wrap = false;
	items.push_back(item);
}

void Menu::addItem(const std::string& label, std::string* value, int max, const std::vector<std::string>& strValues)
{
	addItem(label, value, max);
	items.back().strValues = strValues;
}

void Menu::appendSuffix(const std::string& suffix)
{
	if (!items.empty())
	{
		items.back().suffix = suffix;
	}
}

void Menu::setSuffix(int* value, const std::string& suffix)
{
	assert(value);
	for (std::size_t i = 0; i < items.size(); i++)
	{
		if (value == items[i].intValue)
		{
			items[i].suffix = suffix;
			updateItem(i);
			break;
		}
	}
}

void Menu::action()
{
	if (!isVisible() || items.empty())
	{
		return;
	}

	Item& item = items[selectedItem];
	
	if (selectedItemState != ITEM_ACTIVE && !item.fireAction)
	{
		for (std::size_t i = selectedItem; i < items.size(); ++i)
		{
			if (items[i].fireAction)
			{
				selectedItemState = ITEM_SELECTED;
				std::size_t prevItem = selectedItem;
				selectedItem = i;
				updateItem(prevItem);
				updateItem(selectedItem);
				break;
			}
		}
	}
	else if (selectedItemState == ITEM_SELECTED)
	{
		if (item.fireAction)
		{
			*item.fireAction = true;
		}
	}
	else if (selectedItemState == ITEM_ACTIVE)
	{
		applyInput();
		selectedItemState = ITEM_SELECTED;
		updateItem(selectedItem);
	}
}

void Menu::next()
{	
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	if (selectedItemState != ITEM_UNSELECTED)
	{
		applyInput();
		std::size_t prevItem = selectedItem++;
		if (selectedItem == items.size())
		{
			selectedItem = 0;
		}
		updateItem(prevItem);
	}
	selectedItemState = ITEM_SELECTED;
	updateItem(selectedItem);
}
void Menu::prev()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	if (selectedItemState != ITEM_UNSELECTED)
	{
		applyInput();
		std::size_t prevItem = selectedItem--;
		if (selectedItem >= items.size())
		{
			selectedItem = items.size() - 1;
		}
		updateItem(prevItem);
	}
	selectedItemState = ITEM_SELECTED;
	updateItem(selectedItem);
}

void Menu::inc()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	Item& item = items[selectedItem];
	if (selectedItemState == ITEM_SELECTED)
	{
		if (item.intValue)
		{
			if (++(*item.intValue) > item.max)
			{
				*item.intValue = (item.wrap ? item.min : item.max);
			}
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.intValue);
			}
		}
		else if (item.strValue && !item.strValues.empty())
		{
			std::vector<std::string>::iterator it = std::find(item.strValues.begin(), item.strValues.end(), *item.strValue);
			if (it == item.strValues.end())
			{
				it = item.strValues.begin();
			}
			else
			{
				++it;
				if (it == item.strValues.end())
				{
					it = item.strValues.begin();
				}
			}
			*item.strValue = *it;
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.strValue);
			}
		}
	}
}

void Menu::dec()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	Item& item = items[selectedItem];
	if (selectedItemState == ITEM_SELECTED)
	{
		if (item.intValue)
		{
			if (--(*item.intValue) < item.min)
			{
				*item.intValue = (item.wrap ? item.max : item.min);
			}
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.intValue);
			}
		}
		else if (item.strValue && !item.strValues.empty())
		{
			std::vector<std::string>::iterator it = std::find(item.strValues.begin(), item.strValues.end(), *item.strValue);
			if (it == item.strValues.begin())
			{
				it = item.strValues.end() - 1;
			}
			else
			{
				--it;
			}
			*item.strValue = *it;
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.strValue);
			}
		}
	}
}

bool Menu::escape()
{
	if (!isVisible() || items.empty())
	{
		return false;
	}
	
	if (selectedItemState == ITEM_ACTIVE)
	{
		selectedItemState = ITEM_SELECTED;
		updateItem(selectedItem);
		return true;
	}
	return false;
}

bool Menu::del()
{
	if (!isVisible() || items.empty())
	{
		return false;
	}
	
	if (selectedItemState == ITEM_UNSELECTED)
	{
		return false;
	}
	Item& item = items[selectedItem];
	if (item.strValue || (item.intValue && item.strValues.empty()))
	{
		inputString.clear();
		selectedItemState = ITEM_ACTIVE;
		updateItem(selectedItem);
		return true;
	}
	return false;
}

void Menu::first()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	applyInput();
	selectedItemState = ITEM_SELECTED;
	std::size_t prevItem = selectedItem;
	selectedItem = 0;
	if (selectedItem != prevItem)
	{
		updateItem(prevItem);
	}
	updateItem(selectedItem);
}

void Menu::last()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	applyInput();
	selectedItemState = ITEM_SELECTED;
	std::size_t prevItem = selectedItem;
	selectedItem = items.size() - 1;
	if (selectedItem != prevItem)
	{
		updateItem(prevItem);
	}
	updateItem(selectedItem);
}

void Menu::min()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	Item& item = items[selectedItem];
	if (selectedItemState == ITEM_SELECTED)
	{
		if (item.intValue && *item.intValue != item.min)
		{
			*item.intValue = item.min;
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.intValue);
			}
		}
		else if (item.strValue && !item.strValues.empty() && *item.strValue != item.strValues.front())
		{
			*item.strValue = item.strValues.front();
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.strValue);
			}
		}
	}
}

void Menu::max()
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	Item& item = items[selectedItem];
	if (selectedItemState == ITEM_SELECTED)
	{
		if (item.intValue && *item.intValue != item.max)
		{
			*item.intValue = item.max;
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.intValue);
			}
		}
		else if (item.strValue && !item.strValues.empty() && *item.strValue != item.strValues.back())
		{
			*item.strValue = item.strValues.back();
			updateItem(selectedItem);
			if (state)
			{
				state->onValueUpdate(item.strValue);
			}
		}
	}
}

void Menu::mouseMove(int x, int y)
{
	if (titleDrag)
	{
		Vector p1 = titleDragPos - titlePos;
		Vector p2 = Vector(x, y) - titlePos;
		titleRotation = Vector::angle(p1, p2);
	}
	
	if (!isVisible())
	{
		return;
	}
	
	if (selectedItemState != ITEM_ACTIVE)
	{
		int newHoverItem = getItem(x, y);
		if (newHoverItem != -1)
		{
			if (newHoverItem != static_cast<int>(selectedItem) || selectedItemState == ITEM_UNSELECTED)
			{
				applyInput();
				std::size_t prevItem = selectedItem;
				selectedItem = newHoverItem;
				selectedItemState = ITEM_SELECTED;
				if (selectedItem != prevItem)
				{
					updateItem(prevItem);
				}
				updateItem(selectedItem);
			}
		}
		else if (hoverItem == static_cast<int>(selectedItem) && selectedItemState != ITEM_ACTIVE)
		{
			selectedItemState = ITEM_UNSELECTED;
			updateItem(selectedItem);
		}
		hoverItem = newHoverItem;
	}
}

void Menu::click(int button, int x, int y)
{
	if (titleRenderObject.isHit(x, y) && button == 1)
	{
		titleDragPos = Vector(x, y);
		titleDragPos.rotate(titlePos, -titleRotation);
		titleDrag = true;
		return;
	}
	
	if (!isVisible())
	{
		return;
	}
	
	int clickedItem = getItem(x, y);
	if (clickedItem == -1)
	{
		if (selectedItemState != ITEM_UNSELECTED)
		{
			applyInput();
			selectedItemState = ITEM_UNSELECTED;
			updateItem(selectedItem);
		}
	}
	else
	{
		if (clickedItem != static_cast<int>(selectedItem))
		{
			if (button == 1)
			{
				applyInput();
			}
			std::size_t prevItem = selectedItem;
			selectedItem = clickedItem;
			selectedItemState = ITEM_SELECTED;
			updateItem(prevItem);
			updateItem(selectedItem);
		}
		else 
		{
			Item& item = items[selectedItem];
			if (item.fireAction)
			{
				if (button == 1)
				{
					action();
				}
				else
				{
					
				}
			}
			else
			{
				if (button == 1)
				{
					inc();
				}
				else
				{
					dec();
				}
			}
		}
	}
}

void Menu::release()
{
	titleDrag = false;
}

void Menu::input(char c)
{
	if (!isVisible() || items.empty())
	{
		return;
	}
	
	if (selectedItemState == ITEM_UNSELECTED ||
	   (c == ' ' && (inputString.empty() || 
	                 inputString[inputString.size() - 1] == ' ')))
	{
		return;
	}
	Item& item = items[selectedItem];
	if (item.strValue)
	{
		bool startTypeNow = false;
		if (selectedItemState == ITEM_SELECTED)
		{
			startTypeNow = true;
			inputString.clear();
			selectedItemState = ITEM_ACTIVE;
		}
		if (inputString.size() < static_cast<std::size_t>(item.max))
		{
			inputString += c;
			if (state)
			{
				state->validateTypedValue(item.strValue, inputString);
			}
			if (startTypeNow && inputString.empty())
			{
				selectedItemState = ITEM_SELECTED;
			}
		}
		updateItem(selectedItem);
	}
	else if (item.intValue)
	{
		if (!item.strValues.empty())
		{
			for (std::size_t i = 1; i < item.strValues.size(); ++i)
			{
				std::size_t j = (i + *item.intValue - item.min) % item.strValues.size();
				if (c == std::tolower(item.strValues[j][0]))
				{
					*item.intValue = item.min + j;
					updateItem(selectedItem);
					break;
				}
			}
		}
		else if (std::isdigit(c))
		{
			if (selectedItemState == ITEM_SELECTED)
			{
				inputString.clear();
				selectedItemState = ITEM_ACTIVE;
			}
			if (inputString.size() < int2string(item.max).length())
			{
				inputString += c;
			}
			updateItem(selectedItem);
		}
	}
}

void Menu::erase()
{
	if (!isVisible())
	{
		return;
	}
	
	if (selectedItemState == ITEM_ACTIVE && !inputString.empty())
	{
		inputString.resize(inputString.size() - 1);
		updateItem(selectedItem);
	}
}

void Menu::addRenderObjects(GameState* gs, int x, int y, int dx, int dy)
{
	const Font& font = getBigFont();
	int leftWidthNeeded = 0;
	int rightWidthNeeded = 0;
	std::vector<int> offsets;
	std::vector<Item>::iterator it;
	for (it = items.begin(); it != items.end(); ++it)
	{
		Item& item = *it;
		int leftWidth = font.calculateTextWidth(it->label);
		int rightWidth = 0;
		if (item.strValue)
		{
			rightWidth = font.calculateTextWidth(*item.strValue) + 1;
		}
		else if (item.intValue)
		{
			if (item.strValues.empty())
			{
				rightWidth = int2string(item.max).length() * 6;
			}
			else
			{
				for (std::vector<std::string>::iterator sit = item.strValues.begin(); sit != item.strValues.end(); ++sit)
				{
					rightWidth = std::max(rightWidth, font.calculateTextWidth(*sit) + 1);
				}
			}
		}
		else if (item.label.find(':') != std::string::npos)
		{
			// Makes so that labels with colon also get aligned correctly.
			rightWidth = 1;
		}
		
		if (rightWidth == 0)
		{
			rightWidth = leftWidth = leftWidth / 2;
		}
		else
		{
			// Move the space after the colon to the right side.
			leftWidth -= 4;
			rightWidth += 4;
		}
		
		leftWidthNeeded  = std::max(leftWidthNeeded, leftWidth);
		rightWidthNeeded = std::max(rightWidthNeeded, rightWidth);
		
		offsets.push_back(leftWidth);
	}
	
	if (rightWidthNeeded > GAME_WIDTH - x)
	{
		x = GAME_WIDTH - rightWidthNeeded;
	}
	if (leftWidthNeeded > x)
	{
		x = leftWidthNeeded;
	}
	
	y -= items.size() * dy / 2;
	
	for (it = items.begin(); it != items.end(); ++it)
	{
		std::size_t i = it - items.begin();
		updateItem(i);
		it->renderObject.moveTo(x - offsets[i], y);
		gs->addRenderObject(it->renderObject);
		x += dx;
		y += dy;
	}
	
	gs->addRenderObject(titleRenderObject);
	
	state = gs;
}

void Menu::notifyValueUpdate(int* value)
{
	assert(value);
	for (std::size_t i = 0; i < items.size(); i++)
	{
		if (value == items[i].intValue)
		{
			assert(*value >= items[i].min);
			assert(*value <= items[i].max);
			updateItem(i);
			break;
		}
	}
}

void Menu::notifyValueUpdate(std::string* value)
{
	assert(value);
	for (std::size_t i = 0; i < items.size(); i++)
	{
		if (value == items[i].strValue)
		{
			updateItem(i);
			break;
		}
	}
}

void Menu::makeActive(ItemState state)
{
	if (selectedItemState != state)
	{
		selectedItemState = state;
		updateItem(selectedItem);
	}
}

void Menu::setLabel(int i, const std::string& label)
{
	items[i].label = label;
	updateItem(i);
}

bool Menu::activeItemValue(int* value) const
{
	if (items.empty())
	{
		return false;
	}
	
	return items[selectedItem].intValue == value;
}

void Menu::applyInput()
{
	if (items.empty())
	{
		return;
	}
	Item& item = items[selectedItem];
	if (selectedItemState == ITEM_ACTIVE)
	{
		if (item.strValue)
		{
			if (!inputString.empty() && inputString[inputString.size() - 1] == ' ')
			{
				inputString.resize(inputString.size() - 1);
			}
			
			*item.strValue = inputString;
			if (state)
			{
				state->onValueUpdate(item.strValue);
			}
		}
		else if (item.intValue)
		{
			if (!inputString.empty())
			{
				*item.intValue = std::atoi(inputString.c_str());
			}
			if (*item.intValue > item.max)
			{
				*item.intValue = item.max;
			}
			else if (*item.intValue < item.min)
			{
				*item.intValue = item.min;
			}
			if (state)
			{
				state->onValueUpdate(item.intValue);
			}
		}
		inputString.clear();
	}
}

void Menu::reset()
{
	if (state)
	{
		for (std::size_t i = 0; i < items.size(); ++i)
		{
			state->removeRenderObject(items[i].renderObject);
		}
		state->removeRenderObject(titleRenderObject);
	}
	items.clear();
	selectedItem = 0;
	hoverItem = 0;
	selectedItemState = ITEM_SELECTED;
	inputString.clear();
}

void Menu::setTitle(const std::string& newTitle)
{
	title = newTitle;
	Surface text(title, BIG_FONT, TITLE_COLOR, true);
	Surface box(text.width() + 7, text.height() + 8); // (right-most column is space between characters)
	
	box.fill(BOX_BORDER_COLOR);
	box.fill(BOX_SHADOW_COLOR, 1, 1, box.width() - 2, box.height() - 2);
	box.fill(BOX_INNER_COLOR, 2, 2, box.width() - 4, box.height() - 4);
	
	text.draw(box, 4, 4);
	titleRenderObject.setSurface(box);
	titleRenderObject.moveTo((GAME_WIDTH - box.width()) / 2 , 30);
	
	titlePos = Vector(GAME_WIDTH / 2 , 30 + box.height() / 2);
}

void Menu::update()
{
	if (titleRotation > PI)
	{
		titleRotation -= PI_TWO;
	}
	else if (titleRotation < -PI)
	{
		titleRotation += PI_TWO;
	}
	
	if (!titleDrag)
	{
		if (titleRotationSpeed > titleRotationSpeedLimit)
		{
			titleRotationSpeed = titleRotationSpeedLimit;
		}
		if (titleRotationSpeed < -titleRotationSpeedLimit)
		{
			titleRotationSpeed = -titleRotationSpeedLimit;
		}
		
		if ((titleOldRotation <  0 && titleRotation >= 0) ||
		    (titleOldRotation >= 0 && titleRotation <  0))
		{
			// This is when the speed is at its highest. The upper speed
			// limit is slowly lowered. Normally this has no effect at 
			// all but 1 out of 100 used to get stuck without it.
			titleRotationSpeedLimit = std::abs(titleRotationSpeed) * 0.98;
		}
		
		if (titleRotation > 0)
		{
			titleRotationSpeed -= 0.01; 
		}
		else if (titleRotation < 0)
		{
			titleRotationSpeed += 0.01; 
		}
		
		if (std::abs(titleRotation) < 0.025 && std::abs(titleRotationSpeed) < 0.025)
		{
			titleRotationSpeed = 0; 
			titleRotation = 0; 
		}
		else
		{
			titleRotationSpeed *= 0.9;
		}
		
		titleOldRotation = titleRotation;
		titleRotation += titleRotationSpeed;
	}
	else
	{
		titleRotationSpeed = titleRotation - titleOldRotation;
		if (titleRotationSpeed < -PI)
		{
			titleRotationSpeed += PI_TWO;
		}
		else if (titleRotationSpeed > PI)
		{
			titleRotationSpeed -= PI_TWO;
		}

		titleOldRotation = titleRotation;
		titleRotationSpeedLimit = 1;
	}
	titleRenderObject.setRotation(titleRotation);
}

std::size_t Menu::getItem() const
{
	return selectedItem;
}

void Menu::setItem(std::size_t i)
{
	if (i < items.size())
	{
		std::size_t prevItem = selectedItem;
		selectedItem = i;
		selectedItemState = ITEM_SELECTED;
		updateItem(selectedItem);
		updateItem(prevItem);
	}
}

bool Menu::isActive() const
{
	return selectedItemState == ITEM_ACTIVE;
}

ItemState Menu::getItemState() const
{
	return selectedItemState;
}

bool Menu::isVisible() const
{
	return state != 0;
}

void Menu::hide()
{
	std::vector<Item>::iterator it;
	for (it = items.begin(); it != items.end(); ++it)
	{
		state->removeRenderObject(it->renderObject);
	}
	
	state = 0;
}

void Menu::show(GameState* gs)
{
	std::vector<Item>::iterator it;
	for (it = items.begin(); it != items.end(); ++it)
	{
		gs->addRenderObject(it->renderObject);
	}
	
	state = gs;
	
	std::size_t prevItem = selectedItem;
	selectedItem = 0;
	selectedItemState = ITEM_SELECTED;
	hoverItem = -1;
	updateItem(selectedItem);
	updateItem(prevItem);
}
