/*
 * Copyright 2010, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "NearestNeighbor.h"
#include <cstddef>
#include <cstring>
#include <cassert>

template <typename T>
static inline void zoomBlit(const SDL_Rect& srcArea, int zoom, Sint16 x, Sint16 y, SDL_Surface* source, SDL_Surface* destination)
{
	T *dst = (T*) destination->pixels;
	T *src = (T*) source->pixels;
	
	dst += x + srcArea.x * zoom + (srcArea.y * zoom + y) * destination->pitch / sizeof(T);
	src += srcArea.x + srcArea.y * source->pitch / sizeof(T);
	
	std::ptrdiff_t src_skip = source->pitch / sizeof(T) - srcArea.w;
	std::ptrdiff_t dst_skip = destination->pitch / sizeof(T) - srcArea.w * zoom;
	
	std::ptrdiff_t dst_rowskip = destination->pitch / sizeof(T);
	
	// blit!
	int h = srcArea.h;
	while (h--) {
		T *dstTemp = dst;
		T *end_of_line = src + srcArea.w;
		while (src != end_of_line) {
			for (int zx = 0; zx < zoom; zx++) {
				*dstTemp = *src;
				dstTemp++;
			}
			src++;
		}
		dstTemp += dst_skip;
		// copy above row
		for (int zy = 1; zy < zoom; zy++) {
			std::memcpy(dstTemp, dst, srcArea.w * zoom * sizeof (T));
			dstTemp += dst_rowskip;
		}
		src += src_skip;
		dst += dst_rowskip * zoom;
	}
}

static inline void zoomBlit24(const SDL_Rect& srcArea, int zoom, Sint16 x, Sint16 y, SDL_Surface* source, SDL_Surface* destination)
{
	Uint8 *dst = (Uint8*) destination->pixels;
	Uint8 *src = (Uint8*) source->pixels;
	
	dst += 3 * (x + srcArea.x * zoom) + (srcArea.y * zoom + y) * destination->pitch;
	src += 3 * srcArea.x + srcArea.y * source->pitch;
	
	std::ptrdiff_t src_skip = source->pitch - 3 * srcArea.w;
	std::ptrdiff_t dst_skip = destination->pitch - 3 * srcArea.w * zoom;
	
	std::ptrdiff_t dst_rowskip = destination->pitch;
	
	// blit!
	int h = srcArea.h;
	while (h--) {
		Uint8 *dstTemp = dst;
		Uint8 *end_of_line = src + 3 * srcArea.w;
		while (src != end_of_line) {
			for (int zx = 0; zx < zoom; zx++) {
				dstTemp[0] = src[0];
				dstTemp[1] = src[1];
				dstTemp[2] = src[2];
				dstTemp += 3;
			}
			src += 3;
		}
		dstTemp += dst_skip;
		// copy above row
		for (int zy = 1; zy < zoom; zy++) {
			std::memcpy(dstTemp, dst, srcArea.w * zoom * 3);
			dstTemp += dst_rowskip;
		}
		src += src_skip;
		dst += dst_rowskip * zoom;
	}
}

void NearestNeighbor::renderZoom(const SDL_Rect& srcArea, SDL_Rect& dstArea, SDL_Surface* src, SDL_Surface* dest)
{
	dstArea.x = screenArea.x + srcArea.x * zoom;
	dstArea.y = screenArea.y + srcArea.y * zoom;
	dstArea.w = srcArea.w * zoom;
	dstArea.h = srcArea.h * zoom;
	
	int bpp = dest->format->BytesPerPixel;
	switch (bpp) {
	case 3:
		zoomBlit24(srcArea, zoom, screenArea.x, screenArea.y, src, dest);
		break;
	case 4:
		zoomBlit<Uint32>(srcArea, zoom, screenArea.x, screenArea.y, src, dest);
		break;
	}
}

void NearestNeighbor::renderStretch(const SDL_Rect& srcArea, SDL_Rect& dstArea, SDL_Surface* source, SDL_Surface* destination)
{
	SDL_Rect bufferArea;
	bufferArea.x = srcArea.x * zoom;
	bufferArea.y = srcArea.y * zoom;
	bufferArea.w = srcArea.w * zoom;
	bufferArea.h = srcArea.h * zoom;
	
	int bpp = destination->format->BytesPerPixel;
	switch (bpp) {
	case 3:
		zoomBlit24(srcArea, zoom, 0, 0, source, buffer);
		break;
	case 4:
		zoomBlit<Uint32>(srcArea, zoom, 0, 0, source, buffer);
		break;
	}
	
	dstArea.x = src2dst_x(srcArea.x);
	dstArea.y = src2dst_y(srcArea.y);
	dstArea.w = src2dst_x(srcArea.x + srcArea.w) - dstArea.x;
	dstArea.h = src2dst_y(srcArea.y + srcArea.h) - dstArea.y;
	
	stretchArea(bufferArea, dstArea, 0, 0, buffer, destination);
}
