/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef NETWORK_H
#define NETWORK_H

#include "NetworkMessage.h"
#include "Chat.h"
#include "ChatInput.h"
#include "GameList.h"
#include "TrackState.h"
#include "CircularBuffer.h"
#include "Map.h"
#include "RaceSetup.h"
#include <set>
#include <map>
#include <string>

class Track;

enum
{
	NETWORK_GAME_STATE_BLANK,
	NETWORK_GAME_STATE_RACE,
	NETWORK_GAME_STATE_TIMES,
	NETWORK_GAME_STATE_POINTS
};

enum
{
	JOIN_NOT_STARTED,
	JOIN_MORE_CARS,
	JOIN_NO_MORE_CARS,
	JOIN_HAS_ENDED,
	JOIN_ERROR
};

struct JoinStatus
{
	bool updated;
	IPaddress address;
	int state;
	int joinedPlayers;
	Uint8 reservedCars;
	Uint32 lastTimestamp;
};

enum ConnectionType
{
	NET_LOBBY,
	NET_GAME_OWNER,
	NET_GAME_PLAYER
};

struct NetworkGame
{
	std::string name;
	int maxPlayers;
	int maxPlayersPerClient;
	int laps;
	int track;
	int tracks;
	bool tournament;
	bool collisions;
	//int players;
};

struct NetworkGameInfo
{
	std::string name;
	int players;
	int maxPlayers;
	int track;
	int tracks;
	bool tournament;
	bool collisions;
	Uint32 ping;
};


struct Connection
{
	Connection()
	:	lastRecive(0),
		lastSend(0),
		lastChatMessageRecived(0),
		lastChatMessageDelivered(0),
		lastTimestamp(0),
		isBot(false),
		gameInfo()
	{
	}
	Uint32 lastRecive;
	Uint32 lastSend;
	Uint32 lastChatMessageRecived;
	Uint32 lastChatMessageDelivered;
	Uint32 lastTimestamp;
	bool isBot;
	NetworkGameInfo *gameInfo;
};

struct SentChatMessage
{
	std::string text;
	Uint32 timestamp;
	int car;
};

struct JoinInProgress
{
	Uint32 timeout;
	int reservedPlayers;
	Uint8 reservedCars;
	std::vector<int> cars;
	Uint32 lastTimestamp;
	int maxPlayers;
	bool noMore;
};

struct NetworkRaceCar
{
	class GamePlayer* player;
	std::string name;
	int car;
	int id;
	int time;
	int points;
	int driver;
	bool human;

};

struct GamePlayer
{
	std::vector<NetworkRaceCar> cars;
	std::vector<int> removedCars;
	int gameStateId;
	Uint32 lastTimestamp;
	Uint32 lastRecive;
	Uint32 lastTick;
	Uint32 inputMask;
	Uint32 ticksTooFast;
	Uint32 lastSent;
	
	Uint32 lastChatMessageRecived;
	Uint32 lastChatMessageDelivered;
	bool mute;
};

inline bool operator<(const IPaddress& addr1, const IPaddress& addr2)
{
	return (addr1.host == addr2.host) ? (addr1.port < addr2.port) : (addr1.host < addr2.host);
}

inline bool operator==(const IPaddress& addr1, const IPaddress& addr2)
{
	return addr1.host == addr2.host && addr1.port == addr2.port;
}

inline bool operator!=(const IPaddress& addr1, const IPaddress& addr2)
{
	return !(addr1 == addr2);
}

class Network
{
public:
	Network();
	virtual ~Network();
	virtual void step();
	void updateTrack(bool noInput);
	void setStatus(const std::string& str);
	const std::string& getError() const;
	virtual std::string name() const = 0;
	void setGameState(GameState* gameState, bool chatOn = true, bool gameListOn = false);
	void enableChatInput(bool chatInputOn);
	bool onEvent(const SDL_Event& event);
	void setGame(const std::string& name, int maxPlayers, int maxPlayersPerClient, int laps, int tracks, int collisions);
	void createGame(Uint8 reservedCars, std::vector<int> cars);
	GameList& getGameList();
	void sendJoinStart();
	void sendJoinCar(int car);
	void sendJoinEnd();
	void join(const IPaddress& address);
	bool getJoinStatus(JoinStatus&);
	GameState* getNewGameState();
	void join();
	void onRaceFinished(const RaceSetup& rs);
	void onShowPoints(const RaceSetup& rs);
	bool isOwner() const;
	void startRace();
	void sleep(Uint32 duration);
	void sendChatMessage(const std::string& text, int car = -1);
	void quitGame();
	void erasePlayer(std::map<IPaddress, GamePlayer>::iterator it, const std::string& reason = "");
	std::vector<NetworkRaceCar>::iterator eraseCar(std::vector<NetworkRaceCar>::iterator it, const std::string& reason = "");
	void eraseCar(int carId);
	void writeChatMessage();
	void writeChatMessage(const GamePlayer& gp);
	void readChatMessage(GamePlayer& gp);
	bool readChatMessage();
	std::string uniqueName(const std::string& name) const;
	void lobby();
	int numberOfPlayersWaiting() const;
protected:
	void setNewGameState(GameState*);
	virtual void handleMessage(Uint8 msgType, IPaddress address);
	virtual void onConnectionTimeout(const IPaddress& addr);
	virtual void refreshAsNeeded() = 0;
	void send(IPaddress address);
	bool receive();
private:
	void sendJoinStatus(const IPaddress& address, int state, Uint8 joinedPlayers, Uint8 reservedCars);
	void sendLobbyMessage(Connection& connection, const IPaddress& address);
	
protected:
	std::map<IPaddress, JoinInProgress> joinInProgress;
	
	UDPsocket socket;
	
	std::map<IPaddress, Connection> connections;
	std::map<IPaddress, Connection>::iterator connectionsIt;
	
	NetworkMessage msg;

	enum 
	{
		LOBBY_CONNECTION = 0x1,
		LOBBY_CHAT       = 0x2
	};
	ConnectionType connectionType;
	std::string error;
	
	
private:
	RenderObject statusRenderObject;
	Chat chat;
	ChatInput chatInput;
	GameState* gameState;
	
	std::deque<SentChatMessage> chatBuffer;
	Uint32 lastChatMessage;
	bool chatOn;
	NetworkGame game;
	GameList gameList;
	
	std::map<IPaddress, GamePlayer> gameClients;
	Uint8 reservedCars;
	int reservedPlayers;
	
	JoinStatus joinStatus;
	GameState* newGameState;
	int gameStateId;
	Uint32 gameStateCount;
	
	IPaddress gameOwner;
	Uint32 lastGameTimestamp;
	Uint32 lastGameRecive;
	Uint32 lastGameSent;
	static const std::size_t TRACK_BUFFER_SIZE = 10;
	CircularBuffer<TrackState, TRACK_BUFFER_SIZE> trackStateBuffer;
	bool newTrackInfoReady;
	int newTrackInfoCounter;
	Uint32 newTrackStateTick;
	static const std::size_t INPUT_BUFFER_SIZE = 50;
	CircularBuffer<Uint32, INPUT_BUFFER_SIZE> inputBuffer;
	Uint32 raceTick;
	Uint32 inputMask;
	Uint32 lastSlowDown;
	Map map;
	Track* track;
	std::vector<NetworkRaceCar> localCars;
	std::vector<NetworkRaceCar*> raceCars;
	Uint32 lastChatMessageDeliveredToGameOwner;
	Uint32 lastChatMessageRecivedFromGameOwner;
	int gamePlayers;
	std::vector<int> removedCars;
	std::set<Uint32> blacklist;
};

enum
{
	MSG_PING,
	MSG_PONG,
	MSG_LOBBY,
	MSG_GAME_INFO,
	MSG_JOIN_START,
	MSG_JOIN_CAR,
	MSG_JOIN_END,
	MSG_JOIN_STATUS,
	MSG_NEW_STATE_RACE,
	MSG_NEW_STATE_TIMES,
	MSG_NEW_STATE_POINTS,
	MSG_STATE_STATUS,
	MSG_GAME_CHAT_INFO,
	MSG_GAME_CHAT_ADD,
	MSG_RACE_STATE,
	MSG_RACE_INPUT,
	MSG_QUIT_GAME,
	MSG_END_GAME,
	MSG_JOIN_QUIT
};

inline IPaddress createAddress(Uint32 ip, Uint16 port)
{
	IPaddress addr;

	SDLNet_Write32(ip, &addr.host);
	SDLNet_Write16(port, &addr.port);
	
	return addr;
}

inline Uint32 getIP(IPaddress addr)
{
	return SDLNet_Read32(&addr.host);
}

inline Uint16 getPort(IPaddress addr)
{
	return SDLNet_Read16(&addr.port);
}

#endif
