/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "NetworkMenu.h"
#include "ConnectToNetwork.h"
#include "CreateNetworkGame.h"
#include "JoinNetworkGame.h"
#include "StateManager.h"
#include "Network.h"
#include <sstream>

NetworkMenu::NetworkMenu(Network* net)
:	net(net),
	joinAction(false),
	hostAction(false),
	disconnectAction(false),
	gameList(net->getGameList()),
	numberOfPlayersWaiting(-1)
{
	net->lobby();
	
	menu.setTitle(net->name() + " games");
	
	menu.addItem("Join", &joinAction);
	menu.addItem("Create", &hostAction);
	menu.addItem("Disconnect", &disconnectAction);
	
	menu.addRenderObjects(this, CENTER_X, 210);
	net->setGameState(this, true, true);
	
	addRenderObject(playersWaitingRenderObject);
}

void NetworkMenu::update()
{
	updatePlayersOnlineCounter();
	
	net->step();
	menu.update();
}

void NetworkMenu::onEvent(const SDL_Event& event)
{
	IPaddress joinAddress = IPaddress();
	if (!net->onEvent(event))
	{
		StateManager& sm = StateManager::getInstance();
		MenuEvent me = sm.getMenuEvent(event);
		if (gameList.focus())
		{
			joinAddress = gameList.onEvent(event);
			if (me.type == MENU_MOUSE_MOVE ||
				(!gameList.focus() &&
			    (me.type == MENU_MOUSE_CLICK1 ||
			     me.type == MENU_MOUSE_CLICK2)) ||
				 me.type == MENU_MOUSE_RELEASE)
			{
				if (!gameList.focus() ||
				    menu.getItem(me.x, me.y) > 0)
				{
					gameList.focus(false);
					menu.makeActive(ITEM_UNSELECTED);
				}
				menu.onEvent(me);
				
				if (!menu.isActive())
				{
					gameList.focus(false);
				}
			}
			else if (me.type == MENU_ESC)
			{
				gameList.focus(false);
				menu.makeActive(ITEM_SELECTED);
			}
		}
		else
		{
			menu.onEvent(me);
			if (me.type == MENU_MOUSE_MOVE ||
			    me.type == MENU_MOUSE_CLICK1 ||
			    me.type == MENU_MOUSE_CLICK2 ||
				me.type == MENU_MOUSE_RELEASE ||
				(event.type == SDL_MOUSEBUTTONDOWN && (event.button.button == SDL_BUTTON_WHEELUP || event.button.button == SDL_BUTTON_WHEELDOWN)))
			{
				joinAddress = gameList.onEvent(event);
				if (gameList.focus())
				{
					gameList.focus(true);
					menu.setItem(0);
					menu.makeActive();
				}
			}
			
			if (disconnectAction || me.type == MENU_ESC)
			{
				sm.enter(new ConnectToNetwork(), -1);
			}
			else if (hostAction)
			{
				sm.enter(new CreateNetworkGame(net), +1);
			}
			else if (joinAction)
			{
				gameList.focus(true);
				joinAction = false;
				menu.makeActive();
			}
		}

		if (joinAddress.host != 0)
		{
			sm.enter(new JoinNetworkGame(net, joinAddress), +1);
		}
	}
}

void NetworkMenu::updatePlayersOnlineCounter()
{
	int newNumberOfplayersWaiting = net->numberOfPlayersWaiting();
	if (numberOfPlayersWaiting != newNumberOfplayersWaiting)
	{
		numberOfPlayersWaiting = newNumberOfplayersWaiting;
		
		// Build text.
		std::ostringstream oss;
		if (numberOfPlayersWaiting == 0)
		{
			oss.str("You are the only player waiting here");
		}
		else if (numberOfPlayersWaiting == 1)
		{
			oss.str("There is one other player waiting here");
		}
		else
		{
			oss << "There are " << numberOfPlayersWaiting << " other players waiting here";
		}
		
		// Update render object.
		Surface surface(oss.str(), SMALL_FONT, TEXT_COLOR, TEXT_BORDER_COLOR);
		playersWaitingRenderObject.setSurface(surface);
		playersWaitingRenderObject.moveTo((GAME_WIDTH - surface.width()) / 2, GAME_HEIGHT - 9);
	}
}
