/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "NetworkMessage.h"
#include "str.h"
#include <cstring>
#include <cassert>

// Use these functions instead of SDLNet_WriteXX and SDLNet_ReadXX to 
// avoid breaking strict-aliasing rules when used on a char array.
static void netWrite16(Uint16 value, void *area)
{
	value = SDL_SwapBE16(value);
	std::memcpy(area, &value, sizeof(value));
}
static void netWrite32(Uint32 value, void *area)
{
	value = SDL_SwapBE32(value);
	std::memcpy(area, &value, sizeof(value));
}
static Uint16 netRead16(void *area)
{
	Uint16 value;
	std::memcpy(&value, area, sizeof(value));
	return SDL_SwapBE16(value);
}
static Uint32 netRead32(void *area)
{
	Uint32 value;
	std::memcpy(&value, area, sizeof(value));
	return SDL_SwapBE32(value);
}

NetworkMessage::NetworkMessage()
:	buffer(),
	readPos(0)
{
	packet.channel = -1;
	packet.data = buffer;
	packet.len = 0;
	packet.maxlen = MAX_MESSAGE_LENGTH;
	packet.status = 0;
	packet.address.host = INADDR_NONE;
	packet.address.port = 0;
}

void NetworkMessage::write8(Uint8 value)
{
	if (packet.len + 1 > MAX_MESSAGE_LENGTH)
	{
		return;
	}
	buffer[packet.len] = value;
	packet.len += 1;
}
void NetworkMessage::write16(Uint16 value)
{
	if (packet.len + 2 > MAX_MESSAGE_LENGTH)
	{
		return;
	}
	netWrite16(value, &buffer[packet.len]);
	packet.len += 2;
}
void NetworkMessage::write32(Uint32 value)
{
	if (packet.len + 4 > MAX_MESSAGE_LENGTH)
	{
		return;
	}
	netWrite32(value, &buffer[packet.len]);
	packet.len += 4;
}
void NetworkMessage::write64(Uint64 value)
{
	write32(value >> 32);
	write32(value);
}
void NetworkMessage::writeAddress(const IPaddress& addr)
{
	if (packet.len + 6 > MAX_MESSAGE_LENGTH)
	{
		return;
	}
	memcpy(&buffer[packet.len], &addr.host, 4);
	packet.len += 4;
	memcpy(&buffer[packet.len], &addr.port, 2);
	packet.len += 2;
}
void NetworkMessage::writeString(const std::string& str)
{
	if (packet.len + int(str.length()) + 1 > MAX_MESSAGE_LENGTH)
	{
		return;
	}
	write8(str.length());
	memcpy(&buffer[packet.len], str.data(), str.length());
	packet.len += str.length();
}
Uint8 NetworkMessage::read8()
{
	if (readPos + 1 > packet.len)
	{
		return 0;
	}
	Uint8 value = buffer[readPos];
	readPos += 1;
	return value;
}
Uint16 NetworkMessage::read16()
{
	if (readPos + 2 > packet.len)
	{
		return 0;
	}
	Uint16 value = netRead16(&buffer[readPos]);
	readPos += 2;
	return value;
}
Uint32 NetworkMessage::read32()
{
	if (readPos + 4 > packet.len)
	{
		return 0;
	}
	Uint32 value = netRead32(&buffer[readPos]);
	readPos += 4;
	return value;
}

Uint64 NetworkMessage::read64()
{
	Uint64 hi = read32();
	Uint64 lo = read32();
	return (hi << 32) | lo;
}

IPaddress NetworkMessage::readAddress()
{
	IPaddress addr = IPaddress();
	if (readPos + 6 <= packet.len)
	{
		memcpy(&addr.host, &buffer[readPos], 4);
		readPos += 4;
		memcpy(&addr.port, &buffer[readPos], 2);
		readPos += 2;
	}
	return addr;
}

std::string NetworkMessage::readString()
{
	Uint8 stringLength = read8();
	std::string str;
	if (readPos + stringLength <= packet.len)
	{
		str.resize(stringLength);
		memcpy(&str[0], &buffer[readPos], stringLength);
		readPos += stringLength;
	}
	toLower(str);
	return str;
}

UDPpacket& NetworkMessage::getPacket()
{
	return packet;
}

void NetworkMessage::reset()
{
	readPos = 0;
	packet.len = 0;
}
