/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "NetworkSettingsMenu.h"
#include "StateManager.h"
#include "SettingsMenu.h"
#include "Settings.h"
#include "str.h"
#include <sstream>
#include <algorithm>

template <typename T>
std::string vec2string(const std::vector<T>& vec)
{
	std::string str;
	if (!vec.empty())
	{
		str = int2string(vec[0]);
		for (std::size_t i = 1; i < vec.size(); ++i)
		{
			str += ", " + int2string(vec[i]);
		}
	}
	return str;
}

NetworkSettingsMenu::NetworkSettingsMenu()
:	lanPorts(vec2string(Settings::network.lanPorts)),
	internetPorts(vec2string(Settings::network.internetPorts)),
	server(Settings::network.server),
	maxConnections(Settings::network.maxConnections),
	chatFont(Settings::network.chatFont),
	maxChatMessages(Settings::network.maxChatMessages),
	chatMessageDuration(int2string(Settings::network.chatMessageDuration)),
	save(false),
	cancel(false)
{
	menu.setTitle("Network Settings");
	
	menu.addItem("LAN ports", &lanPorts, 250);
	menu.addItem("Internet ports", &internetPorts, 250);
	menu.addItem("Internet server", &server, 250);
	
	menu.addItem("Max connections", &maxConnections, 1, 1000);
	
	menu.addItem("Chat text size", &chatFont, 0, 1, "small", "big");
	menu.addItem("Max chat messages", &maxChatMessages, 1, 20);
	
	std::vector<std::string> durations;
	durations.push_back("1 second");
	for (int seconds = 2; seconds <= 60; ++seconds)
	{
		durations.push_back(int2string(seconds) + " seconds");
	}
	durations.push_back("infinite");
	menu.addItem("Chat message duration", &chatMessageDuration, 20, durations);
	
	menu.addItem("Save settings", &save);
	menu.addItem("Cancel", &cancel);
	
	// Set proper string values.
	onValueUpdate(&lanPorts);
	onValueUpdate(&internetPorts);
	onValueUpdate(&chatMessageDuration);
	
	menu.addRenderObjects(this);
}

void NetworkSettingsMenu::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	
	if (menu.onEvent(me))
	{
		if (save)
		{
			Settings::network.internetPorts = internetPortsVec;
			Settings::network.lanPorts = lanPortsVec;
			Settings::network.server = server;
			Settings::network.maxConnections = maxConnections;
			Settings::network.chatFont = chatFont;
			Settings::network.maxChatMessages = maxChatMessages;
			Settings::network.chatMessageDuration = chatMessageDurationValue;
		
			sm.enter(new SettingsMenu(), -1);
			return;
		}
	}
	else if (me.type == MENU_ESC)
	{
		cancel = true;
	}
	if (cancel)
	{
		sm.enter(new SettingsMenu(), -1);
	}

}

void NetworkSettingsMenu::onValueUpdate(std::string* strValue)
{
	if (strValue == &lanPorts ||
	    strValue == &internetPorts)
	{
		// Remove everything but digits
		for (std::size_t i = 0; i < strValue->size(); ++i)
		{
			char& c = (*strValue)[i];
			if (!std::isdigit(c))
			{
				c = ' ';
			}
		}
		
		std::replace(strValue->begin(), strValue->end(), ',', ' ');
		std::istringstream iss(*strValue);
		std::vector<Uint16>& ports = (strValue == &lanPorts ? lanPortsVec : internetPortsVec);
		Uint16 port;
		ports.clear();
		while (iss >> port)
		{
			if (port != 0)
			{
				ports.push_back(port);
			}
		}
		
		if (ports.empty())
		{
			if (strValue == &lanPorts)
			{
				for (int port = DEFAULT_LAN_PORTS_FIRST; port <= DEFAULT_LAN_PORTS_LAST; ++port)
				{
					ports.push_back(port);
				}
			}
			else
			{
				*strValue = "auto";
				return;
			}
		}
		
		*strValue = vec2string(ports);
	}
	else if (strValue == &chatMessageDuration)
	{
		std::istringstream iss(chatMessageDuration);
		int duration;
		if (iss >> duration && duration > 0)
		{
			char unit;
			if (iss >> unit)
			{
				switch (unit)
				{
				case 'h': // hours
					duration *= 60;
					//* fall through! *//
				case 'm': // minutes
					duration *= 60;
					break;
				}
			}
			duration = std::min(100000, std::max(1, duration));
			chatMessageDuration = int2string(duration) + " second" + (duration > 1 ? "s" : "");
			chatMessageDurationValue = duration;
		}
		else
		{
			chatMessageDuration = "infinite";
			chatMessageDurationValue = -1;
		}
	}
}
