/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Pointer.h"
#include "Path.h"
#include <algorithm>

// This class handles the visual pointer on the screen, together with 
// the StateManager class.
//
// The pointer is hidden if the user doesn't use the mouse, but only in
// fullscreen mode. In window mode it's easy to get the pointer out of 
// the way and the user probably want to know where the pointer is at 
// all time. 
// 
// By default the pointer is only scaled in fullscreen mode because then 
// we try to create the perfect illusion of an old-school pixelated 
// game. When in window mode the user probably want the mouse to behave 
// pretty much like in any other window.
//
// For technical reasons the scaled pointer is not shown in the black 
// padding area. If the user chose to use scaled pointer in window mode 
// this wouldn't be very user-friendly so for that reason a small cursor
// is displayed in window mode when the pointer is located in the
// padding area just to let the user know where the pointer is.

const int HIDE_POINTER = 500; // 10 seconds

Pointer::Pointer()
:	offsetX(0),
	offsetY(0),
	hideTimer(0),
	visible(true),
	scale(true),
	cursor(0),
	fullscreen(false),
	outside(false)
{
}

void Pointer::move(int x, int y)
{
	if (!visible)
	{
		show();
	}
	ro.moveTo(x - offsetX, y - offsetY);
	hideTimer = 0;
	
	if (scale && !fullscreen)
	{
		if (x < 0 || x >= 320 ||
		    y < 0 || y >= 240)
		{
			if (!outside)
			{
				SDL_ShowCursor(SDL_ENABLE);
				outside = true;
			}
		}
		else
		{
			if (outside)
			{
				SDL_ShowCursor(SDL_DISABLE);
				outside = false;
			}
		}
	}
}

void Pointer::hide(bool force)
{
	if (!fullscreen && !force)
	{
		// This disables hidden pointers in window mode. Simply remove 
		// this if statement to allow hidden pointer window mode.
		return;
	}
	visible = false;
	if (!scale)
	{
		SDL_ShowCursor(SDL_DISABLE);
	}
	else
	{
		ro.flagUpdated();
		if (!fullscreen && outside)
		{
			SDL_ShowCursor(SDL_DISABLE);
			
			// This is a lie but it makes so that the cursor 
			// will be shown again when it's moved.
			outside = false;
		}
	}
}

void Pointer::show()
{
	visible = true;
	hideTimer = 0;
	if (!scale)
	{
		SDL_ShowCursor(SDL_ENABLE);
	}
	else
	{
		ro.flagUpdated();
	}
}

void Pointer::setScaling(bool scaling)
{
	scale = scaling;
	outside = false;
	
	if (scale)
	{
		SDL_FreeCursor(cursor);
		
		if (fullscreen)
		{
			// To prevent a bug on Linux that messes up the mouse sensitivity 
			// in fullscreen mode we have to show an invisible SDL cursor.
			Uint8 data = 0;
			Uint8 mask = 0;
			cursor = SDL_CreateCursor(&data, &mask, 8, 1, 0, 0);
			SDL_ShowCursor(SDL_ENABLE);
		}
		else
		{
			// Scaled pointers are not visible in the black border for technical 
			// reasons, but that is not very user friendly in window mode so for 
			// that reason we use a small cross as the cursor.
			Uint8 data[3] = {};
			Uint8 mask[3] = {0x40, 0xE0, 0x40};
			cursor = SDL_CreateCursor(data, mask, 8, 3, 1, 1);
			SDL_ShowCursor(SDL_DISABLE);
		}
		
		SDL_SetCursor(cursor);
	}
}

void Pointer::setFullScreen(bool fullscreen)
{
	this->fullscreen = fullscreen;
}

void Pointer::setType(const std::string& name)
{
	Surface surf("pointers/" + name + ".png");
	
	// The cursor width must be a multiple of 8 bits. 
	int w = surf.width() + (8 - surf.width() % 8);
	int h = surf.height();
	int bytes = w * h / 8;
	
	std::vector<Uint8> data(bytes);
	std::vector<Uint8> mask(bytes);
	
	offsetX = 0;
	offsetY = 0;
	
	for (int y = 0; y < h; ++y)
	{
		for (int x = 0; x < w; ++x)
		{
			int dataValue = 0;
			int maskValue = 0;
			if (x < surf.width())
			{
				Color c = surf.getColor(x, y);
				if (c != DEFAULT_COLORKEY)
				{
					maskValue = 1;
					Uint8 maxColor = std::max(red(c), std::max(green(c), blue(c)));
					if (maxColor < 128)
					{
						dataValue = 1;
					}
					if (c != BLACK && c != WHITE)
					{
						// This marks the hot spot!
						offsetX = x;
						offsetY = y;
						
						surf.setColor(x, y, dataValue ? BLACK : WHITE);
					}
				}
			}
			int i = (x + y * w) / 8;
			data[i] = (data[i] << 1) | dataValue;
			mask[i] = (mask[i] << 1) | maskValue;
		}
	}
	
	if (scale)
	{
		surf.setColorKey(DEFAULT_COLORKEY);
		ro.setSurface(surf);
	}
	else
	{
		SDL_FreeCursor(cursor);
		cursor = SDL_CreateCursor(&data[0], &mask[0], w, h, offsetX, offsetY);
		SDL_SetCursor(cursor);
		SDL_ShowCursor(SDL_ENABLE);
	}
}

void Pointer::draw(Surface& screen)
{
	if (scale && visible)
	{
		ro.draw(screen);
	}
}

bool Pointer::isUpdated() const
{
	return scale && ro.isUpdated();
}

bool Pointer::isVisible() const
{
	return visible;
}

void Pointer::updateRects(Rect& newRect, Rect& oldRect)
{
	ro.updateRects(newRect, oldRect);
}

void Pointer::update(bool notNeeded)
{
	if (notNeeded)
	{
		++hideTimer;
		if (hideTimer == HIDE_POINTER)
		{
			hide();
		}
	}
	else
	{
		hideTimer = 0;
	}
}

bool isPNG(const std::string& filename);

const std::vector<std::string>& Pointer::getList()
{
	static std::vector<std::string> pointers;
	if (pointers.empty())
	{
		std::vector<std::string> files = (dataPath() + "pointers").files();
		for (std::size_t i = 0; i < files.size(); ++i)
		{
			if (isPNG(files[i]))
			{
				pointers.push_back(files[i].substr(0, files[i].length() - 4));
			}
		}
		
		// Make sure not empty.
		if (pointers.empty())
		{
			pointers.push_back("?");
		}
		
		std::sort(pointers.begin(), pointers.end());
	}
	return pointers;
}
