/*
 * Copyright 2010, 2011, 2012, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Map.h"
#include "RaceState.h"
#include "SurfaceBank.h"
#include "StateManager.h"
#include "ResultState.h"
#include "TimeTrialResultState.h"
#include "TileSet.h"
#include "Network.h"
#include "NetworkMenu.h"
#include "MainMenu.h"
#include "SelectNewTrack.h"
#include "SelectTrack.h"
#include "str.h"
#include <algorithm>
#include <cassert>
#include <climits>

const Color STAT_COLOR = 0;
const Uint32 TIME_STATE_DURATION = 800;

const Color READY_SET_GO_COLOR  = 0xFFFFFF;
const Color READY_SET_GO_BORDER = 0x000000;

bool RaceState::CarProgress::operator<(const RaceState::CarProgress& cp) const
{
	return progress < cp.progress;
}

RaceState::RaceState(const Map& map, const RaceSetup& rs, Network* net)
:   raceSetup(rs),
	track(raceSetup.numberOfLaps(), raceSetup.numberOfPlayers(), raceSetup.getDrivers(), map, raceSetup.carCollisionsEnabled()),
	raceTime(0),
	endRace(false),
	timeState(TIME_STATE_READY),
	map(map),
	net(net),
	closeMenu(false),
	quit(false),
	restartRace(false),
	newTrack(false),
	selectTrack(false)
{
	readyRO.setSurface(Surface("Ready", BIG_FONT, READY_SET_GO_COLOR, READY_SET_GO_BORDER));
	setRO  .setSurface(Surface("Set"  , BIG_FONT, READY_SET_GO_COLOR, READY_SET_GO_BORDER));
	goRO   .setSurface(Surface("Go"   , BIG_FONT, READY_SET_GO_COLOR, READY_SET_GO_BORDER));
	
	int rsgOffsetX = 0;
	int rsgOffsetY = 0;
	switch (track.getFinishDir())
	{
	case UP:
		rsgOffsetY = 1;
		break;
	case DOWN:
		rsgOffsetY = 22;
		break;
	case LEFT:
		rsgOffsetX = -21;
		rsgOffsetY = 12;
		break;
	case RIGHT:
		rsgOffsetX = 22;
		rsgOffsetY = 12;
		break;
	}
	
	int rsgX = track.getFinishX() * TILE_SIZE + rsgOffsetX;
	int rsgY = track.getFinishY() * TILE_SIZE + rsgOffsetY;
	readyRO.moveTo(rsgX, rsgY);
	setRO.moveTo(rsgX + 6, rsgY);
	goRO.moveTo(rsgX + 9, rsgY);
	
	for (int i = 0; i < 10; i++)
	{
		smallNumbers[i] = Surface(int2string(i), SMALL_FONT, STAT_COLOR);
		bigNumbers[i] = Surface(int2string(i), BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR);
	}
	
	TileSet::getTileSet(map.getTilesetName()).decorate(map, background);
	
	if (raceSetup.isTimeTrial())
	{
		TimeTrialTrack* ttt = raceSetup.getTimeTrialTrack();
		ttt->prepareGhosts(*this);
	}
	
	int numCars = raceSetup.numberOfPlayers();
	
	// reserve to avoid invalidated pointers with addRenderObject
	cars.reserve(numCars);

	for (int i = 0; i < numCars; i++)
	{
		Car car;
		car.pos = 0;
		car.lap = 1;
		car.time = 0;
		const Vector& v = track.getCarPosition(i);
		car.carRO.update(v.x - CAR_RADIUS, v.y - CAR_RADIUS, track.getCarRotation(i));
		car.carRO.setSurface(SurfaceBank::getCar(raceSetup.getPlayerCar(i)));
		cars.push_back(car);
		addRenderObject(cars[i].carRO);
		addRenderObject(cars[i].timeRO);
	}
	placeStats();
	
	for (int i = 0; i < 6; i++)
	{
		timeRO[i].setSurface(bigNumbers[0]);
		timeNumbers[i] = 0;
		addRenderObject(timeRO[i]);
	}
	
	addRenderObject(readyRO);
	
	// Start the timer after all other initilization
	gameTime = SDL_GetTicks();
	
	if (net)
	{
		menu.addItem("Back to race", &closeMenu);
		if (net->isOwner())
		{
			menu.addItem("End Race", &endRace);
			menu.addItem("End network game", &quit);
		}
		else
		{
			menu.addItem("Leave network game", &quit);
		}
	}
	else
	{
		menu.addItem("Continue race", &closeMenu);
		if (rs.isTimeTrial())
		{
			menu.addItem("Restart race", &restartRace);
			menu.addItem("New track", &newTrack);
			menu.addItem("Select track", &selectTrack);
		}
		else
		{
			menu.addItem("End Race", &endRace);
		}
		menu.addItem("Quit to main menu", &quit);
	}
	
	menu.addRenderObjects(this, CENTER_X, 112);
	menu.hide();
	StateManager::getInstance().forgetLastMenuEvent();
	
	if (net)
	{
		net->setGameState(this);
		net->enableChatInput(false);
	}
}

const int CATCH_UP_LIMIT = 50;

void RaceState::update()
{
	StateManager& sm = StateManager::getInstance();

	if (net)
	{
		GameState* newState = net->getNewGameState();
		if (newState)
		{
			sm.enter(newState, 0);
			return;
		}
	}
	
	int updates = 0;
	do
	{
		if (net)
		{
			net->updateTrack(menu.isVisible());
		}
		else if (!menu.isVisible())
		{
			track.update(TRACK_TIME_STEP * 0.001);
			if (raceSetup.isTimeTrial() && track.getTimeState() >= TIME_STATE_GO)
			{
				const Vector& v = track.getCarPosition(0);
				double rotation = track.getCarRotation(0);
				
				int x = static_cast<int>(v.x - CAR_RADIUS);
				int y = static_cast<int>(v.y - CAR_RADIUS);
				
				TimeTrialTrack* ttt = raceSetup.getTimeTrialTrack();
				ttt->ghostUpdate(x, y, rotation);
			}
		}
		gameTime += TRACK_TIME_STEP;
		if (++updates >= CATCH_UP_LIMIT)
		{
			// The game is not able to keep up with the pace so we have 
			// to slow down a bit to avoid an infinite loop here.
			gameTime = SDL_GetTicks();
			break;
		}
		
	} while (gameTime <= SDL_GetTicks());
	
	if (net)
	{
		net->step();
	}
	
	std::vector<int>& removedCars = track.getRemovedCars();
	if (!removedCars.empty())
	{
		removePlayers(removedCars);
		removedCars.clear();
	}
	
	if (endRace) // The user wants to end the race.
	{
		const int MIN_UPDATES = 3000; // 1 minute
		const int MAX_WAIT    = 1000; // 1 second
		track.finish(TRACK_TIME_STEP * 0.001, MIN_UPDATES, MAX_WAIT);
	}
	
	int newTimeState = track.getTimeState();
	if (newTimeState != timeState)
	{
		timeState = newTimeState;
		updateTimeState();
	}
	
	unsigned newRaceTime = 1000 * track.getRaceTime();
	if (newRaceTime != raceTime)
	{
		raceTime = newRaceTime;
		updateTimeRO();
		if (raceTime >= MAX_RACE_TIME )
		{
			if (!net || net->isOwner())
			{
				endRace = true;
			}
		}
	}
	
	int numCars = raceSetup.numberOfPlayers();
	
	for (int i = 0; i < numCars; i++)
	{
		const Vector& v = track.getCarPosition(i);
		double rotation = track.getCarRotation(i);
		int x = static_cast<int>(v.x - CAR_RADIUS);
		int y = static_cast<int>(v.y - CAR_RADIUS);
		cars[i].carRO.update(x, y, rotation);
		
		int finishTime = 1000 * track.getCarFinishTime(i);
		
		// The lap should not be decreased after the car has finished (i.e. car has non-zero finish time).
		const double ALMOST_ZERO = 1.0;
		int lap = (finishTime <= ALMOST_ZERO ? std::max(1, track.getCarLap(i)) : raceSetup.numberOfLaps() + 1);
		
		if (lap != cars[i].lap || finishTime != cars[i].time)
		{
			cars[i].lap = lap;
			cars[i].time = finishTime;
			updateLapRO(i);
		}
		carProgress[i].carId = i;
		carProgress[i].progress = track.getCarProgress(i);
	}
	
	// update position stat
	std::sort(carProgress, carProgress + numCars);
	for (int i = 0; i < numCars; i++)
	{
		int carId = carProgress[i].carId;
		int pos = numCars - i;
		if (cars[carId].pos != pos)
		{
			cars[carId].pos = pos;
			updatePosRO(carId);
		}
	}
	
	if (track.finished() || endRace)
	{		
		if (raceSetup.isTournament())
		{
			raceSetup.nextTrack();
			raceSetup.givePoints();
		}
		if (net)
		{
			net->onRaceFinished(raceSetup);
		}
		else
		{
			if (raceSetup.isTimeTrial())
			{
				sm.enter(new TimeTrialResultState(raceSetup), 0);
			}
			else
			{
				sm.enter(new ResultState(raceSetup), 0);
			}
			return;
		}
	}
}

void RaceState::updateTimeState()
{
	removeRenderObject(readyRO);
	removeRenderObject(setRO);
	removeRenderObject(goRO);
	
	if (!menu.isVisible())
	{
		switch (timeState)
		{
		case TIME_STATE_READY:
			addRenderObject(readyRO);
			break;
		case TIME_STATE_SET:
			addRenderObject(setRO);
			break;
		case TIME_STATE_GO:
			addRenderObject(goRO);
			break;
		}
	}
}

int RaceState::sleepTime()
{
	return gameTime - SDL_GetTicks();
}

void RaceState::updateLapRO(int i)
{
	Car& car = cars[i];
	removeRenderObject(car.lapRO);
	
	if (car.lap <= raceSetup.numberOfLaps())
	{
		if (raceSetup.isTimeTrial())
		{
			car.lapRO.setSurface(bigNumbers[car.lap]);
			addRenderObject(car.lapRO);
			Surface("lap  /" + int2string(raceSetup.numberOfLaps()), BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR).draw(background, 3, 228);
			Surface("   :", BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR).draw(background, 3, 228);
		}
		else
		{
			car.timeRO.moveTo(3 + i * 31, 233);
			car.lapRO.setSurface(smallNumbers[car.lap]);
			addRenderObject(car.lapRO);
			car.timeRO.setSurface(Surface("lap: /" + int2string(raceSetup.numberOfLaps()), SMALL_FONT, STAT_COLOR));
		}
		raceSetup.setPlayerTime(i, INT_MAX); // DNF
	}
	else
	{
		if (!raceSetup.isTimeTrial())
		{
			car.timeRO.moveTo(2 + i * 31, 233);
			car.timeRO.setSurface(Surface(time2string(car.time), SMALL_FONT, STAT_COLOR));
		}
		raceSetup.setPlayerTime(i, car.time);
	}
}

void RaceState::updatePosRO(int i)
{
	assert(cars[i].pos >= 0 && cars[i].pos <= 9);
	cars[i].posRO.setSurface(smallNumbers[cars[i].pos]);
}

void RaceState::updateTimeRO()
{
	Uint32 newTimeNumbers[6] =
	{
		(raceTime / 10) % 10,
		(raceTime / 100) % 10,
		(raceTime / 1000) % 10,
		(raceTime / 10000) % 6,
		(raceTime / 60000) % 10,
		(raceTime / 600000) % 10
	};
	
	for (int i = 0; i < 6; i++)
	{
		if (newTimeNumbers[i] != timeNumbers[i])
		{
			timeRO[i].setSurface(bigNumbers[newTimeNumbers[i]]);
			timeNumbers[i] = newTimeNumbers[i];
		}
	}
}

void RaceState::removePlayers(const std::vector<int>& removedCars)
{
	for (std::size_t i = 0; i < cars.size(); ++i)
	{
		removeRenderObject(cars[i].posRO);
		removeRenderObject(cars[i].lapRO);
	}
	
	for (std::size_t i = 0; i < removedCars.size(); ++i)
	{
		std::size_t carToBeRemoved = removedCars[i];
		assert(i < cars.size());
		raceSetup.removePlayer(carToBeRemoved);
		removeRenderObject(cars.back().carRO);
		removeRenderObject(cars.back().timeRO);
		cars.erase(cars.begin() + carToBeRemoved);
	}
	TileSet::getTileSet(map.getTilesetName()).decorateLower(map, background);
	placeStats();
	
	// Best to remove everything because this stuff messes with the 
	// update rects in ways that wasn't intended.
	Rect updateRect = {0, GAME_WIDTH, 0, GAME_HEIGHT};
	StateManager::getInstance().addUpdateRect(updateRect);
}

void RaceState::placeStats()
{
	if (raceSetup.isTimeTrial())
	{
		assert(cars.size() == 1);
		
		cars[0].lapRO.moveTo(26, 228);
		updateLapRO(0);
	}
	else
	{
		for (std::size_t i = 0; i < cars.size(); ++i)
		{
			Car& car = cars[i];
			car.posRO.moveTo(17 + i * 31, 226);
			car.lapRO.moveTo(17 + i * 31, 233);
			addRenderObject(car.posRO);
			
			SurfaceBank::getCarStatBackground(raceSetup.getPlayerCar(i)).draw(background, i * 31, 224);
			updateLapRO(i);
			Surface("pos: /"+int2string(cars.size()), SMALL_FONT, STAT_COLOR).draw(background, 3 + i * 31, 226);
		}
	}
	
	Surface timeColonSurface(":", BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR);
	timeRO[0].moveTo(310, 228);
	timeRO[1].moveTo(304, 228);
	timeColonSurface.draw(background, 302, 228);
	timeRO[2].moveTo(296, 228);
	timeRO[3].moveTo(290, 228);
	timeColonSurface.draw(background, 288, 228);
	timeRO[4].moveTo(282, 228);
	timeRO[5].moveTo(276, 228);
}

Track* RaceState::getTrack()
{
	return &track;
}

void RaceState::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	if (menu.isVisible())
	{
		if (net && !net->onEvent(event))
		{
			menu.onEvent(event);
			if (closeMenu)
			{
				menu.hide();
				sm.forgetLastMenuEvent();
				closeMenu = false;
				updateTimeState();
				net->enableChatInput(false);
				return;
			}
			else if (quit)
			{
				net->quitGame();
				sm.enter(new NetworkMenu(net), -1);
				return;
			}
		}
		else if (!net)
		{
			menu.onEvent(event);
			if (closeMenu)
			{
				menu.hide();
				sm.forgetLastMenuEvent();
				closeMenu = false;
				updateTimeState();
			}
			else if (quit)
			{
				sm.enter(new MainMenu());
				return;
			}
			else if (restartRace)
			{
				sm.enter(new RaceState(map, raceSetup), 0);
				return;
			}
			else if (newTrack)
			{
				sm.enter(new SelectNewTrack(), 0);
				return;
			}
			else if (selectTrack)
			{
				sm.enter(new SelectTrack(), -1);
				return;
			}
		}
	}
	if (event.type == SDL_KEYDOWN)
	{
		switch (event.key.keysym.sym)
		{
		case SDLK_ESCAPE:
			if (menu.isVisible())
			{
				menu.hide();
				sm.forgetLastMenuEvent();
			}
			else
			{
				menu.show(this);
			}
			updateTimeState();
			if (net)
			{
				net->enableChatInput(menu.isVisible());
			}
			break;
		case SDLK_TAB:
			if (net && !menu.isVisible())
			{
				// Allow toggle chat while racing
				net->onEvent(event);
				net->enableChatInput(false);
			}
		default:
			break;
		}
	}
}
