/*
 * Copyright 2011, 2013, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ResultState.h"
#include "SurfaceBank.h"
#include "StateManager.h"
#include "Network.h"
#include "RaceState.h"
#include "NetworkMenu.h"
#include "MainMenu.h"
#include "str.h"
#include <cmath>

ResultState::ResultState(const RaceSetup& rs, Network* net, bool showPoints)
:	raceSetup(rs),
	nextAction(false),
	quitAction(false),
	pointsAction(false),
	haveWinner(showPoints && rs.getCurrentTrack() == rs.getTracks()),
	winningCarRotation(0),
	winnerTextTimer(0),
	net(net),
	showPoints(showPoints)
{
	if (showPoints)
	{		
		menu.setTitle(std::string(haveWinner ? "final " : "") + "Tournament table");
		raceSetup.sortPlayersByPoints();
	}
	else
	{
		menu.setTitle("Race Results" + (raceSetup.isTournament() ? " (" + int2string(raceSetup.getCurrentTrack()) + "/" + int2string(raceSetup.getTracks()) + ")" : ""));
		raceSetup.sortPlayersByTime();
	}
	
	int y = 60;
	int ySpacing;
	switch (raceSetup.numberOfPlayers())
	{
	case 8:
		ySpacing = 15;
		break;
	case 7:
		ySpacing = 17;
		break;
	default:
		ySpacing = 20;
	}
	for (int i = 0; i < raceSetup.numberOfPlayers(); ++i)
	{
		if (!haveWinner || i > 0)
		{
			Surface carSurface = SurfaceBank::getCar(raceSetup.getPlayerCar(i));
			carSurface.draw(background, 92, y, PI_HALF);
		}
		Surface(raceSetup.getPlayerName(i), BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR).draw(background, 102, y);
		Surface valueSurface(showPoints ? int2string(raceSetup.getPlayerPoints(i)) : time2string(raceSetup.getPlayerTime(i)), BIG_FONT, TEXT_COLOR, TEXT_BORDER_COLOR);
		valueSurface.draw(background, 207 - valueSurface.width() / 2, y);
		y += ySpacing;
	}
	
	if (!net || net->isOwner())
	{
		if (raceSetup.isTournament())
		{
			if (showPoints)
			{
				if (raceSetup.getCurrentTrack() == raceSetup.getTracks())
				{
					menu.addItem("New tournament", &nextAction);
				}
				else
				{
					menu.addItem("Next Race", &nextAction);
				}
			}
			else
			{
				menu.addItem("Show points", &pointsAction);
			}
		}
		else
		{
			menu.addItem("New Race", &nextAction);
		}
	}
	
	
	if (net)
	{
		if (net->isOwner())
		{
			menu.addItem("End network game", &quitAction);
		}
		else
		{
			menu.addItem("Hide menu", &nextAction);
			menu.addItem("Leave network game", &quitAction);
		}
	}
	else
	{
		menu.addItem("Main menu", &quitAction);
	}
	
	if (haveWinner)
	{
		const std::string winnerText = "Winner";
		Surface leftWinnerSurface(winnerText + " >", SMALL_FONT, WINNER_COLOR, TEXT_BORDER_COLOR);
		Surface rightWinnerSurface("< " + winnerText, SMALL_FONT, WINNER_COLOR, TEXT_BORDER_COLOR);
		leftWinnerTextRO.setSurface(leftWinnerSurface);
		rightWinnerTextRO.setSurface(rightWinnerSurface);
		winningCarRO.setSurface(SurfaceBank::getCar(raceSetup.getPlayerCar(0)));
		
		leftWinnerTextRO.moveTo(28, 61);
		rightWinnerTextRO.moveTo(257, 61);
		winningCarRO.moveTo(92, 60);
		
		addRenderObject(leftWinnerTextRO);
		addRenderObject(rightWinnerTextRO);
		
		addRenderObject(winningCarRO);
	}
	
	menu.addRenderObjects(this, CENTER_X, 200);
	
	if (net)
	{
		if (!net->isOwner())
		{
			menu.hide();
		}
		net->setGameState(this);
	}
}

void ResultState::update()
{
	if (net)
	{
		GameState* newState = net->getNewGameState();
		if (newState)
		{
			StateManager::getInstance().enter(newState, 0);
			return;
		}
		net->step();
	}
	menu.update();
	if (haveWinner)
	{
		winningCarRO.setRotation(5 * std::sin(winningCarRotation));
		winningCarRotation += 0.03;
		
		leftWinnerTextRO.moveTo(28 + 15 * std::sin(3*winningCarRotation), 61);
		rightWinnerTextRO.moveTo(257 + 15 * std::sin(-3*winningCarRotation), 61);
		
		++winnerTextTimer;
	}
}

void ResultState::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	
	if (net && net->onEvent(event))
	{
		return;
	}
	
	MenuEvent me = sm.getMenuEvent(event);
	if (net && !net->isOwner() && me.type == MENU_ESC)
	{
		if (menu.isVisible())
		{
			menu.hide();
		}
		else
		{
			menu.show(this);
		}
		return;
	}
	
	menu.onEvent(me);
	if (nextAction)
	{
		if (net)
		{
			if (net->isOwner())
			{
				net->startRace();
			}
			else
			{
				menu.hide();
				nextAction = false;
			}
		}
		else
		{
			// make so the cars start next track in reversed posion order
			raceSetup.sortPlayersByTime();
			raceSetup.reversePlayerOrder();
			
			sm.enter(new RaceState(Map::generate(), raceSetup), 0);
		}
	}
	else if (quitAction)
	{
		if (net)
		{
			net->quitGame();
			sm.enter(new NetworkMenu(net), -1);
		}
		else
		{
			sm.enter(new MainMenu());
		}
	}
	else if (pointsAction)
	{
		if (net)
		{
			net->onShowPoints(raceSetup);
		}
		else
		{
			sm.enter(new ResultState(raceSetup, net, true), 0);
		}
	}
}
