/*
 * Copyright 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Scaler.h"
#include "Settings.h"
#include "DColor.h"
#include "SDL.h"
#include <cmath>
#include <cassert>

Scaler::Scaler()
:	buffer(0)
{
}

Scaler::~Scaler()
{
	SDL_FreeSurface(buffer);
}

void Scaler::update(int w, int h)
{ 
	zoom = std::min(w / GAME_WIDTH, h / GAME_HEIGHT);
	while (!isValidZoom(zoom))
	{
		--zoom;
		assert(zoom > 0);
	}
	xZoom = w / GAME_WIDTH;
	yZoom = h / GAME_HEIGHT;
	
	stretch = Settings::graphics.stretch;
	
	if (stretch)
	{
		screenArea.w = w;
		screenArea.h = h;
		if (Settings::graphics.keepAspectRatio)
		{
			int w_scale = w * GAME_HEIGHT;
			int h_scale = h * GAME_WIDTH;
			if (w_scale > h_scale)
			{
				screenArea.w = h_scale / GAME_HEIGHT;
				xZoom = yZoom;
			}
			else
			{
				screenArea.h = w_scale / GAME_WIDTH;
				yZoom = xZoom;
			}
		}
		
		if (screenArea.w == zoom * GAME_WIDTH &&
		    screenArea.h == zoom * GAME_HEIGHT)
		{
			// Stretching is not needed so it can be disabled here. 
			// This makes the perfomance somewhat inconsistent for
			// different sizes.
			stretch = false;
		}
	}
	else
	{
		screenArea.w = GAME_WIDTH * zoom;
		screenArea.h = GAME_HEIGHT * zoom;
	}
	screenArea.x = (w - screenArea.w) / 2;
	screenArea.y = (h - screenArea.h) / 2;
	
	if (stretch && needStretchBuffer())
	{
		SDL_PixelFormat* fmt = SDL_GetVideoSurface()->format;
		buffer = SDL_CreateRGBSurface(SDL_SWSURFACE, zoom * GAME_WIDTH, zoom * GAME_HEIGHT, fmt->BitsPerPixel, fmt->Rmask, fmt->Gmask, fmt->Bmask, fmt->Amask);
	}
	else
	{
		SDL_FreeSurface(buffer);
		buffer = 0;
	}
}

bool Scaler::isValidZoom(int zoom) const
{
	return true;
}

bool Scaler::isValidBpp(int bpp) const
{
	return bpp == 24 || bpp == 32;
}

bool Scaler::needStretchBuffer() const
{
	return true;
}

void Scaler::render(const Surface& surface, SDL_Rect* area)
{
	SDL_Surface* dest = SDL_GetVideoSurface();
	SDL_Surface* src = surface.surface;
	
	assert(src != 0);
	assert(dest != 0);
	assert(src->format->BytesPerPixel == dest->format->BytesPerPixel);
	
	SDL_Rect srcArea;
	SDL_Rect dstArea;
	if (!area)
	{
		srcArea.w = src->w;
		srcArea.h = src->h;
		srcArea.x = 0;
		srcArea.y = 0;
		area = &dstArea;
	} 
	else
	{
		srcArea.w = area->w;
		srcArea.h = area->h;
		srcArea.x = area->x;
		srcArea.y = area->y;
	}

	if (stretch)
	{
		renderStretch(srcArea, *area, src, dest);
	}
	else
	{
		renderZoom(srcArea, *area, src, dest);
	}
}

template <typename S>
void stretchBilinear(const SDL_Rect& dstArea, const SDL_Rect& screenArea, SDL_Surface* source, SDL_Surface* destination)
{
	typedef typename S::PixelType PixelType;
	
	int xStart = dstArea.x;
	int yStart = dstArea.y;
	int xEnd = dstArea.x + dstArea.w;
	int yEnd = dstArea.y + dstArea.h;
	
	PixelType* src = (PixelType*) source->pixels;
	
	std::ptrdiff_t src_rowskip = source->pitch / sizeof(PixelType);
	std::ptrdiff_t dst_rowskip = destination->pitch / sizeof(PixelType);
	
	double srcStepX = static_cast<double>(source->w) / screenArea.w;
	double srcStepY = static_cast<double>(source->h) / screenArea.h;
	
	double srcY = (yStart - screenArea.y + 0.5) * srcStepY - 0.5;
	
	for (int y = yStart; y < yEnd; ++y, srcY += srcStepY)
	{
		int y1 = std::max(0, static_cast<int>(srcY));
		int y2 = std::min(source->h - 1, static_cast<int>(srcY + 1.0));
		
		double yDistance1 = srcY - std::floor(srcY);
		double yDistance2 = 1.0 - yDistance1;
		
		double srcX = (xStart - screenArea.x + 0.5) * srcStepX - 0.5;
		
		for (int x = xStart; x < xEnd; ++x, srcX += srcStepX)
		{			
			int x1 = std::max(0, static_cast<int>(srcX));
			int x2 = std::min(source->w - 1, static_cast<int>(srcX + 1.0));
			
			PixelType* p1 = src + S::pixelSize * x1 + y1 * src_rowskip;
			PixelType* p2 = src + S::pixelSize * x2 + y1 * src_rowskip;
			PixelType* p3 = src + S::pixelSize * x1 + y2 * src_rowskip;
			PixelType* p4 = src + S::pixelSize * x2 + y2 * src_rowskip;
			
			DColor c1;
			DColor c2;
			DColor c3;
			DColor c4;
			
			S::readColors(p1, c1);
			S::readColors(p2, c2);
			S::readColors(p3, c3);
			S::readColors(p4, c4);
			
			double xDistance1 = srcX - std::floor(srcX);
			double xDistance2 = 1.0 - xDistance1;
			
			double ry1 = c1.r * yDistance2 + c3.r * yDistance1;
			double ry2 = c2.r * yDistance2 + c4.r * yDistance1;
			double gy1 = c1.g * yDistance2 + c3.g * yDistance1;
			double gy2 = c2.g * yDistance2 + c4.g * yDistance1;
			double by1 = c1.b * yDistance2 + c3.b * yDistance1;
			double by2 = c2.b * yDistance2 + c4.b * yDistance1;
			
			DColor c;
			c.r = ry1 * xDistance2 + ry2 * xDistance1;
			c.g = gy1 * xDistance2 + gy2 * xDistance1;
			c.b = by1 * xDistance2 + by2 * xDistance1;
			
			PixelType* dst = (PixelType*) destination->pixels + S::pixelSize * x + y * dst_rowskip;
			S::writeColors(dst, c);
		}
	}
}

void Scaler::stretchArea(const SDL_Rect& srcArea, SDL_Rect& dstArea, int x, int y, SDL_Surface* source, SDL_Surface* destination)
{
	assert(source);
	assert(destination);
	
	int bpp = destination->format->BytesPerPixel;

	int dstPaddingX = (screenArea.w / source->w) + ((screenArea.w % source->w) != 0 ? 1 : 0);
	int dstPaddingY = (screenArea.h / source->h) + ((screenArea.h % source->h) != 0 ? 1 : 0);
	
	dstArea.x -= dstPaddingX;
	dstArea.w += 2 * dstPaddingX;
	if (dstArea.x < screenArea.x)
	{
		dstArea.w -= screenArea.x - dstArea.x;
		dstArea.x = screenArea.x;
	}
	Uint16 mx = screenArea.x + screenArea.w - dstArea.x;
	if (screenArea.x + dstArea.w > mx)
	{
		dstArea.w = mx;
	}
	
	dstArea.y -= dstPaddingY;
	dstArea.h += 2 * dstPaddingY;
	if (dstArea.y < screenArea.y)
	{
		dstArea.h -= screenArea.y - dstArea.y;
		dstArea.y = screenArea.y;
	}
	Uint16 my = screenArea.y + screenArea.h - dstArea.y;
	if (screenArea.y + dstArea.h > my)
	{
		dstArea.h = my;
	}
	
	switch (bpp)
	{
	case 3:
		stretchBilinear<Scale24>(dstArea, screenArea, buffer, destination);
		break;
	case 4:
		stretchBilinear<Scale32>(dstArea, screenArea, buffer, destination);
		break;
	}
}

void Scaler::expand(SDL_Rect& srcArea, SDL_Rect& dstArea, int pixels)
{	
	srcArea.x -= pixels;
	srcArea.w += 2 * pixels;
	if (srcArea.x < 0)
	{
		srcArea.w += srcArea.x;
		srcArea.x = 0;
	}
	Uint16 mx = GAME_WIDTH - srcArea.x;
	if (srcArea.w > mx)
	{
		srcArea.w = mx;
	}
	
	srcArea.y -= pixels;
	srcArea.h += 2 * pixels;
	if (srcArea.y < 0)
	{
		srcArea.h += srcArea.y;
		srcArea.y = 0;
	}
	Uint16 my = GAME_HEIGHT - srcArea.y;
	if (srcArea.h > my)
	{
		srcArea.h = my;
	}
	
	dstArea.x = src2dst_x(srcArea.x);
	dstArea.y = src2dst_y(srcArea.y);
	dstArea.w = src2dst_x(srcArea.x + srcArea.w) - dstArea.x;
	dstArea.h = src2dst_y(srcArea.y + srcArea.h) - dstArea.y;
}
