/*
 * Copyright 2010, 2011, 2012, 2013, 2014, 2016 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "StateManager.h"
#include "GameState.h"
#include "RenderObject.h"
#include "MainMenu.h"
#include "RacePreparationMenu.h"
#include "CarSelectionState.h"
#include "RaceState.h"
#include "ResultState.h"
#include "SettingsMenu.h"
#include "Map.h"
#include "Path.h"
#include "Driver.h"
#include "KeyboardDriver.h"
#include "KeyboardDevice.h"
#include "MouseDevice.h"
#include "JoystickDevice.h"
#include "JoystickDriver.h"
#include "NearestNeighbor.h"
#include "Bilinear.h"
#include "BilinearLessBlur.h"
#include "Hqx.h"
#include "Network.h"
#include "background.h"
#include <fstream>
#include <cassert>
#include <cstdlib>

const int GAME_WIDTH = 320;
const int GAME_HEIGHT = 240;

StateManager::StateManager()
:	currentState(0),
	stateChanged(false),
	screenZoom(1),
	screenPaddingX(0),
	screenPaddingY(0),
	fullscreen(0),
	xAxis(0),
	yAxis(0),
	bpp(0),
	network(),
	lastMenuEvent()
{
	// Load settings.
	Path settingsPath(configPath() + "/settings");
	std::ifstream in(settingsPath.c_str());
	Settings::load(in);
	
	fullscreen = Settings::graphics.fullscreen;
	
	switch (Settings::graphics.scaler)
	{
	case 0:
	default:
		scaler = new NearestNeighbor;
		break;
	case 1:
		scaler = new Bilinear;
		break;
	case 2:
		scaler = new Hqx;
		break;
	case 3:
		scaler = new BilinearLessBlur;
		break;
	}
	
	// Make sure bit depth is supported by the scaler.
	const SDL_VideoInfo* videoInfo = SDL_GetVideoInfo();
	bpp = videoInfo->vfmt->BitsPerPixel;
	if (!scaler->isValidBpp(bpp))
	{
		bpp = 32;
	}
	
	// set the fullscreen dimensions
	if (Settings::graphics.fullscreenWidth == 0 || Settings::graphics.fullscreenHeight == 0)
	{
		screenWidth[1]  = videoInfo->current_w;
		screenHeight[1] = videoInfo->current_h;
	}
	else
	{
		screenWidth[1]  = std::max(Settings::graphics.fullscreenWidth,  GAME_WIDTH);
		screenHeight[1] = std::max(Settings::graphics.fullscreenHeight, GAME_HEIGHT);
	}

	// set the window dimensions
	int maxZoom = std::min(videoInfo->current_w / GAME_WIDTH, videoInfo->current_h / GAME_HEIGHT);
	screenZoom = std::max(maxZoom - 1, 1);
	screenWidth[0] = GAME_WIDTH * screenZoom;
	screenHeight[0] = GAME_HEIGHT * screenZoom;

	updateVideoMode();
	
	screenBuffer = Surface(GAME_WIDTH, GAME_HEIGHT);
	background = createBackground(),
	
	// Read input devices
	devices.push_back(new KeyboardDevice());
	devices.push_back(new MouseDevice());
	int nJoysticks = SDL_NumJoysticks();
	for (int i = 0; i < nJoysticks; ++i)
	{
		devices.push_back(new JoystickDevice(i));
	}

	// load human drivers
	drivers.resize(8);
	if (in)
	{
		std::string deviceName;
		for (std::size_t i = 0; i < 8 && std::getline(in, deviceName); ++i)
		{
			for (std::size_t j = 0; j < devices.size(); ++j)
			{
				if (devices[j]->getName() == deviceName)
				{
					Driver* driver = devices[j]->createDriver();
					driver->load(in);
					driver->setDeviceIndex(j);
					drivers[i] = driver;
					break;
				}
			}
			if (!drivers[i])
			{
				// Ignore key line.
				in.ignore(1000, '\n');
			}
		}
	}

	for (std::size_t i = 0; i < 8; ++i)
	{
		if (!drivers[i])
		{
			Driver* driver = 0;
			switch (i)
			{
			case 0:
				driver = new KeyboardDriver(SDLK_UP, SDLK_DOWN, SDLK_LEFT, SDLK_RIGHT);
				break;
			case 1:
				driver = new KeyboardDriver(SDLK_w, SDLK_s, SDLK_a, SDLK_d);
				break;
			case 2:
				driver = new KeyboardDriver(SDLK_i, SDLK_k, SDLK_j, SDLK_l);
				break;
			case 3:
				driver = new KeyboardDriver(SDLK_t, SDLK_g, SDLK_f, SDLK_h);
				break;
			case 4:
				driver = new KeyboardDriver(SDLK_KP8, SDLK_KP5, SDLK_KP4, SDLK_KP6);
				break;
			case 5:
				driver = new KeyboardDriver(SDLK_HOME, SDLK_END, SDLK_DELETE, SDLK_PAGEDOWN);
				break;
			case 6:
				driver = new KeyboardDriver(SDLK_F4, SDLK_F3, SDLK_F1, SDLK_F2);
				break;
			case 7:
				driver = new KeyboardDriver(SDLK_F10, SDLK_F9, SDLK_F7, SDLK_F8);
				break;
			}
			driver->setDeviceIndex(0);
			drivers[i] = driver;
		}
	}
	
	forceHideFullscreenPointer();
}

void StateManager::resize(int width, int height)
{
	screenWidth[fullscreen] = std::max(GAME_WIDTH, width);
	screenHeight[fullscreen] = std::max(GAME_HEIGHT, height);
	scaler->update(screenWidth[fullscreen], screenHeight[fullscreen]);
	updateVideoMode();
}

void StateManager::updateVideoMode()
{
	scaler->update(screenWidth[fullscreen], screenHeight[fullscreen]);

	// update SDL window
	Uint32 flags = SDL_SWSURFACE|(fullscreen?SDL_FULLSCREEN|SDL_NOFRAME:SDL_RESIZABLE);
	SDL_Surface* screen = SDL_SetVideoMode(screenWidth[fullscreen], screenHeight[fullscreen], bpp, flags);
	if (!screen)
	{
		if (fullscreen)
		{
			// This could happen if the user has a weird fullscreen
			// resolution so switching to window mode might help.
			fullscreen = false;
			updateVideoMode();
			return;
		}
		else
		{
			// I have no idea why this happened so we better quit now.
			std::exit(EXIT_FAILURE);
		}	
	}
	
	pointer.setFullScreen(fullscreen);
	pointer.setScaling(Settings::graphics.scalePointer > (fullscreen ? 0 : 1));
	pointer.setType(Settings::graphics.pointer);
	
	// updateWholeScreen() will have to be called afterwards if needed
}

void StateManager::updateScreen()
{
	// draw background
	assert(currentState);
	currentState->background.draw(screenBuffer,0,0);
	
	pointer.update(!currentState->menu.isVisible());
	
	std::vector<RenderObject*>& renderObjects = currentState->renderObjects;
	std::vector<RenderObject*>::iterator it;

	Rect rect1;
	Rect rect2;
	for (it = renderObjects.begin(); it != renderObjects.end(); ++it)
	{
		RenderObject* ro = *it;
		ro->draw(screenBuffer);
		if (ro->isUpdated())
		{
			ro->updateRects(rect1, rect2);
			addUpdateRect(rect1);
			addUpdateRect(rect2);
		}
	}
	
	pointer.draw(screenBuffer);
	if (pointer.isUpdated())
	{
		pointer.updateRects(rect1, rect2);
		addUpdateRect(rect1);
		addUpdateRect(rect2);
	}

	std::list<Rect>::iterator rit;
	for (rit = updateRects.begin(); rit != updateRects.end(); ++rit)
	{
		SDL_Rect rect;
		
		rect.x = (rit->x1);
		rect.y = (rit->y1);
		rect.w = (rit->x2 - rit->x1);
		rect.h = (rit->y2 - rit->y1);
		
		scaler->render(screenBuffer, &rect);
		
		updateSDLRects.push_back(rect);
	}

	if (!updateSDLRects.empty())
	{
		SDL_UpdateRects(SDL_GetVideoSurface(), updateRects.size(), &updateSDLRects[0]);
		updateSDLRects.clear();
		updateRects.clear();
	}
}

void StateManager::updateWholeScreen()
{
	// Fill screen with black
	SDL_FillRect(SDL_GetVideoSurface(), 0, 0);

	// draw background
	assert(currentState);
	currentState->background.draw(screenBuffer, 0, 0);
	
	// draw render objects
	std::vector<RenderObject*>& renderObjects = currentState->renderObjects;
	std::vector<RenderObject*>::iterator it;
	Rect rect1;
	Rect rect2;
	for (it = renderObjects.begin(); it != renderObjects.end(); ++it)
	{
		RenderObject* ro = *it;
		ro->draw(screenBuffer);
		if (ro->isUpdated())
		{
			// we need to call updateRects even if we don't use the 
			// rects to update them.
			ro->updateRects(rect1, rect2);
		}
	}
	
	pointer.draw(screenBuffer);
	if (pointer.isUpdated())
	{
		pointer.updateRects(rect1, rect2);
	}
	
	scaler->render(screenBuffer, 0);
	
	SDL_Flip(SDL_GetVideoSurface());
}

void StateManager::addUpdateRect(Rect& rect)
{
	if (rect.x1 < 0)
	{
		rect.x1 = 0;
	}
	if (rect.y1 < 0)
	{
		rect.y1 = 0;
	}
	if (rect.x2 > GAME_WIDTH)
	{
		rect.x2 = GAME_WIDTH;
	}
	if (rect.y2 > GAME_HEIGHT)
	{
		rect.y2 = GAME_HEIGHT;
	}
	
	if (rect.x1 >= rect.x2 || rect.y1 >= rect.y2)
	{
		return;
	}
	
	std::list<Rect>::iterator it;
	for (it = updateRects.begin(); it != updateRects.end(); ++it)
	{
		if (overlap(rect, *it))
		{
			rect.combine(*it);

			updateRects.erase(it);
			
			// start loop from start
			// because the new rect can be much larger
			it = updateRects.begin();
		}
	}
	updateRects.push_front(rect);
}

Driver* StateManager::getDriver(int index)
{
	assert(static_cast<unsigned>(index) < drivers.size());
	return drivers[index];
}

InputDevice* StateManager::getDevice(int index)
{
	assert(static_cast<unsigned>(index) < devices.size());
	return devices[index];
}

std::vector<std::string> StateManager::getDeviceNames() const
{
	std::vector<std::string> names;
	for (std::size_t i = 0; i < devices.size(); ++i)
	{
		names.push_back(devices[i]->getName());
	}
	return names;
}

void StateManager::setDriver(int i, Driver* driver)
{
	delete drivers[i];
	drivers[i] = driver;
}

void StateManager::cleanUp()
{	
	// Save settings.
	Path settingsPath(configPath() + "/settings");
	std::ofstream out(settingsPath.c_str());
	Settings::save(out);
	
	// Save (and delete) human drivers.
	for (std::size_t i = 0; i < drivers.size(); ++i)
	{
		drivers[i]->save(out);
		delete drivers[i];
	}
	
	for (std::size_t i = 0; i < devices.size(); ++i)
	{
		delete devices[i];
	}
	
	delete network;
}

StateManager& StateManager::getInstance()
{
	static StateManager instance;
	return instance;
}

void StateManager::run()
{
	SDL_Event event;
	while (currentState)
	{
		stateChanged = false;

		// handle events
		while (SDL_PollEvent(&event))
		{
			handleEvent(event);
			if (stateChanged)
			{
				break;
			}
		}
		if (stateChanged)
		{
			// this one is needed because the other check inside the event
			// loop doesn't break out of the main loop
			continue;
		}
		
		// update state
		currentState->update();
		if (stateChanged)
		{
			continue;
		}
		// render
		updateScreen();
		
		// sleep
		int st = currentState->sleepTime();
		if (st > 0)
		{
			if (network)
			{
				network->sleep(st);
			}
			else
			{
				SDL_Delay(st);
			}
		}
	}
}

void StateManager::handleEvent(const SDL_Event& event)
{	
	switch(event.type)
	{
	case SDL_KEYDOWN:
		if (SDL_GetModState() & KMOD_ALT)
		{
			switch (event.key.keysym.sym)
			{
			case SDLK_F4:
				quit();
				return;
			case SDLK_RETURN:
				toggleFullscreen();
				return;
			default:
				;
			}
		}
		switch (event.key.keysym.sym)
		{
		case SDLK_F11:
			toggleFullscreen();
			return;
		default:
			;
		}
		break;
	case SDL_QUIT:
		quit();
		return;
	case SDL_VIDEORESIZE: 
		resize(event.resize.w, event.resize.h);
		updateWholeScreen();
		return;
	case SDL_VIDEOEXPOSE: 
		updateWholeScreen();
		return;
	case SDL_MOUSEMOTION:
		pointer.move(scaler->dst2src_x(event.motion.x), scaler->dst2src_y(event.motion.y));
		break;
	case SDL_ACTIVEEVENT:
		if (event.active.state & SDL_APPMOUSEFOCUS)
		{
			if (event.active.gain)
			{
				pointer.show();
			}
			else
			{
				pointer.hide(true);
			}
		}
		break;
	default:
		;
	}
	
	currentState->onEvent(event);
}

void StateManager::quit()
{
	delete currentState;
	currentState = 0;
	stateChanged = true;
}

void StateManager::enter(GameState* newState)
{
	newState->setStateInfo(oldStateInfo.empty() ? 0 : oldStateInfo[0]);
	oldStateInfo.clear();
	delete currentState;
	currentState = newState;
	
	stateChanged = true;
	if (currentState)
	{
		updateWholeScreen();
	}
}

void StateManager::enter(GameState* newState, int x)
{
	int stateInfo = 0;
	if (x > 0)
	{
		assert(x == 1);
		if (currentState)
		{
			oldStateInfo.push_back(currentState->getStateInfo());
		}
	}
	else if (x < 0)
	{
		if (newState)
		{
			std::size_t i = oldStateInfo.size() + x;
			assert(i < oldStateInfo.size());
			stateInfo = oldStateInfo[i];
			oldStateInfo.resize(i);
		}
	}
	newState->setStateInfo(stateInfo);

	if  (lastMenuEvent.type != MENU_MOUSE_CLICK1)
	{
		pointer.hide();
	}
	
	delete currentState;
	currentState = newState;
	
	stateChanged = true;
	if (currentState)
	{
		updateWholeScreen();
	}
}

// When entering a new state we sometimes want to use the old content of
// the screen as a background. This is useful for ingame menus. Call this
// method before creating the state object to make use of this.
void StateManager::screenToBackgound()
{
	assert(currentState);
	screenBuffer.draw(currentState->background, 0, 0);
}

void StateManager::toggleFullscreen()
{
	fullscreen = !fullscreen;
	updateVideoMode();
	updateWholeScreen();
	
	forceHideFullscreenPointer();
}

void StateManager::forceHideFullscreenPointer()
{
	if (fullscreen)
	{
		// It doesn't look nice getting that big pointer in the middle 
		// of the screen so lets just hide the pointer when switching to
		// fullscreen mode.
		SDL_Event e;
		while (SDL_PollEvent(&e));
		pointer.hide();
	}
}

MenuEvent StateManager::getMenuEvent(const SDL_Event& event)
{
	MenuEvent e = MenuEvent(); // Zero initialized.
	switch (event.type)
	{
	case SDL_KEYDOWN:
		switch (event.key.keysym.sym)
		{
		case SDLK_DOWN:
			e.type = MENU_NEXT;
			break;
		case SDLK_UP:
			e.type = MENU_PREV;
			break;
		case SDLK_LEFT:
			e.type = MENU_DEC;
			break;
		case SDLK_RIGHT:
			e.type = MENU_INC;
			break;
		case SDLK_RETURN:
		case SDLK_KP_ENTER:
			e.type = MENU_ACTION;
			break;
		case SDLK_ESCAPE:
			e.type = MENU_ESC;
			break;
		case SDLK_HOME:
			e.type = MENU_MIN;
			break;
		case SDLK_END:
			e.type = MENU_MAX;
			break;
		case SDLK_PAGEUP:
			e.type = MENU_FIRST;
			break;
		case SDLK_PAGEDOWN:
			e.type = MENU_LAST;
			break;
		case SDLK_BACKSPACE:
			e.type = MENU_ERASE;
			break;
		case SDLK_DELETE:
			e.type = MENU_DELETE;
			break;
		default:
			e.c = unicode2char(event.key.keysym.unicode);
			if (std::isprint(e.c))
			{
				e.type = MENU_TEXT;
			}
		}
		break;
	case SDL_JOYAXISMOTION:
	{
		int newAxis = 0;
		if (event.jaxis.value > AXIS_THRESHOLD)
		{
			newAxis = 1;
		}
		else if (std::abs(event.jaxis.value) < AXIS_THRESHOLD / 2)
		{
			newAxis = 0;
		}
		else if (event.jaxis.value < -AXIS_THRESHOLD)
		{
			newAxis = -1;
		}
		
		if (event.jaxis.axis & 1)
		{
			if (newAxis != yAxis)
			{
				if (newAxis == 1)
				{
					e.type = MENU_NEXT;
				}
				else if (newAxis == -1)
				{
					e.type = MENU_PREV;
				}
				yAxis = newAxis;
			}
		}
		else
		{
			if (newAxis != xAxis)
			{
				if (newAxis == 1)
				{
					e.type = MENU_INC;
				}
				else if (newAxis == -1)
				{
					e.type = MENU_DEC;
				}
				xAxis = newAxis;
			}
		}
		break;
	}
	case SDL_JOYHATMOTION:
		switch (event.jhat.value)
		{
		case SDL_HAT_DOWN:
			e.type = MENU_NEXT;
			break;
		case SDL_HAT_UP:
			e.type = MENU_PREV;
			break;
		case SDL_HAT_LEFT:
			e.type = MENU_DEC;
			break;
		case SDL_HAT_RIGHT:
			e.type = MENU_INC;
			break;
		}
		break;
	case SDL_JOYBUTTONDOWN:
		switch (event.jbutton.button)
		{
		case 0:
			e.type = MENU_ACTION;
			break;
		case 1:
			e.type = MENU_ESC;
			break;
		}
		break;
	case SDL_MOUSEMOTION: 
		e.type = MENU_MOUSE_MOVE;
		e.x = scaler->dst2src_x(event.motion.x);
		e.y = scaler->dst2src_y(event.motion.y);
		break;
	case SDL_MOUSEBUTTONDOWN:
		if (!pointer.isVisible())
		{
			return e;
		}
		switch (event.button.button)
		{
		case SDL_BUTTON_LEFT:
			e.type = MENU_MOUSE_CLICK1;
			break;
		case SDL_BUTTON_RIGHT:
			e.type = MENU_MOUSE_CLICK2;
			break;
		case SDL_BUTTON_WHEELUP:
			e.type = MENU_INC;
			break;
		case SDL_BUTTON_WHEELDOWN:
			e.type = MENU_DEC;
			break;
		default:
			e.type = MENU_ESC;
		}
		e.x = scaler->dst2src_x(event.motion.x);
		e.y = scaler->dst2src_y(event.motion.y);
		break;
	case SDL_MOUSEBUTTONUP:
		e.type = MENU_MOUSE_RELEASE;
		break;
	}
	lastMenuEvent = e;
	return e;
}

void StateManager::forgetLastMenuEvent()
{
	lastMenuEvent.type = MENU_UNKNOWN;
}

void StateManager::updateGraphicsSettings(const Settings::Graphics& gfx)
{
	// This function is not very efficient because it calls 
	// updateVideoMode() and updateWholeScreen() multiple times.
	
	updateScaler(gfx.scaler);
	updateStretch(gfx.stretch, gfx.keepAspectRatio);
	updatePointerType(gfx.pointer);
	updatePointerScaling(gfx.scalePointer);
	
	// These settings only have an effect when the game is started.
	Settings::graphics.fullscreen = gfx.fullscreen;
	Settings::graphics.fullscreenWidth = gfx.fullscreenWidth;
	Settings::graphics.fullscreenHeight = gfx.fullscreenHeight;
}

void StateManager::updateScaler(int newScaler)
{
	if (newScaler != Settings::graphics.scaler)
	{
		delete scaler;
		switch (newScaler)
		{
		case 0:
		default:
			scaler = new NearestNeighbor;
			break;
		case 1:
			scaler = new Bilinear;
			break;
		case 2:
			scaler = new Hqx;
			break;
		case 3:
			scaler = new BilinearLessBlur;
			break;
		}
		
		const SDL_VideoInfo* videoInfo = SDL_GetVideoInfo();
		bpp = videoInfo->vfmt->BitsPerPixel;
		if (!scaler->isValidBpp(bpp))
		{
			bpp = 32;
		}
		updateVideoMode();
		
		// Make sure the buffer has the correct bit depth.
		screenBuffer = Surface(GAME_WIDTH, GAME_HEIGHT);
		
		Settings::graphics.scaler = newScaler;
		updateWholeScreen();
	}
}

void StateManager::updateStretch(bool newStretch, bool newKeepAspectRatio)
{
	if (newStretch         != Settings::graphics.stretch || 
	    newKeepAspectRatio != Settings::graphics.keepAspectRatio)
	{
		Settings::graphics.stretch = newStretch;
		Settings::graphics.keepAspectRatio = newKeepAspectRatio;
		updateVideoMode();
		updateWholeScreen();
	}
}

void StateManager::updatePointerType(const std::string& pointerType)
{
	if (pointerType != Settings::graphics.pointer)
	{
		pointer.setType(pointerType);
		Settings::graphics.pointer = pointerType;
		
		// Make sure offset is updated correctly.
		int x;
		int y;
		SDL_GetMouseState(&x, &y);
		pointer.move(scaler->dst2src_x(x), scaler->dst2src_y(y));
		
		updateWholeScreen();
	}
}

void StateManager::updatePointerScaling(int scalePointer)
{
	if (scalePointer != Settings::graphics.scalePointer)
	{
		pointer.setScaling(scalePointer > (fullscreen ? 0 : 1));
		pointer.setType(Settings::graphics.pointer);
		Settings::graphics.scalePointer = scalePointer;
		updateWholeScreen();
	}
}


// The purpose of this function is to give the StateManager 
// responsibility to delete the network. 
void StateManager::setNetwork(Network* net)
{
	delete network;
	network = net;
}

Surface StateManager::getBackground() const
{
	return background;
}

const MenuEvent& StateManager::getLastMenuEvent() const
{
	return lastMenuEvent;
}

char unicode2char(Uint16 unicode)
{
	if (unicode < 128)
	{
		return std::tolower(static_cast<char>(unicode));
	}
	switch (unicode)
	{
	case 192:
	case 193:
	case 194:
	case 195:
	case 196:
	case 197:
	case 198:
	case 224:
	case 225:
	case 226:
	case 227:
	case 228:
	case 229:
	case 230:
		return 'a';
	case 199:
	case 231:
		return 'c';
	case 200:
	case 201:
	case 202:
	case 203:
	case 232:
	case 233:
	case 234:
	case 235:
		return 'e';
	case 204:
	case 205:
	case 206:
	case 207:
	case 236:
	case 237:
	case 238:
	case 239:
		return 'i';
	case 208:
	case 240:
		return 'd';
	case 209:
	case 241:
		return 'n';
	case 210:
	case 211:
	case 212:
	case 213:
	case 214:
	case 216:
	case 242:
	case 243:
	case 244:
	case 245:
	case 246:
	case 248:
		return 'o';
	case 217:
	case 218:
	case 219:
	case 220:
	case 249:
	case 250:
	case 251:
	case 252:
		return 'u';
	case 221:
	case 253:
	case 255:
		return 'y';
	}
	
	return '\0';
}
