/*
 * Copyright 2012, 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TileSet.h"
#include "Map.h"
#include "Path.h"
#include "Settings.h"
#include <cstdlib>
#include <algorithm>

const int MIN_TILESET = 9;

TileSet::TileSet(const std::string& name, const Surface& surface)
:	name(name)
{
	wallColor = surface.getPixel(0, 0);
	
	int nTiles = std::max(surface.height() / TILE_SIZE, MIN_TILESET);
	tiles.reserve(nTiles);
	
	for (int n = 0; n < nTiles; n++)
	{
		tiles.push_back(surface.subSurface(0, n * TILE_SIZE, TILE_SIZE, TILE_SIZE));
	}
}

void TileSet::decorate(const Map& map, Surface& background) const
{
	decorateUpper(map, background);
	decorateLower(map, background);
}

void TileSet::decorateUpper(const Map& map, Surface& background) const
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			const Surface& tile = pickTile(map, x, y);

			int px = x * TILE_SIZE;
			int py = y * TILE_SIZE;
			tile.draw(background, px, py);
			
			if (map[y][x] == 0)
			{
				if (y > 0 && map[y-1][x] != 0)
				{
					for (int i = 0; i < TILE_SIZE; i++)
					{
						background.setPixel(px + i, py, wallColor); 
					}
				}
				if (x < (MAX_MAP_X - 1) && map[y][x + 1] != 0)
				{
					for (int i = 0; i < TILE_SIZE; i++)
					{
						background.setPixel(px + TILE_SIZE - 1, py + i, wallColor); 
					}
				}
				
				if (x < (MAX_MAP_X - 1) && y > 0 && map[y - 1][x + 1] != 0)
				{
					background.setPixel(px + TILE_SIZE - 1, py, wallColor); 
				}
			}
		}
	}
}

void TileSet::decorateLower(const Map& map, Surface& background) const
{
	// This decoration is positioned below the track and not all of it 
	// will be visible. It has low priority so it doesn't have it's own 
	// decoration in Map, but instead we use the decoration sum of each 
	// column to decide the decoration.
	for (int x = 0; x < MAX_MAP_X; x++)
	{
		int colSum = 0;
		for (int y = 0; y < MAX_MAP_Y; y++)
		{
			colSum += map.getDecoration(x, y);
		}
		const Surface& tile = getTile(colSum);
		int px = x * TILE_SIZE;
		int py = MAX_MAP_Y * TILE_SIZE;
		tile.draw(background, px, py);
	}
}

const std::string& TileSet::getName() const
{
	return name;
}

int TileSet::getGfxTileLimit() const
{
	return tiles.size() - MIN_TILESET + 1;
}

const Surface& TileSet::pickTile(const Map& map, int x, int y) const
{
	int finishX = map.getFinishX();
	int finishY = map.getFinishY();
	if (x == finishX && y == finishY)
	{
		if (map[finishY][finishX] == ROAD_HORIZONTAL)
		{
			return tiles[6]; 
		}
		else
		{
			return tiles[7];
		}
	}
	switch (map[y][x])
	{
		case ROAD_HORIZONTAL:
			return tiles[0];
		case ROAD_VERTICAL:
			return tiles[1];
		case ROAD_LEFT_UP:
			return tiles[2];
		case ROAD_LEFT_DOWN:
			return tiles[3];
		case ROAD_RIGHT_UP:
			return tiles[4];
		case ROAD_RIGHT_DOWN:
			return tiles[5];
		default:
			return getTile(map.getDecoration(x, y));
	}
}

const Surface& TileSet::getTile(int nr) const
{
	return tiles[MIN_TILESET + nr % (tiles.size() - MIN_TILESET + 1) - 1];
}

bool isPNG(const std::string& filename)
{
	return filename.length() > 4 && filename.compare(filename.length() - 4, 4, ".png") == 0;
}

const std::vector<std::string>& TileSet::getTileSetNames()
{
	static std::vector<std::string> names;
	if (names.empty())
	{
		std::vector<std::string> files = (dataPath() + "tilesets").files();
		for (std::size_t i = 0; i < files.size(); ++i)
		{
			if (isPNG(files[i]))
			{
				names.push_back(files[i].substr(0, files[i].length() - 4));
			}
		}
		
		// Make sure not empty.
		if (names.empty())
		{
			names.push_back("blank");
		}
		
		std::sort(names.begin(), names.end());
	}
	return names;
}

const std::vector<TileSet>& TileSet::getTileSets()
{
	static std::vector<TileSet> tileSets;
	if (tileSets.empty()) // initialize
	{
		std::vector<std::string> names = getTileSetNames();
		
		tileSets.reserve(names.size());
		for (std::size_t i = 0; i < names.size(); ++i)
		{
			tileSets.push_back(TileSet(names[i], Surface("tilesets/" + names[i] + ".png")));
		}
	}
	return tileSets;
}

const TileSet& TileSet::getTileSet(const std::string& name)
{
	// Return tileset with matching name or the closest following 
	// tileset in alphabetic order. This gives some level of consistency
	// in case the tileset does not exist but unfortunately the 
	// distribution is not uniform.
	const std::vector<TileSet>& sets = getTileSets();
	std::vector<TileSet>::const_iterator it;
	for (it = sets.begin(); it != sets.end(); ++it)
	{
		if (it->getName() >= name)
		{
			break;
		}
	}
	if (it == sets.end())
	{
		it = sets.begin();
	}
	return *it;
}

const TileSet& TileSet::getRandomTileSet()
{
	const std::vector<TileSet>& sets = getTileSets();
	int percentSum = 0;
	std::vector<TileSet>::const_iterator it;
	for (it = sets.begin(); it != sets.end(); ++it)
	{
		percentSum += Settings::tracks.tilesets[it->getName()];
	}
	
	if (percentSum > 0)
	{
		// Make sure the random number is big enough to have good properties.
#if RAND_MAX < 2147483647
		int r = (std::rand() & 0x7FFF) | 15 << (std::rand() & 0xFFFF);
#else
		int r = std::rand();
#endif
		int num = r % percentSum;
		percentSum = 0;
		for (it = sets.begin(); it != sets.end(); ++it)
		{
			percentSum += Settings::tracks.tilesets[it->getName()];
			if (num < percentSum)
			{
				return *it;
			}
		}
	}
	return sets[std::rand() % sets.size()];
}
