/*
 * Copyright 2010, 2011, 2012, 2013, 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Track.h"
#include "Driver.h"
#include "TrackState.h"
#include <cmath>
#include <cstdlib>
#include <limits>
#include <climits>
#include <algorithm>
#include <cassert>

const int TILE_COLLISION_LEFT       = 0x01;
const int TILE_COLLISION_RIGHT      = 0x02;
const int TILE_COLLISION_UP         = 0x04;
const int TILE_COLLISION_DOWN       = 0x08;
const int TILE_COLLISION_LEFT_UP    = 0x10;
const int TILE_COLLISION_RIGHT_UP   = 0x20;
const int TILE_COLLISION_LEFT_DOWN  = 0x40;
const int TILE_COLLISION_RIGHT_DOWN = 0x80;

const int TILE_COLLISION_CORNERS = TILE_COLLISION_LEFT_UP
                                 | TILE_COLLISION_RIGHT_UP
                                 | TILE_COLLISION_LEFT_DOWN
                                 | TILE_COLLISION_RIGHT_DOWN;

const double TIME_STATE_DURATION = 0.8;

// angle to +- PI 
static double angle2range(double angle)
{
	return angle - 2 * PI * std::floor(angle / (2 * PI)) - PI;
}


Track::Track(int laps, int numCars, const std::vector<Driver*>& drivers, const Map& map, bool carCollisionsEnabled)
:	finishX(map.getFinishX()),
	finishY(map.getFinishY()),
	carsLeft(0),
	tiles(),
	trackLength(map.getRoadCount()),
	laps(laps),
	carCollisionsEnabled(carCollisionsEnabled),
	time(0.0),
	timeState(TIME_STATE_READY)
{
	for (int y = 0; y < MAX_MAP_Y; y++)
	{
		for (int x = 0; x < MAX_MAP_X; x++)
		{
			switch (map[y][x])
			{
			case ROAD_HORIZONTAL:
				tiles[y][x].collisionFlags = TILE_COLLISION_UP
				                           | TILE_COLLISION_DOWN; 
				break;
			case ROAD_VERTICAL:
				tiles[y][x].collisionFlags = TILE_COLLISION_LEFT
				                           | TILE_COLLISION_RIGHT;
				break;
			case ROAD_LEFT_UP:
				tiles[y][x].collisionFlags = TILE_COLLISION_LEFT_UP
				                           | TILE_COLLISION_RIGHT
				                           | TILE_COLLISION_DOWN;
				break;
			case ROAD_LEFT_DOWN:
				tiles[y][x].collisionFlags = TILE_COLLISION_LEFT_DOWN
				                           | TILE_COLLISION_RIGHT
				                           | TILE_COLLISION_UP; 
				break;
			case ROAD_RIGHT_UP:
				tiles[y][x].collisionFlags = TILE_COLLISION_RIGHT_UP
				                           | TILE_COLLISION_LEFT
				                           | TILE_COLLISION_DOWN;
				break;
			case ROAD_RIGHT_DOWN: 
				tiles[y][x].collisionFlags = TILE_COLLISION_RIGHT_DOWN
				                           | TILE_COLLISION_LEFT
				                           | TILE_COLLISION_UP; 
				break;
			default:
				tiles[y][x].collisionFlags = 0; 
				break;
			}
		}
	}

	int x = finishX;
	int y = finishY;
	int finishDir = map.getFinishDir();
	int prevDir = finishDir;
	int n = 0;
	do
	{
		tiles[y][x].orderNum = n++;
		int dir = map[y][x];
		switch (prevDir)
		{
		case LEFT:
			dir &= ~RIGHT;
			break;
		case RIGHT:
			dir &= ~LEFT;
			break;
		case UP:
			dir &= ~DOWN;
			break;
		case DOWN:
			dir &= ~UP;
			break;
		}
		tiles[y][x].dir = dir;
		switch (dir)
		{
		case LEFT:
			x--;
			break;
		case RIGHT:
			x++;
			break;
		case UP:
			y--;
			break;
		case DOWN:
			y++;
			break;
		}
		prevDir = dir;
	} while (x != finishX || y != finishY);
	
	
	cars.resize(numCars);
	
	Vector finishMiddle(finishX * TILE_SIZE + TILE_SIZE / 2.0, 
	                  finishY * TILE_SIZE + TILE_SIZE / 2.0);
	for (int i = 0; i < numCars; i++)
	{
		cars[i].position.x = numCars > 1 ? ((i % 2) ? 1 : -1) * TILE_SIZE * 0.2 : 0;
		cars[i].position.y =  9.0 + (i / 2) * 11.0;
		
		switch (finishDir)
		{
		case LEFT:
			cars[i].position.rotate(-PI_HALF);
			++cars[i].position.x; // Some kind of rounding error.
			cars[i].rotation = -PI_HALF;
			break;
		case RIGHT:
			cars[i].position.rotate(PI_HALF);
			cars[i].rotation = PI_HALF;
			break;
		case UP:
			//cars[i].position.rotate(0.0);
			cars[i].rotation = 0.0;
			break;
		case DOWN:
			cars[i].position.rotate(PI);
			cars[i].rotation = PI;
			break;
		}
		
		cars[i].position += finishMiddle;
		
		cars[i].velocity = Vector(0, 0);
		cars[i].rotationVelocity = 0.0;
		cars[i].driver = drivers[i];
		cars[i].lap = 0;
		cars[i].tileOrderNum = trackLength;
		cars[i].posFinished = 0;
		cars[i].timeFinished = 0.0;
	}
}

int Track::getFinishX() const
{
	return finishX;
}
int Track::getFinishY() const
{
	return finishY;
}

// These values needs testing!
const double CAR_CAR_FRICTION = 0.03;
const double CAR_WALL_FRICTION = 0.0078;
const double SIDE_WHEEL_FRICTION = 320.0;
const double ROLL_WHEEL_FRICTION = 5.0;
const double BRAKE_WHEEL_FRICTION = 160;

const double TURN_ANGLE = 0.41;
const double GAS_SPEED = 0.50 * 50;
const double BRAKE_SPEED = 0.40 * 50;

bool carFinishSorter(const Track::Car* car1, const Track::Car* car2)
{
	return car1->timeFinished < car2->timeFinished;
}

void Track::update(double dt, Uint32* input)
{
	time += dt;
	
	if (timeState < TIME_STATE_GO)
	{
		if (time > TIME_STATE_DURATION)
		{
			++timeState;
			time = 0.0;
		}
		return;
	}
	else if (timeState == TIME_STATE_GO)
	{
		if (time > 2 * TIME_STATE_DURATION)
		{
			++timeState;
		}
	}
	
	std::vector<Car>::iterator c1;
	std::vector<Car>::iterator c2;
	if (carCollisionsEnabled)
	{
		for (c1 = cars.begin(); c1 != cars.end(); ++c1)
		{
			for (c2 = c1 + 1; c2 != cars.end(); ++c2)
			{
				Vector c1c2Vec = c2->position - c1->position;
				
				double distance = c1c2Vec.length();
				if (distance <= 2 * CAR_RADIUS)
				{
					// relative velocity
					Vector relVel = c2->velocity - c1->velocity;

					double pt = findCollisionTime(c1c2Vec, relVel, 2 * CAR_RADIUS);
					
					if (-pt > dt)
					{
						Vector v = c1c2Vec;
						v.normalize();
						v *= 0.1;
						c1->position -= v;
						c2->position += v;
					}
					
					// calculate the collision direction
					Vector colDir(c1c2Vec.x + relVel.x * pt, c1c2Vec.y + relVel.y * pt);
					
					double colAngle = Vector::angle(colDir, VECTOR_Y);
					
					// rotate the velocity in the collision direction 
					c1->velocity.rotate(colAngle);
					c2->velocity.rotate(colAngle);
					
					// the velocity in the y direction will be swapped 
					// due to the momentum. It's as easy as this because
					// the masses of the two cars is always the same
					std::swap(c1->velocity.y, c2->velocity.y);
					
					// the velocity in the x direction will change the
					// rotation speed
					double friction = CAR_CAR_FRICTION * std::fabs(c2->velocity.y - c1->velocity.y);
					if (friction > 1.0)
					{
						friction = 1.0;
					}
					double rotChange1 = (c2->rotationVelocity - c1->rotationVelocity) * 0.5;
					double rotChange2 = (c1->velocity.x - c2->velocity.x) * 0.5;

					c1->rotationVelocity = friction * (rotChange2 - rotChange1) + (1.0 - friction) * c1->rotationVelocity;
					c2->rotationVelocity = friction * (rotChange2 + rotChange1) + (1.0 - friction) * c2->rotationVelocity;

					// change the rotation back to world space
					c1->velocity.rotate(-colAngle);
					c2->velocity.rotate(-colAngle);
					
				}
			}
		}
	}
	
	std::vector<Car*> finishedCars;
	for (c1 = cars.begin(); c1 != cars.end(); ++c1)
	{
		int tilex = (int) c1->position.x / TILE_SIZE;
		int tiley = (int) c1->position.y / TILE_SIZE;
		
		if (tilex < 0 || tilex >= MAX_MAP_X
		 || tiley < 0 || tiley >= MAX_MAP_Y)
		{
			// this should not be possible
			continue;
		}
		
		int colFlags = tiles[tiley][tilex].collisionFlags;
		
		// Corner collisions
		if (colFlags & TILE_COLLISION_CORNERS)
		{
			Vector prevPos = c1->position;
			Vector prevVel = c1->velocity;
			double prevRotVel = c1->rotationVelocity;
			Vector tilePos(tilex * TILE_SIZE, tiley * TILE_SIZE);
			if (colFlags & TILE_COLLISION_LEFT_UP)
			{
				handleCarCornerCollision(*c1, tilePos + Vector(0, 0));
			}
			else if (colFlags & TILE_COLLISION_RIGHT_UP)
			{
				handleCarCornerCollision(*c1, tilePos + Vector(TILE_SIZE, 0));
			}
			else if (colFlags & TILE_COLLISION_LEFT_DOWN)
			{
				handleCarCornerCollision(*c1, tilePos + Vector(0, TILE_SIZE));
			}
			else if (colFlags & TILE_COLLISION_RIGHT_DOWN)
			{
				handleCarCornerCollision(*c1, tilePos + Vector(TILE_SIZE, TILE_SIZE));
			}
			
			// When a car that has contact with a wall reaches the corner
			// the corner collision will push the car back into the previous
			// tile so the car gets stuck. To avoid this we need to detect
			// when this happens and in that case undo the corner collision 
			// and do wall collision on that tile instead.
			int tilex2 = (int) c1->position.x / TILE_SIZE;
			int tiley2 = (int) c1->position.y / TILE_SIZE;
			if (tilex != tilex2 || tiley != tiley2)
			{
				int colFlags2 = tiles[tiley2][tilex2].collisionFlags;
				if (((colFlags & TILE_COLLISION_LEFT_UP)    && colFlags2 & (tilex != tilex2 ? TILE_COLLISION_UP   : TILE_COLLISION_LEFT))  ||
				    ((colFlags & TILE_COLLISION_LEFT_DOWN)  && colFlags2 & (tilex != tilex2 ? TILE_COLLISION_DOWN : TILE_COLLISION_LEFT))  ||
				    ((colFlags & TILE_COLLISION_RIGHT_UP)   && colFlags2 & (tilex != tilex2 ? TILE_COLLISION_UP   : TILE_COLLISION_RIGHT)) ||
				    ((colFlags & TILE_COLLISION_RIGHT_DOWN) && colFlags2 & (tilex != tilex2 ? TILE_COLLISION_DOWN : TILE_COLLISION_RIGHT)))
				{
					c1->position = prevPos;
					c1->velocity = prevVel;
					c1->rotationVelocity = prevRotVel;
					colFlags = colFlags2;
				}
			}
		}
		
		// handle collision with walls

		Vector posInTile(c1->position.x - tilex * TILE_SIZE, c1->position.y - tiley * TILE_SIZE);
		
		// collision with walls we know that the collisions on one of
		// the axis. We only have to flip the velocity on the that axis
		// but to make sure we don't get stuck in walls or something we
		// use std::fabs to make sure the velocity is flipped in the
		// correct direction
		// We also place the car so that it does not overlap the wall.
		// This is to prevent cars from going through walls by standing 
		// against the wall and try to accelerate.
		if (colFlags & TILE_COLLISION_LEFT && CAR_RADIUS > posInTile.x)
		{
			c1->velocity.x = std::fabs(c1->velocity.x);
			c1->position.x = tilex * TILE_SIZE + CAR_RADIUS;
			
			double friction = CAR_WALL_FRICTION * std::fabs(c1->velocity.x);
			if (friction > 1.0)
			{
				friction = 1.0;
			}
			c1->rotationVelocity = friction * (c1->velocity.y / CAR_RADIUS) * 0.5 + (1.0 - friction) * c1->rotationVelocity;
		} 
		else if (colFlags & TILE_COLLISION_RIGHT && TILE_SIZE - CAR_RADIUS < posInTile.x)
		{
			c1->velocity.x = -std::fabs(c1->velocity.x);
			c1->position.x = tilex * TILE_SIZE + TILE_SIZE - CAR_RADIUS;
			
			double friction = CAR_WALL_FRICTION * std::fabs(c1->velocity.x);
			if (friction > 1.0)
			{
				friction = 1.0;
			}
			c1->rotationVelocity = friction * (-c1->velocity.y / CAR_RADIUS) * 0.5 + (1.0 - friction) * c1->rotationVelocity;
		}

		if (colFlags & TILE_COLLISION_UP && CAR_RADIUS > posInTile.y)
		{
			c1->velocity.y = std::fabs(c1->velocity.y);
			c1->position.y = tiley * TILE_SIZE + CAR_RADIUS;
			
			double friction = CAR_WALL_FRICTION * std::fabs(c1->velocity.y);
			if (friction > 1.0)
			{
				friction = 1.0;
			}
			c1->rotationVelocity = friction * (-c1->velocity.x / CAR_RADIUS) * 0.5 + (1.0 - friction) * c1->rotationVelocity;
		} 
		else if (colFlags & TILE_COLLISION_DOWN && TILE_SIZE < posInTile.y + CAR_RADIUS)
		{
			c1->velocity.y = -std::fabs(c1->velocity.y);
			c1->position.y = tiley * TILE_SIZE + TILE_SIZE - CAR_RADIUS;
			
			double friction = CAR_WALL_FRICTION * std::fabs(c1->velocity.y);
			if (friction > 1.0)
			{
				friction = 1.0;
			}
			c1->rotationVelocity = friction * (c1->velocity.x / CAR_RADIUS) * 0.5 + (1.0 - friction) * c1->rotationVelocity;
		}
		
		
		// count laps
		
		bool finished = false;
		double timeSinceFinish = 0.0;
		if (tiles[tiley][tilex].orderNum == 0)
		{
			switch (tiles[finishY][finishX].dir)
			{
				case LEFT:
					finished = (posInTile.x < TILE_SIZE * 0.5);
					timeSinceFinish = (TILE_SIZE * 0.5 - posInTile.x) / -c1->velocity.x;
					break;
				case RIGHT:
					finished = (posInTile.x > TILE_SIZE * 0.5);
					timeSinceFinish = (posInTile.x - TILE_SIZE * 0.5) / c1->velocity.x;
					break;
				case UP:
					finished = (posInTile.y < TILE_SIZE * 0.5);
					timeSinceFinish = (TILE_SIZE * 0.5 - posInTile.y) / -c1->velocity.y;
					break;
				case DOWN:
					finished = (posInTile.y > TILE_SIZE * 0.5);
					timeSinceFinish = (posInTile.y - TILE_SIZE * 0.5) / c1->velocity.y;
					break;
			}
			if (!finished)
			{
				if (c1->tileOrderNum == 0)
				{
					c1->lap--;
					c1->tileOrderNum = trackLength;
				}
				else if (c1->tileOrderNum == trackLength - 1)
				{
					c1->tileOrderNum++;
				}
			}
		}
		if (finished)
		{
			if (c1->tileOrderNum > 1)
			{
				c1->lap++;
				c1->tileOrderNum = 0;
				if (c1->lap > laps)
				{
					// This car has finished the race!

					// Calculate finish time.
					if (timeSinceFinish < 0)
					{
						timeSinceFinish = 0;
					}
					else if (timeSinceFinish > dt)
					{
						timeSinceFinish = dt;
					}

					c1->timeFinished = time - timeSinceFinish;
					
					// Add car to vector so that finish 
					// position can be calculated correctly.
					finishedCars.push_back(&*c1);
				}
			}
		}
		else if (c1->tileOrderNum + 1 == tiles[tiley][tilex].orderNum)
		{
			c1->tileOrderNum++;
		}
		else if (c1->tileOrderNum - 1 == tiles[tiley][tilex].orderNum)
		{
			c1->tileOrderNum--;
		}
	}
	
	// Calculate finish positions.
	if (!finishedCars.empty())
	{
		std::sort(finishedCars.begin(), finishedCars.end(), carFinishSorter);
		std::vector<Car*>::iterator it;
		for (it = finishedCars.begin(); it != finishedCars.end(); ++it)
		{
			(*it)->posFinished = ++carsLeft;
		}
	}
	
	for (c1 = cars.begin(); c1 != cars.end(); ++c1)
	{
		// handle wheel friction
		Vector frontWheelVelocity(c1->velocity);
		frontWheelVelocity.rotate(-c1->rotation);
		Vector backWheelVelocity(c1->velocity);
		backWheelVelocity.rotate(-c1->rotation);
		
		frontWheelVelocity.x += CAR_RADIUS * c1->rotationVelocity;
		backWheelVelocity.x -= CAR_RADIUS * c1->rotationVelocity;
		
		int carId = std::distance(cars.begin(), c1);
		
		int driverFlags = DRIVER_BRAKE|DRIVER_ACCELERATE;
		if (c1->posFinished == 0)
		{
			if (c1->driver && (!input || c1->driver->isAI()))
			{
				driverFlags = c1->driver->drive(carId, *this);
				if (input)
				{
					 *input = (*input & ~(0xF << (4 * carId))) | (driverFlags << (4 * carId));
				}
			}
			else if (input)
			{
				driverFlags = (*input >> (4 * carId)) & 0xF;
			}
		}
		
		double turnAngle = 0.0;
		if (DRIVER_LEFT & driverFlags)
		{
			turnAngle -= TURN_ANGLE;
		}
		if (DRIVER_RIGHT & driverFlags)
		{
			turnAngle += TURN_ANGLE;
		}
		
		double backWheelRollFriction = ROLL_WHEEL_FRICTION;
		double backWheelSpeedAcc = 0.0;
		
		if (driverFlags & DRIVER_BRAKE)
		{
			if (driverFlags & DRIVER_ACCELERATE) // accelerate + reverse = brake
			{
				backWheelRollFriction = BRAKE_WHEEL_FRICTION;
			}
			else if (backWheelVelocity.y > -BRAKE_SPEED * dt) // reverse
			{
				backWheelSpeedAcc = -BRAKE_SPEED;
			}
			else // brake
			{
				backWheelRollFriction = BRAKE_WHEEL_FRICTION;
			}
		}
		else if (driverFlags & DRIVER_ACCELERATE)
		{
			if (backWheelVelocity.y < GAS_SPEED * dt) // forward
			{
				backWheelSpeedAcc = GAS_SPEED;
			}
			else // brake
			{
				backWheelRollFriction = BRAKE_WHEEL_FRICTION;
			}
		}
		
		// the front wheel can have different angles
		frontWheelVelocity.rotate(-turnAngle);
		if (std::fabs(frontWheelVelocity.x) > SIDE_WHEEL_FRICTION * dt)
		{
			frontWheelVelocity.x += (frontWheelVelocity.x > 0.0 ? -1.0 : 1.0) * SIDE_WHEEL_FRICTION * dt;
		}
		else
		{
			frontWheelVelocity.x = 0;
		}
		
		if (std::fabs(frontWheelVelocity.y) > ROLL_WHEEL_FRICTION * dt)
		{
			frontWheelVelocity.y += (frontWheelVelocity.y > 0.0 ? -1.0 : 1.0) * ROLL_WHEEL_FRICTION * dt;
		}
		else
		{
			frontWheelVelocity.y = 0;
		}
		frontWheelVelocity.rotate(turnAngle);
		
		if (std::fabs(backWheelVelocity.x) > SIDE_WHEEL_FRICTION * dt)
		{
			backWheelVelocity.x += (backWheelVelocity.x > 0.0 ? -1.0 : 1.0) * SIDE_WHEEL_FRICTION * dt;
		}
		else
		{
			backWheelVelocity.x = 0;
		}
		
		if (std::fabs(backWheelVelocity.y) > backWheelRollFriction * dt)
		{
			backWheelVelocity.y += (backWheelVelocity.y > 0.0 ? -1.0 : 1.0) * backWheelRollFriction * dt;
		}
		else
		{
			backWheelVelocity.y = 0;
		}
		c1->velocity.y = (frontWheelVelocity.y + backWheelVelocity.y) * 0.5 - backWheelSpeedAcc * dt;
		c1->velocity.x = (frontWheelVelocity.x + backWheelVelocity.x) * 0.5;
		c1->rotationVelocity = (frontWheelVelocity.x - backWheelVelocity.x) * 0.5 / CAR_RADIUS; // ?
		
		// avoid that car has too high speed.
		if (c1->velocity.y < -CAR_SPEED_MAX)
		{
			c1->velocity.y = -CAR_SPEED_MAX;
		}
		else if (c1->velocity.y > CAR_SPEED_MAX)
		{
			c1->velocity.y = CAR_SPEED_MAX;
		}
		
		c1->velocity.rotate(c1->rotation);
		
		// update positions
		c1->position.x += dt * c1->velocity.x;
		c1->position.y += dt * c1->velocity.y;
		c1->rotation += dt * c1->rotationVelocity;
	}
}

void Track::handleCarCornerCollision(Car& car, const Vector& corner)
{
	Vector carCornerVector = car.position - corner;

	if (carCornerVector.lengthSquared() < CAR_RADIUS_SQUARED)
	{		
		// Car collisions can affect the velocity direction so we need 
		// to make sure we use the shortest collision time to avoid
		// that the car is teleported to the other side of the wall.
		double pt1 = findCollisionTime(carCornerVector, car.velocity, CAR_RADIUS);
		double pt2 = findCollisionTime(carCornerVector, -car.velocity, CAR_RADIUS);
		double pt;
		if (pt1 >= pt2)
		{
			pt = pt1;
		}
		else
		{
			pt = -pt2;
		}

		Vector colDir(carCornerVector.x + car.velocity.x * pt, carCornerVector.y + car.velocity.y * pt);
		double colAngle = Vector::angle(colDir, VECTOR_Y);
		
		car.velocity.rotate(colAngle);
		car.position.rotate(colAngle);
		
		// move car to y collision point
		car.position.y += pt * car.velocity.y;
		
		car.velocity.y = std::fabs(car.velocity.y);
		
		double friction = CAR_WALL_FRICTION * std::fabs(car.velocity.y);
		if (friction > 1.0)
		{
			friction = 1.0;
		}
		
		car.rotationVelocity = friction * (-car.velocity.x  / CAR_RADIUS) + (1.0 - friction) * car.rotationVelocity;
		car.position.rotate(-colAngle);
		car.velocity.rotate(-colAngle);
	}
}

double Track::findCollisionTime(const Vector& vec, 
                                const Vector& velocity,
                                double distance)
{
	// Calculate how much time pt have passed since the collision.
	// To do this we use an equation that says that the
	// distance beetween the two points should be equal to radius
	// because this is when the collision occured. 
	// sqrt((vec.x+vec.x*dt)^2+(vec.y+vec.y*dt)^2)=distance
	// Solving this on paper we get an quadratic equation
	// that can be solved by the quadratic formula
	// ax^2 + bx + c = 0
	// x = -(b/2a) ± sqrt((b^2)/(4a^2)-c/a)
	double a = velocity.x * velocity.x + velocity.y * velocity.y;
	double b = 2 * (vec.x * velocity.x + vec.y * velocity.y);
	double c = vec.x * vec.x + vec.y * vec.y - distance * distance;
	
	// avoid division by zero
	if (std::fabs(a) < std::numeric_limits<double>::epsilon())
	{
		a = std::numeric_limits<double>::epsilon();
	}
	
	// f1 = -(b/2a)
	double f1 = -(b / (2 * a));
	// f2 = sqrt((b^2)/(4a^2)-c/a)
	double f2 = std::sqrt((b * b) / ((4 * a) * a) - c / a);
	// We have two solutions to the equation, f1+f2 and f1-f2
	// but we are only interrested in the negative one
	double dt = std::min(f1 + f2, f1 - f2);

	return dt;
}

const Vector& Track::getCarPosition(int carId) const
{
	return cars[carId].position;
}
double Track::getCarRotation(int carId) const
{
	return cars[carId].rotation;
}

int Track::getCarLap(int carId) const
{
	return cars[carId].lap;
}

double Track::getCarProgress(int carId) const
{
	if (cars[carId].posFinished)
	{
		return INT_MAX - cars[carId].posFinished;
	}
	
	int tilex = (int) cars[carId].position.x / TILE_SIZE;
	int tiley = (int) cars[carId].position.y / TILE_SIZE;
	Vector posInTile(cars[carId].position.x - tilex * TILE_SIZE, 
	                 cars[carId].position.y - tiley * TILE_SIZE);

	double progressInTile = 0.0;
	switch (tiles[tiley][tilex].dir)
	{
	case LEFT:
		progressInTile = TILE_SIZE - posInTile.x;
		break;
	case RIGHT:
		progressInTile = posInTile.x;
		break;
	case UP:
		progressInTile = TILE_SIZE - posInTile.y;
		break;
	case DOWN:
		progressInTile = posInTile.y;
		break;
	}
	
	// The tileOrderNum is updated once after the collitions and are
	// used later for the drivers. After that the positions are updated
	// so the tileOrderNum can be old. To avoid wrong positions we
	// calculate the updated tileOrderNum here.
	int ton = cars[carId].tileOrderNum;
	if (ton + 1 == tiles[tiley][tilex].orderNum)
	{
		ton++;
	}
	else if (cars[carId].tileOrderNum - 1 == tiles[tiley][tilex].orderNum)
	{
		ton--;
	}
	
	int tiles = (cars[carId].lap - 1) * trackLength + ton;
	return tiles * TILE_SIZE + progressInTile;
}

static void vdistance(const Vector& p1, const Vector& p2, double& minDistance, double& angle)
{
	Vector v = p2 - p1;
	double distance = v.length();
	if (distance < minDistance)
	{
		minDistance = distance;
		angle = Vector::angle(v, VECTOR_Y);
	}
}

double Track::getCarDistanceNearestWall(int carId, double& angle) const
{
	Vector p = cars[carId].position;
	int tilex = (int) p.x / TILE_SIZE;
	int tiley = (int) p.y / TILE_SIZE;
	Vector posInTile(cars[carId].position.x - tilex * TILE_SIZE, 
	                 cars[carId].position.y - tiley * TILE_SIZE);
	int colFlags = tiles[tiley][tilex].collisionFlags;
	
	angle = 0;
	double distance = TILE_SIZE;
	if (colFlags & TILE_COLLISION_LEFT)
	{
		vdistance(posInTile, Vector(0, posInTile.y), distance, angle);
	}
	if (colFlags & TILE_COLLISION_RIGHT)
	{
		vdistance(posInTile, Vector(TILE_SIZE, posInTile.y), distance, angle);
	}
	if (colFlags & TILE_COLLISION_UP)
	{
		vdistance(posInTile, Vector(posInTile.x, 0), distance, angle);
	}
	if (colFlags & TILE_COLLISION_DOWN)
	{
		vdistance(posInTile, Vector(posInTile.x, TILE_SIZE), distance, angle);
	}
	
	if (colFlags & TILE_COLLISION_LEFT_UP)
	{
		vdistance(posInTile, Vector(0, 0), distance, angle);
	}
	if (colFlags & TILE_COLLISION_LEFT_DOWN)
	{
		vdistance(posInTile, Vector(0, TILE_SIZE), distance, angle);
	}
	if (colFlags & TILE_COLLISION_RIGHT_UP)
	{
		vdistance(posInTile, Vector(TILE_SIZE, 0), distance, angle);
	}
	if (colFlags & TILE_COLLISION_RIGHT_DOWN)
	{
		vdistance(posInTile, Vector(TILE_SIZE, TILE_SIZE), distance, angle);
	}
	angle = angle2range(angle + cars[carId].rotation);

	return distance;
}

double Track::getCarTileAngle(int carId) const
{
	double angle = 0.0;
	
	int x = (int) cars[carId].position.x / TILE_SIZE;
	int y = (int) cars[carId].position.y / TILE_SIZE;
	Vector posInTile(cars[carId].position.x - x * TILE_SIZE,
	                 cars[carId].position.y - y * TILE_SIZE);
		
	switch (tiles[y][x].dir)
	{
	case LEFT:
		angle = PI / 2;
		break;
	case RIGHT:
		angle = 3 * PI / 2;
		break;
	case UP:
		angle = 0;
		break;
	case DOWN:
		angle = PI;
		break;
	}
	
	return angle2range(PI - (angle + cars[carId].rotation));
}

bool Track::finished() const
{
	return carsLeft == (int) cars.size();
}

int Track::getTrackLength() const
{
	return trackLength;
}

double Track::getCarForwardSpeed(int carId) const
{
	Vector v = cars[carId].velocity;
	v.rotate(-cars[carId].rotation);
	return -v.y;
}

int Track::getFinishDir() const
{
	return tiles[finishY][finishX].dir;
}

double Track::getBestDrivingAngle(int carId, double safeMargin, double maxDepth) const
{
	double minAngle = cars[carId].rotation - 0.5 * PI;
	double maxAngle = cars[carId].rotation + 0.5 * PI;
	
	Vector p = getBestDrivingPoint(cars[carId], CAR_RADIUS + safeMargin, maxDepth, minAngle, maxAngle);
	Vector v = p - cars[carId].position;

	return angle2range(Vector::angle(v, VECTOR_Y) + cars[carId].rotation);
}


// shoots two rays from the car,
Vector Track::rayWallTest(const Car& car, double angle, double radius) const
{
	Vector pos1(-radius, 0);
	pos1.rotate(angle);
	Vector pos2(radius, 0);
	pos2.rotate(angle);
	pos1 = rayWallTest(car.position + pos1, angle);
	pos2 = rayWallTest(car.position + pos2, angle);
	
	if ((pos1 - car.position).length() < (pos2 - car.position).length())
	{
		return pos1;
	}
	else
	{
		return pos2;
	}
}

// shoots one ray in the angle direction
Vector Track::rayWallTest(Vector pos, double angle) const
{
	Vector dir = -VECTOR_Y;
	dir.rotate(angle);
	
	int xColFlag;
	int yColFlag;
	int xTileChange;
	int yTileChange;
	int xWrap;
	int yWrap;
	
	if (dir.x < 0.0)
	{
		xColFlag = TILE_COLLISION_LEFT;
		xTileChange = -1;
		xWrap = 0;
	}
	else
	{
		xColFlag = TILE_COLLISION_RIGHT;
		xTileChange = 1;
		xWrap = TILE_SIZE;
	}
	
	if (dir.y < 0.0)
	{
		yColFlag = TILE_COLLISION_UP;
		yTileChange = -1;
		yWrap = 0;
	}
	else
	{
		yColFlag = TILE_COLLISION_DOWN;
		yTileChange = 1;
		yWrap = TILE_SIZE;
	}
	
	int tilex = (int) pos.x / TILE_SIZE;
	int tiley = (int) pos.y / TILE_SIZE;
	Vector posInTile(pos.x - tilex * TILE_SIZE, 
	                 pos.y - tiley * TILE_SIZE);

	while (true)
	{
		assert(tilex >= 0);
		assert(tiley >= 0);
		assert(tilex < MAX_MAP_X);
		assert(tiley < MAX_MAP_Y);
		int colFlags = tiles[tiley][tilex].collisionFlags;
		
		double kx = (xWrap - posInTile.x) / dir.x;
		double ky = (yWrap - posInTile.y) / dir.y;

		if (std::fabs(kx) < std::fabs(ky))
		{
			posInTile.y = posInTile.y + dir.y * kx;
			posInTile.x = TILE_SIZE - xWrap;
			tilex += xTileChange;
			if (colFlags & xColFlag)
			{
				break;
			}
		}
		else
		{
			posInTile.x = posInTile.x + dir.x * ky;
			posInTile.y = TILE_SIZE - yWrap;
			tiley += yTileChange;
			if (colFlags & yColFlag)
			{
				break;
			}
		}
	}
	// Back up a little to make sure that we don't end up in the wrong tile.
	posInTile += -0.1 * dir;
		
	return Vector(tilex * TILE_SIZE + posInTile.x,
	              tiley * TILE_SIZE + posInTile.y);
}

Vector Track::rayCarTest(const Car& car, double angle, double radius) const
{
	Vector shortesRayPoint = Vector(2212, 2113);
	double shortestRayLength = 10000000.0; // a huge number
	
	Vector dir = -VECTOR_Y;
	dir.rotate(angle);
	double r = 2 * CAR_RADIUS + radius / 2;
 
	// maybe I do too many test  below   
	std::vector<Car>::const_iterator it;
	for (it = cars.begin(); it != cars.end(); ++it)
	{
		if (&car == &*it)
		{
			continue;
		}
		Vector carVector = car.position - it->position;
		double p = 2.0 * Vector::dot(carVector, dir);
		double q = Vector::dot(carVector, carVector) - r * r;
		if (p > 0.0 && q < 0.0)
		{
			continue;
		}
		double t1 = -0.5 * p;
		double t2Sqr = 0.25 * p * p - q;
		if (t2Sqr < 0)
		{
			continue;
		}
		double t2 = std::sqrt(t2Sqr);
		double t = std::min(t1 + t2, t1 - t2);
		if (t > 0.0)
		{
			Vector rayPoint = car.position + dir * t;
			double rayLength = (rayPoint - car.position).length();
			if (rayLength < shortestRayLength)
			{
				shortestRayLength = rayLength;
				shortesRayPoint = rayPoint;
			}
		}
	}
	return shortesRayPoint;
}

double Track::PointProgress(Vector pos) const
{
	int tilex = (int) pos.x / TILE_SIZE;
	int tiley = (int) pos.y / TILE_SIZE;
	Vector posInTile(pos.x - tilex * TILE_SIZE,
	                 pos.y - tiley * TILE_SIZE);
	const Tile& tile = tiles[tiley][tilex];
	double progressInTile = 0.0;
	switch (tile.dir)
	{
	case LEFT:
		progressInTile = TILE_SIZE - posInTile.x;
		break;
	case RIGHT:
		progressInTile = posInTile.x;
		break;
	case UP:
		progressInTile = TILE_SIZE - posInTile.y;
		break;
	case DOWN:
		progressInTile = posInTile.y;
		break;
	}
	return tile.orderNum * TILE_SIZE + progressInTile;
}

double Track::rayProgress(Vector pos1, Vector pos2) const
{
	// no two points can have more than half the track progress
	double maxProgress = 0.5 * trackLength * TILE_SIZE;
	
	double progress = PointProgress(pos2) - PointProgress(pos1);
	
	if (progress > maxProgress)
	{
		progress -= trackLength * TILE_SIZE;
	}
	else if (progress < -maxProgress)
	{
		progress += trackLength * TILE_SIZE;
	}
	
	return progress;
}

double Track::rayProgress(const Car& car, Vector rayHitpos) const
{
	return rayProgress(car.position, rayHitpos);
}

Vector Track::getBestDrivingPoint(const Car& car, double radius, double maxDepth, double minAngle, double maxAngle) const
{
	Vector bestRayPoint = car.position;
	double bestRayProgress = -10000.0;
	
	while (maxDepth)
	{
		--maxDepth;
		double middleAngle = (minAngle + maxAngle) * 0.5;
		
		Vector rayWallPoint = rayWallTest(car, middleAngle, radius);
		double rayWallDistance = (rayWallPoint - car.position).length();
		
		
		if (carCollisionsEnabled)
		{
			Vector rayCarPoint = rayCarTest(car, middleAngle, radius);
			double rayCarDistance = (rayCarPoint - car.position).length();
		
			if (rayCarDistance < rayWallDistance)
			{
				Vector rayPoint1 = getBestDrivingPoint(car, radius, maxDepth, minAngle, middleAngle);
				Vector rayPoint2 = getBestDrivingPoint(car, radius, maxDepth, middleAngle, maxAngle);
				double rayProgress1 = rayProgress(car, rayPoint1);
				double rayProgress2 = rayProgress(car, rayPoint2);
				double rayCarProgress = rayProgress(car, rayCarPoint);
				if (rayCarProgress > rayProgress1 && rayCarProgress > rayProgress2)
				{
					return rayCarPoint;
				}
				else if (rayProgress1 > rayProgress2)
				{
					return rayPoint1;
				}
				else
				{
					return rayPoint2;
				}
			}
		}
		
		int tilex = (int) rayWallPoint.x / TILE_SIZE;
		int tiley = (int) rayWallPoint.y / TILE_SIZE;
		Vector dir;
		switch (tiles[tiley][tilex].dir)
		{
		case LEFT:
			dir = Vector(-1, 0);
			break;
		case RIGHT:
			dir = Vector(1, 0);
			break;
		case UP:
			dir = Vector(0, -1);
			break;
		case DOWN:
			dir = Vector(0, 1);
			break;
		}
		dir.rotate(-middleAngle);
		(dir.x > 0 ? minAngle : maxAngle) = middleAngle;
		
		double rayWallProgress = rayProgress(car, rayWallPoint);
		
		if (rayWallProgress > bestRayProgress)
		{
			bestRayProgress = rayWallProgress;
			bestRayPoint = rayWallPoint;
		}
	}
	return bestRayPoint;
}

void Track::save(TrackState& state) const
{
	state.time = time;
	state.timeState = timeState;
	for (std::size_t i = 0; i < cars.size(); ++i)
	{
		TrackCarState& stateCar = state.cars[i];
		const Car& trackCar = cars[i];
		
		stateCar.velocity         = trackCar.velocity;
		stateCar.position         = trackCar.position;
		stateCar.rotationVelocity = trackCar.rotationVelocity;
		stateCar.rotation         = trackCar.rotation;
		stateCar.tileOrderNum     = trackCar.tileOrderNum;
		stateCar.lap              = trackCar.lap;
		stateCar.posFinished      = trackCar.posFinished;
		stateCar.timeFinished     = trackCar.timeFinished;
	}
}

void Track::restore(const TrackState& state)
{
	time = state.time;
	timeState = state.timeState;
	
	carsLeft = 0;
	
	for (std::size_t i = 0; i < cars.size(); ++i)
	{
		const TrackCarState& stateCar = state.cars[i];
		Car& trackCar = cars[i];
		
		trackCar.velocity         = stateCar.velocity;
		trackCar.position         = stateCar.position;
		trackCar.rotationVelocity = stateCar.rotationVelocity;
		trackCar.rotation         = stateCar.rotation;
		trackCar.tileOrderNum     = stateCar.tileOrderNum;
		trackCar.lap              = stateCar.lap;
		trackCar.posFinished      = stateCar.posFinished;
		trackCar.timeFinished     = stateCar.timeFinished;
		
		if (trackCar.posFinished)
		{
			++carsLeft;
		}
	}
}

void Track::removeCar(unsigned carId)
{
	if (carId < cars.size())
	{
		Car& car = cars[carId];
		if (car.posFinished)
		{
			--carsLeft;
		}
		cars.erase(cars.begin() + carId);
		removedCars.push_back(carId);
	}
}

std::vector<int>& Track::getRemovedCars()
{
	return removedCars;
}

double Track::getRaceTime() const
{
	return timeState < TIME_STATE_GO ? 0 : time;
}

int Track::getTimeState() const
{
	return timeState;
}

double Track::getCarFinishTime(int carId) const
{
	return cars[carId].timeFinished;
}

int Track::getCarCount() const
{
	return cars.size();
}

void Track::finish(double dt, int minUpdates, int maxRealDuration)
{
	int updates = 0;
	Uint32 stop = SDL_GetTicks() + maxRealDuration;
	while (!finished())
	{
		update(dt);
		++updates;
		if (updates >= minUpdates)
		{
			if (stop <= SDL_GetTicks())
			{
				break;
			}
			bool noComputerDrivers = true;
			std::vector<Car>::iterator it;
			for (it = cars.begin(); it != cars.end(); ++it)
			{
				Car& car = *it;
				if (car.posFinished == 0 && car.driver && car.driver->isAI())
				{
					noComputerDrivers = false;
					break;
				}
			}
			if (noComputerDrivers)
			{
				break;
			}
		}
	}
}
