/*
 * Copyright 2010, 2011, 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TRACK_H
#define TRACK_H

#include <vector>
#include "Vector.h"
#include "Map.h"

class Driver;
class TrackState;

const double CAR_SPEED_MAX = 160.0;
const double CAR_RADIUS = 3.5;
const double CAR_RADIUS_SQUARED = CAR_RADIUS * CAR_RADIUS;
const int TRACK_TIME_STEP = 20;

enum
{
	TIME_STATE_READY,
	TIME_STATE_SET,
	TIME_STATE_GO,
	TIME_STATE_TIME
};

class Track
{
public:
	Track(int laps, int numCars, const std::vector<Driver*>& drivers, const Map& map, bool carCollisionsEnabled = true);
	void update(double dt, Uint32* input = 0);
	const Vector& getCarPosition(int carId) const;
	double getCarRotation(int carId) const;
	double getCarForwardSpeed(int carId) const;
	int getCarLap(int carId) const;
	double getCarProgress(int carId) const;
	double getCarDistanceNearestWall(int carId, double& angle) const;
	double getCarTileAngle(int carId) const;
	int getFinishX() const;
	int getFinishY() const;
	bool finished() const;
	int getTrackLength() const;
	int getFinishDir() const;
	double getBestDrivingAngle(int carId, double safeMargin, double maxDepth) const;
	void save(TrackState& state) const;
	void restore(const TrackState& state);
	void removeCar(unsigned carId);
	std::vector<int>& getRemovedCars();
	double getRaceTime() const;
	int getTimeState() const;
	double getCarFinishTime(int carId) const;
	int getCarCount() const;
	void finish(double dt, int minUpdates, int maxRealDuration);
private:
	int finishX;
	int finishY;
	int carsLeft;
	struct Car
	{
		Vector velocity;
		double rotationVelocity;
		Vector position;
		double rotation;
		Driver* driver;
		int tileOrderNum;
		int lap;
		int posFinished;
		double timeFinished;
	};
	friend bool carFinishSorter(const Car*, const Car*);
	std::vector<Car> cars;
	static void handleCarCornerCollision(Car& car, const Vector& corner);
	Vector getBestDrivingPoint(const Car& car, double radius, double maxDepth, double minAngle, double maxAngle) const;
	Vector rayWallTest(const Car& car, double angle, double radius) const;
	Vector rayWallTest(Vector pos, double angle) const;
	Vector rayCarTest(const Car& car, double angle, double radius) const;
	double PointProgress(Vector pos) const;
	double rayProgress(Vector pos1, Vector pos2) const;
	double rayProgress(const Car& car, Vector rayHitpos) const;
	
	// finds the relative time when the vec had length distance 
	// radius from if vec change with velocity.
	static double findCollisionTime(const Vector& vec, 
	                                const Vector& velocity,
	                                double distance);
	struct Tile
	{
		int collisionFlags;
		int orderNum;
		int dir;
	};
	Tile tiles[MAX_MAP_Y][MAX_MAP_X];
	int trackLength;
	int laps;
	bool carCollisionsEnabled;
	double time;
	int timeState;
	std::vector<int> removedCars;
};

#endif
