/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TrackList.h"
#include "GameState.h"
#include "StateManager.h"

const std::size_t MAX_VISIBLE_TRACKS = 10;

TrackList::TrackList(GameState& state)
:	selected(true),
	iTrack(0),
	scrollPos(0),
	gameState(state),
	hoverTrack(-1),
	leftArrowHover(false),
	leftArrowVisible(false),
	rightArrowHover(false),
	rightArrowVisible(false)
{
	const std::vector<TimeTrialTrack>& timeTrialTracks = TimeTrialTrack::loadFromFile();
	
	// One extra for the New Track option.
	std::size_t trackItemCount = timeTrialTracks.size() + 1;
	tracks.reserve(trackItemCount);
	tracks.push_back(Track());
	updateRenderObject(0);

	for (std::size_t i = 0; i < timeTrialTracks.size(); ++i)
	{
		Track t;
		t.timeTrialTrack = timeTrialTracks[i];
		tracks.push_back(t);
		updateRenderObject(i + 1);
	}
		
	updateRenderObjects();
}

TimeTrialTrack* TrackList::onEvent(const SDL_Event& event)
{
	switch (event.type)
	{
	case SDL_MOUSEBUTTONDOWN:
		switch (event.button.button)
		{
		case SDL_BUTTON_WHEELUP:
			scrollLeft();
			return 0;
		case SDL_BUTTON_WHEELDOWN:
			scrollRight();
			return 0;
		}
	}
	
	TimeTrialTrack* ttt = 0;
		
	MenuEvent me = StateManager::getInstance().getMenuEvent(event);
	switch (me.type)
	{
	case MENU_MOUSE_MOVE:
		mouseMove(me.x, me.y);
		break;
	case MENU_MOUSE_CLICK1:
		click(me.x, me.y, ttt);
		break;
	case MENU_INC:
		next();
		break;
	case MENU_DEC:
		prev();
		break;
	case MENU_MIN:
		for (int i = iTrack; i >= 0; --i)
		{
			prev();
		}
		break;
	case MENU_MAX:
		for (std::size_t i = iTrack; i < tracks.size(); ++i)
		{
			next();
		}
		break;
	case MENU_FIRST:
		for (std::size_t i = 0; i < MAX_VISIBLE_TRACKS; ++i)
		{
			if (scrollLeft())
			{
				prev();
			}
		}
		break;
	case MENU_LAST:
		for (std::size_t i = 0; i < MAX_VISIBLE_TRACKS; ++i)
		{
			if (scrollRight())
			{
				next();
			}
		}
		break;
	case MENU_ACTION:
		if (selected)
		{
			ttt = &tracks[iTrack].timeTrialTrack;
		}
		break;
	}
	
	return ttt;
}

TimeTrialTrack* TrackList::getCurrentTrack()
{
	return (selected ? &tracks[iTrack].timeTrialTrack : 0);
}

void TrackList::updateRenderObject(std::size_t i)
{
	Surface box(27, 21);
	box.fill(BOX_BORDER_COLOR);
	box.fill(BOX_SHADOW_COLOR, 1, 1, box.width() - 2, box.height() - 2);
	box.fill(BOX_INNER_COLOR, 2, 2, box.width() - 4, box.height() - 4);
	
	Color color = (i == iTrack && selected ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
	
	Track& track = tracks[i];
	
	if (i == iTrack)
	{
		gameState.removeRenderObject(topTimes);
	}
	
	if (i == 0)
	{
		Surface newTxt("new", SMALL_FONT, color, TEXT_BORDER_COLOR);
		newTxt.draw(box, 6, 3);
		
		Surface trackTxt("track", SMALL_FONT, color, TEXT_BORDER_COLOR);
		trackTxt.draw(box, 3, 11);
	}
	else
	{
		if (i == iTrack && selected)
		{
			Surface s(GAME_WIDTH, GAME_HEIGHT);
			s.fill(DEFAULT_COLORKEY);
			s.setColorKey(DEFAULT_COLORKEY);
			track.timeTrialTrack.drawTopTimes(s, false);
			topTimes.setSurface(s);
			gameState.addRenderObject(topTimes);
		}
		
		const Map& map = track.timeTrialTrack.getMap();
		
		for (int y = 0; y < MAX_MAP_Y; y++)
		{
			for (int x = 0; x < MAX_MAP_X; x++)
			{
				if (map[y][x])
				{
					box.fill(TEXT_BORDER_COLOR, 3 + x * 2, 3 + y * 2, 3, 3);
				}
			}
		}
		
		for (int y = 0; y < MAX_MAP_Y; y++)
		{
			for (int x = 0; x < MAX_MAP_X; x++)
			{
				if (map[y][x])
				{
					int destx = 4 + x * 2;
					int desty = 4 + y * 2;
					
					box.setColor(destx, desty, color);
					if (map[y][x] & LEFT)
					{
						box.setColor(destx - 1, desty, color);
					}
					if (map[y][x] & RIGHT)
					{
						box.setColor(destx + 1, desty, color);
					}
					if (map[y][x] & UP)
					{
						box.setColor(destx, desty - 1, color);
					}
					if (map[y][x] & DOWN)
					{
						box.setColor(destx, desty + 1, color);
					}
				}
			}
		}
		
		box.setColor(4 + map.getFinishX() * 2, 4 + map.getFinishY() * 2, 0xff0000);
	}
	
	track.renderObject.setSurface(box);
}

void TrackList::updateRenderObjects()
{
	if (iTrack >= tracks.size())
	{
		iTrack = tracks.size() - 1;
		updateRenderObject(iTrack);
		selected = false;
	}
	
	if (iTrack < scrollPos)
	{
		scrollPos = iTrack;
	}
	else if (iTrack >= scrollPos + MAX_VISIBLE_TRACKS)
	{
		scrollPos = iTrack - (MAX_VISIBLE_TRACKS - 1);
	}
	
	std::size_t scrollPosEnd = scrollPos + MAX_VISIBLE_TRACKS;
	if (scrollPosEnd > tracks.size())
	{
		scrollPos -= scrollPosEnd - tracks.size();
		if (scrollPos > tracks.size())
		{
			scrollPos = 0;
		}
	}
	int x = 161 - std::min(tracks.size(), MAX_VISIBLE_TRACKS) * 15;
	for (std::size_t i = 0; i < tracks.size(); ++i)
	{
		gameState.removeRenderObject(tracks[i].renderObject);
		if (i >= scrollPos && i < scrollPosEnd)
		{
			RenderObject& ro = tracks[i].renderObject;
			ro.moveTo(x, 216);
			x += ro.getSurface().width() + 3;
			gameState.addRenderObject(ro);
		}
	}
	
	updateLeftArrow(leftArrowHover, scrollPos > 0);
	updateRightArrow(rightArrowHover, scrollPosEnd < tracks.size());
}

void TrackList::updateLeftArrow(bool arrowHover, bool arrowVisible)
{
	if (arrowHover != leftArrowHover || arrowVisible != leftArrowVisible)
	{
		leftArrowHover = arrowHover;
		leftArrowVisible = arrowVisible;
		
		gameState.removeRenderObject(leftArrow);
		if (arrowVisible)
		{
			Color color = (arrowHover ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
			leftArrow.setSurface(Surface("^", BIG_FONT, color, TEXT_BORDER_COLOR).subSurface(0, 0, 7, 5));
			leftArrow.moveTo(2, 224);
			leftArrow.setRotation(-PI_HALF);
			gameState.addRenderObject(leftArrow);
		}
	}
}

void TrackList::updateRightArrow(bool arrowHover, bool arrowVisible)
{
	if (arrowHover != rightArrowHover || arrowVisible != rightArrowVisible)
	{
		rightArrowHover = arrowHover;
		rightArrowVisible = arrowVisible;
		
		gameState.removeRenderObject(rightArrow);
		if (arrowVisible)
		{
			Color color = (arrowHover ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
			rightArrow.setSurface(Surface("^", BIG_FONT, color, TEXT_BORDER_COLOR).subSurface(0, 0, 7, 5));
			rightArrow.moveTo(310, 224);
			rightArrow.setRotation(PI_HALF);
			gameState.addRenderObject(rightArrow);
		}
	}
}

void TrackList::next()
{
	if (selected)
	{
		if (iTrack + 1 < tracks.size() )
		{
			std::size_t old = iTrack;
			++iTrack;
			updateRenderObject(old);
			updateRenderObject(iTrack);
			updateRenderObjects();
		}
	}
	else
	{
		selected = true;
		updateRenderObject(iTrack);
	}
}

void TrackList::prev()
{
	if (selected)
	{
		if (iTrack > 0)
		{
			std::size_t old = iTrack;
			--iTrack;
			updateRenderObject(old);
			updateRenderObject(iTrack);
			updateRenderObjects();
		}
	}
	else
	{
		selected = true;
		updateRenderObject(iTrack);
	}
}

bool TrackList::scrollRight()
{
	if (scrollPos + MAX_VISIBLE_TRACKS < tracks.size())
	{
		++scrollPos;
		if (iTrack < scrollPos)
		{
			++iTrack;
			selected = false;
			updateRenderObject(iTrack - 1);
			gameState.removeRenderObject(topTimes);
		}
		updateRenderObjects();
		return true;
	}
	return false;
}

bool TrackList::scrollLeft()
{
	if (scrollPos > 0)
	{
		--scrollPos;
		if (iTrack >= scrollPos + MAX_VISIBLE_TRACKS)
		{
			--iTrack;
			selected = false;
			updateRenderObject(iTrack + 1);
			gameState.removeRenderObject(topTimes);
		}
		updateRenderObjects();
		return true;
	}
	return false;
}

void TrackList::mouseMove(int x, int y)
{
	std::size_t scrollPosEnd = scrollPos + MAX_VISIBLE_TRACKS;
	updateLeftArrow(leftArrow.isHit(x, y), scrollPos > 0);
	updateRightArrow(rightArrow.isHit(x, y), scrollPosEnd < tracks.size());
	
	std::size_t scrollEnd = std::min(scrollPosEnd, tracks.size());
	for (std::size_t i = scrollPos; i < scrollEnd; ++i)
	{
		if (tracks[i].renderObject.isHit(x, y))
		{
			hoverTrack = i;
			if (iTrack != i || !selected)
			{
				selected = true;
				std::size_t old = iTrack;
				iTrack = i;
				updateRenderObject(old);
				updateRenderObject(iTrack);
			}
			return;
		}
	}
	
	// Not found.
	
	if (hoverTrack != -1 && selected)
	{
		selected = false;
		updateRenderObject(iTrack);
	}
	hoverTrack = -1;
}

void TrackList::click(int x, int y, TimeTrialTrack*& ttt)
{
	if (leftArrowHover)
	{
		scrollLeft();
	}
	else if (rightArrowHover)
	{
		scrollRight();
	}
	else 
	{
		std::size_t scrollEnd = std::min(scrollPos + MAX_VISIBLE_TRACKS, tracks.size());
		for (std::size_t i = scrollPos; i < scrollEnd; ++i)
		{
			if (tracks[i].renderObject.isHit(x, y))
			{
				if (iTrack != i || !selected)
				{
					selected = true;
					std::size_t old = iTrack;
					iTrack = i;
					updateRenderObject(old);
					updateRenderObject(iTrack);
				}
				else
				{
					ttt = &tracks[i].timeTrialTrack;
				}
				return;
			}
		}
		selected = false;
		updateRenderObject(iTrack);
	}
}

void TrackList::setState(int state)
{
	scrollPos = state >> 16;
	std::size_t old = iTrack;
	iTrack = std::min(state & 0xFFFF, int(tracks.size() - 1));
	updateRenderObject(old);
	updateRenderObject(iTrack);
	updateRenderObjects();
}

int TrackList::getState()
{
	return (scrollPos << 16) | (iTrack & 0xFFFF);
}
