/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TracksSettingsMenu.h"
#include "TileSet.h"
#include "StateManager.h"
#include "SettingsMenu.h"
#include <sstream>

TracksSettingsMenu::RangeItem::RangeItem(const std::string& text, int* value1, int* value2, int min, int max, const std::string& unit)
:	value1(value1),
	value2(value2),
	min(min),
	max(max),
	text(text),
	unit(unit)
{
	updateValue();
}

void TracksSettingsMenu::RangeItem::addTo(Menu& menu)
{
	menu.addItem(text, &value);
}

void TracksSettingsMenu::RangeItem::validateTypedValue(const std::string* strValue, std::string& inputString)
{
	if (strValue == &value)
	{
		if (!inputString.empty())
		{
			int i = inputString.size() - 1;
			char& c = inputString[i];
			if (!std::isdigit(c))
			{
				c = ' '; // Avoid finding hyphen.
				if (i == 0 || inputString.find('-') != std::string::npos)
				{
					inputString.resize(i);
				}
				else
				{
					c = '-';
				}
			}
		}
	}
}

void TracksSettingsMenu::RangeItem::onValueUpdate(std::string* strValue)
{
	if (strValue == &value)
	{
		std::istringstream iss(value);
		int newValue1;
		int newValue2;
		if (iss >> newValue1)
		{
			iss.ignore();
			if (!(iss >> newValue2))
			{
				newValue2 = newValue1;
			}
			newValue1 = std::min(std::max(newValue1, min), max);
			newValue2 = std::min(std::max(newValue2, min), max);
			if (newValue1 > newValue2)
			{
				std::swap(newValue1, newValue2);
			}
			
			*value1 = newValue1;
			*value2 = newValue2;
		}
		updateValue();
	}
}
void TracksSettingsMenu::RangeItem::updateValue()
{
	std::ostringstream oss;
	oss << *value1 << '-' << *value2 << unit;
	value = oss.str();
}

void TracksSettingsMenu::RangeItem::inc(Menu& menu)
{
	if (*value1 < *value2)
	{
		++*value1;
	}
	else if (*value2 < max)
	{
		++*value2;
		*value1 = min;
	}
	updateValue();
	menu.notifyValueUpdate(&value);
}

void TracksSettingsMenu::RangeItem::dec(Menu& menu)
{	
	if (*value1 > min)
	{
		--*value1;
	}
	else if (*value2 > min)
	{
		--*value2;
		*value1 = *value2;
	}
	updateValue();
	menu.notifyValueUpdate(&value);
}

void TracksSettingsMenu::RangeItem::maximize(Menu& menu)
{
	if (*value2 < max)
	{
		*value2 = max;
	}
	else
	{
		*value1 = max;
	}
	updateValue();
	menu.notifyValueUpdate(&value);
}

void TracksSettingsMenu::RangeItem::minimize(Menu& menu)
{
	if (*value1 > min)
	{
		*value1 = min;
	}
	else
	{
		*value2 = min;
	}
	updateValue();
	menu.notifyValueUpdate(&value);
}

TracksSettingsMenu::TracksSettingsMenu()
:	trackSettings(Settings::tracks),
	save(false),
	cancel(false)
{
	menu.setTitle("Track Settings");
	
	rangeItems.push_back(RangeItem("Track length", &trackSettings.minLength, &trackSettings.maxLength, 6, 70));
	rangeItems.push_back(RangeItem("Track width", &trackSettings.minWidth, &trackSettings.maxWidth, 2, 10));
	rangeItems.push_back(RangeItem("Track height", &trackSettings.minHeight, &trackSettings.maxHeight, 2, 7));
	rangeItems.push_back(RangeItem("Corners", &trackSettings.minCorners, &trackSettings.maxCorners, 0, 100, "%"));
	rangeItems.push_back(RangeItem("Longest straightaway", &trackSettings.minLongestStraighaway, &trackSettings.maxLongestStraighaway, 1, 8));
	rangeItems.push_back(RangeItem("Longest corner series", &trackSettings.minLongestCornerSeries, &trackSettings.maxLongestCornerSeries, 1, 70));
	
	for (std::vector<RangeItem>::iterator it = rangeItems.begin(); it != rangeItems.end(); ++it)
	{
		it->addTo(menu);
	}
	
	const std::vector<std::string>& tilesetNames = TileSet::getTileSetNames();
	for (std::vector<std::string>::const_iterator it = tilesetNames.begin(); it != tilesetNames.end(); ++it)
	{
		const std::string& name = *it;
		menu.addItem(name, &trackSettings.tilesets[name], 0, 10000);
	}
	updateTilesetInfo();
	
	menu.addItem("Save settings", &save);
	menu.addItem("Cancel", &cancel);
	
	menu.addRenderObjects(this);
}

void TracksSettingsMenu::onEvent(const SDL_Event& event)
{
	StateManager& sm = StateManager::getInstance();
	MenuEvent me = sm.getMenuEvent(event);
	
	std::size_t i = menu.getItem();
	if (i < rangeItems.size() && menu.getItemState() == ITEM_SELECTED)
	{
		
		if (me.type == MENU_INC ||
		    (me.type == MENU_MOUSE_CLICK1 && menu.getItem(me.x, me.y) == int(i)))
		{
			rangeItems[i].inc(menu);
		}
		else if (me.type == MENU_DEC ||
		         (me.type == MENU_MOUSE_CLICK2 && menu.getItem(me.x, me.y) == int(i)))
		{
			rangeItems[i].dec(menu);
		}
		else if (me.type == MENU_MAX)
		{
			rangeItems[i].maximize(menu);
		}
		else if (me.type == MENU_MIN)
		{
			rangeItems[i].minimize(menu);
		}
	}
	
	if (menu.onEvent(me))
	{
		if (save)
		{
			Settings::tracks = trackSettings;
			sm.enter(new SettingsMenu(), -1);
			return;
		}
	}
	else if (me.type == MENU_ESC)
	{
		cancel = true;
	}
	if (cancel)
	{
		sm.enter(new SettingsMenu(), -1);
	}
}

void TracksSettingsMenu::validateTypedValue(const std::string* strValue, std::string& inputString)
{
	std::vector<RangeItem>::iterator it;
	for (it = rangeItems.begin(); it != rangeItems.end(); ++it)
	{
		it->validateTypedValue(strValue, inputString);
	}
}

void TracksSettingsMenu::onValueUpdate(std::string* strValue)
{
	std::vector<RangeItem>::iterator it;
	for (it = rangeItems.begin(); it != rangeItems.end(); ++it)
	{
		it->onValueUpdate(strValue);
	}
}

void TracksSettingsMenu::onValueUpdate(int* intValue)
{
	std::map<std::string, int>::iterator it;
	for (it = trackSettings.tilesets.begin(); it != trackSettings.tilesets.end(); ++it)
	{
		if (intValue == &it->second)
		{
			updateTilesetInfo();
			break;
		}
	}
}

void TracksSettingsMenu::updateTilesetInfo()
{
	int total = 0;
	std::map<std::string, int>::iterator it;
	for (it = trackSettings.tilesets.begin(); it != trackSettings.tilesets.end(); ++it)
	{
		total += it->second;
	}
	
	for (it = trackSettings.tilesets.begin(); it != trackSettings.tilesets.end(); ++it)
	{
		int permille = (total != 0 ? (1000.0 * it->second) / total : 1000.0 / trackSettings.tilesets.size()) + 0.5;
		int percentIntegerPart = permille / 10;
		int percentFractionalPart = permille % 10;
		
		std::ostringstream oss;
		oss << "/" << total << " (" << percentIntegerPart;
		if (percentFractionalPart != 0)
		{
			oss << '.' << percentFractionalPart;
		}
		oss << "%)";
		
		menu.setSuffix(&it->second, oss.str());
	}
}
